/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-2000 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "io_spec.h"
#include "buffer.h"
#include "linklist.h"
#include "util.h"


int GotMagic=0;

int BlockSize=0;
int NumChannels=0;
int SampleRate=0;
int NoPanel=1; /* Panel does not yet exist */
char *WindowTitle="Filter and downsample";


struct SBuf *Buffer;
struct SBuf *OutBuf;

int readpipe=0;
int writepipe=1;

int downsamplefactor=2;

char *filterfile=NULL;

static char **global_argv;

double startfreq=20.0;
double freqstepfactor=1.3;
double endfreq=20000.0;
char *outfilename="analyzer.out";

double outputlength=8.0; /* seconds */
double listendelay=4.0;
double outputampl=.5; /* this factor times maximum possible output range */

double testfreq(double freq)
{
  int outputblocks;
  int listendelayblocks;
  int Cnt;
  int OBCnt;
  int CCnt;
  int BCnt;
  double Val;
  double t=0.0;
  double t2=0.0;
  /*  double rms=0.0;
      int rmscnt=0;*/
  double real=0.0,imag=0.0;
  double multiplier;
  int pointcnt=0;
  int listensamples;

  outputblocks=outputlength*SampleRate/BlockSize;
  listendelayblocks=listendelay*SampleRate/BlockSize;
  
  listensamples=(outputblocks-listendelayblocks)*BlockSize;
  
  for (Cnt=0;Cnt < outputblocks;Cnt++) {
    
    if (!NoPanel) {
      /* handle GUI i/o */
      /*  fdpi_readchars();
	  if (queued_outbuffer_data_available(guilink)) write_queued_outbuffer(guilink); */
    }
    
    /* Read in data from input */
    SBufPrepareHeader(Buffer);
    ReadSBuf(Buffer,readpipe);
    assert(SBufCheckHeader(Buffer));
    
    SBufReallocateIfNeeded(Buffer);
    SBufPrepareFinish(Buffer);
    ReadSBuf(Buffer,readpipe);
    
    if (Buffer->Header->BlockSize != BlockSize) {
      BlockSize=Buffer->Header->BlockSize;
    }
    
    if (Buffer->Header->NumChannels != NumChannels) {
      NumChannels=Buffer->Header->NumChannels;
    }
    
    if (Buffer->Header->SampleRate != SampleRate) {
      SampleRate=Buffer->Header->SampleRate;
    }
    
    SBufCopyHeader(Buffer,OutBuf);
    if (Cnt >= listendelayblocks) {
      /* analyze output */
      for (BCnt=0;BCnt < Buffer->Size;BCnt++) {
	Val=0.0;

	multiplier=1-cos(2*M_PI*((double)pointcnt)/((double)listensamples));

	assert(pointcnt < listensamples);

	for (CCnt=0;CCnt < NumChannels;CCnt++) {
	  Val+=multiplier*Buffer->Data[NumChannels*BCnt+CCnt];
	  if  (Buffer->Data[NumChannels*BCnt+CCnt] >= 32766.0 || Buffer->Data[NumChannels*BCnt+CCnt] <= -32766.0) {
	    fprintf(stderr,"%s: WARNING: Clipped input\n",global_argv[0]);
	  }
	  
	}
	Val/=32767*NumChannels;
	
	real+=Val*sin(2*M_PI*freq*t2);
	imag+=Val*cos(2*M_PI*freq*t2);
	
	t2=t2+1.0/SampleRate;
	pointcnt++;
	
      }
    }
    
    /* write out wave at specified frequency */
    SBufReallocateIfNeeded(OutBuf);
    for (OBCnt=0;OBCnt < OutBuf->Size;OBCnt++) {
      Val=outputampl*32767*sin(2*M_PI*freq*t);
      t=t+1.0/SampleRate;
      for (CCnt=0;CCnt < OutBuf->NumChannels;CCnt++) {
	OutBuf->Data[NumChannels*OBCnt+CCnt]=(int)Val;
      }
      
    }
    SBufPrepare(OutBuf);
    WriteSBuf(OutBuf,writepipe); 
    
  }
  
  /* done with this freq. -- process results */
  /*  rms/=rmscnt;
  rms=sqrt(rms);
  */

  real/=pointcnt;
  imag/=pointcnt;
  
  /*fprintf(stderr,"%d %d\n",pointcnt,listensamples);*/
  
  return sqrt(real*real+imag*imag);
}


int main(int argc,char *argv[])
{
  int Cnt;
  struct queued_outbuffer *guilink=NULL;

  double freq;
  double epsilon;
  double amplitude;

  FILE *outfile;
  
  global_argv=argv;

  for (Cnt=1;argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    else if (!strncmp(argv[Cnt],"-startfreq=",11)) {
      startfreq=strtod(argv[Cnt]+11,NULL);
    }
    else if (!strncmp(argv[Cnt],"-freqstepfactor=",16)) {
      freqstepfactor=strtod(argv[Cnt]+16,NULL);
    }
    else if (!strncmp(argv[Cnt],"-endfreq=",9)) {
      endfreq=strtod(argv[Cnt]+9,NULL);
    } else if (!strncmp(argv[Cnt],"-outfilename=",13)) {
      outfilename=argv[Cnt]+13;
    
    } else if (!strcmp(argv[Cnt],"-nopanel")) {
      NoPanel=1;
    } else if (!strncmp(argv[Cnt],"-title=",7)) {
      WindowTitle=strdup(argv[Cnt]+7);
    }
    else if (!strncmp(argv[Cnt],"-outputlength=",14)) {
      outputlength=strtod(argv[Cnt]+14,NULL);
    }
    else if (!strncmp(argv[Cnt],"-listendelay=",13)) {
      outputlength=strtod(argv[Cnt]+13,NULL);
    }

  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   aglaophone format input (connect to soundcard record output)\n"
	    "        #1   aglaophone format output (connect to soundcard playback input)\n"
	    "        #2   error messages (do not connect)\n"
      	    "\n"
	    "Usage:\n"
	    "   %s -magic -startfreq=%f -freqstepfactor=%f -endfreq=%f\n"
	    "           -outfilename=%s -outputlength=%f -listendelay=%f\n"
	    /*"         -nopanel \"-title=Filter and downsample\"  \n"*/
	    "\n"
	    "   All frequencies are given in Hz.\n"
	    ,argv[0],startfreq,freqstepfactor,endfreq,outfilename,outputlength,listendelay);
    exit(1);
  }
  
  if (!NoPanel) {
    /* initfdpi(argv[0],&guilink);
    if (!guilink) {
      fprintf(stderr,"%s: Error creating GUI\n",argv[0]);
      exit(1);
      }*/
  }

  outfile=fopen(outfilename,"w");
  if (!outfile) {
    fprintf(stderr,"%s: Error opening output file %s for write\n",argv[0],outfilename);
    exit(1);
  }
  
  Buffer=CreateSBuf();
  OutBuf=CreateSBuf();

  /* Read in one block from input, so that we can find sample rate, etc. */
  SBufPrepareHeader(Buffer);
  ReadSBuf(Buffer,readpipe);
  assert(SBufCheckHeader(Buffer));
  
  SBufReallocateIfNeeded(Buffer);
  SBufPrepareFinish(Buffer);
  ReadSBuf(Buffer,readpipe);
  
  BlockSize=Buffer->Header->BlockSize;
  NumChannels=Buffer->Header->NumChannels;
  SampleRate=Buffer->Header->SampleRate;


  epsilon=endfreq/1.e7;
  for (freq=startfreq;freq < (endfreq+epsilon);freq*=freqstepfactor) {
    fprintf(stderr,"%3.3f Hz... ",freq);
    fflush(stderr);
    amplitude=testfreq(freq);
    fprintf(stderr,"%3.3f dB\n",20*log10(amplitude));
    fprintf(outfile,"%16.16f\t%16.16f\n",freq,amplitude);
  }

  fclose(outfile);
  return 0;


  
}

