/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "io_spec.h"
#include "buffer.h"
#include "linklist.h"
#include "util.h"
#include "fftw.h"

#include "bits.h"

int LowBits=0;
int HighBits=0;

int LowDisabled=0;
int HighDisabled=0;

int PanelX=-1;
int PanelY=-1;

int MuteOption=0;
int NoHighBits=0;
int Muted=0;

int AutoScale=0;
int FreqDomain=0;

int GotMagic=0;
int NoPanel=0;
char *Title="bits";

int InputPipe=0;
int OutputPipe=1;

int BlkSize=0;
int NumChannels=0;

FFTW_COMPLEX *FFTBuf;
FFTW_COMPLEX *FFTOutBuf;
int FFTBufLen;

int *FDBuf=NULL;
int FDBufNumChannels=0;
int FDBufLen=0;
int FFTLenFactorNumerator=5;
int FFTLenFactorDenominator=4; /* should always be a power of 2 */
int *FDStore=NULL;
int FDStoreLen=0;

int bpifd=-1;
struct queued_outbuffer *bpibuf;

struct SBuf *Buffer;


void FDQuantBuf(int *Buf,int BufNumChannels,int Len,int Ceil,int Floor,int LowMask)
     /* Quantize the specified buffer and store it in FFTBuf */
{

  int ChanCnt;
  fftw_plan plan;
  double ScaleFactor;
  double Max,Val;
  register int Data;
  int Cnt;
  
  
  /* Frequency domain quantization */
  if (FFTBufLen != Len*2) {
    if (FFTBuf) free(FFTBuf);
    if (FFTOutBuf) free(FFTOutBuf);
    FFTBuf=malloc(sizeof(FFTW_COMPLEX)*Len*2*BufNumChannels);
    FFTOutBuf=malloc(sizeof(FFTW_COMPLEX)*Len*2*BufNumChannels);
    FFTBufLen=Len*2;
    
  }
  for (ChanCnt=0;ChanCnt < BufNumChannels;ChanCnt++) {
    for (Cnt=0;Cnt < (Len);Cnt++) {
      FFTBuf[Cnt*BufNumChannels+ChanCnt].re=(double)Buf[Cnt*BufNumChannels+ChanCnt];
      FFTBuf[Cnt*BufNumChannels+ChanCnt].im=0.0;
      if (Cnt!=0) {
	FFTBuf[(Len*2-Cnt)*BufNumChannels+ChanCnt].re=FFTBuf[Cnt*BufNumChannels+ChanCnt].re;
	FFTBuf[(Len*2-Cnt)*BufNumChannels+ChanCnt].im=0.0;
      }
      
    }
    FFTBuf[Len*BufNumChannels+ChanCnt].re=0.0;
    FFTBuf[Len*BufNumChannels+ChanCnt].im=0.0;
    
    plan=fftw_create_plan(Len*2,FFTW_FORWARD,FFTW_ESTIMATE);
    fftw(plan,1,FFTBuf+ChanCnt,BufNumChannels,0,FFTOutBuf+ChanCnt,BufNumChannels,0);
    fftw_destroy_plan(plan);
    /* Determine Scalefactor */
    Max=0.0;
    for (Cnt=0;Cnt < (Len);Cnt++) {
      /* only need to go up half way -- to blksize, since it is symmetric */
      Val=FFTOutBuf[Cnt*BufNumChannels+ChanCnt].re;
      
      if (Val < 0.0) Val=-Val;
      if (Val > Max) Max=Val;
      
    }
    ScaleFactor=32767/Max;
    
    for (Cnt=0;Cnt < (Len*2);Cnt++) {
      /* while we're at it, make sure imaginary parts are still zero */
      FFTOutBuf[Cnt*BufNumChannels+ChanCnt].im=0;
      
      /* Scale real parts */
      Data=(int)(FFTOutBuf[Cnt*BufNumChannels+ChanCnt].re*ScaleFactor);
      
      /* quantize bits, as desired */
      Data&=~LowMask; 
      /* BUG!!! -- why does it sound bad if we uncomment these two??? */
      //if (Data > Ceil) Data=Ceil&(~LowMask);
      //else if (Data < Floor) Data=Floor&(~LowMask);
      if (Muted) 
	Data=0;
      
      /* unscale and store quantized data */
      
      
      FFTOutBuf[Cnt*BufNumChannels+ChanCnt].re=((double)Data)/ScaleFactor;
      
    }
    /* now inverse fft back into original buffer */
    plan=fftw_create_plan(Len*2,FFTW_BACKWARD,FFTW_ESTIMATE);
    fftw(plan,1,FFTOutBuf+ChanCnt,BufNumChannels,0,FFTBuf+ChanCnt,BufNumChannels,0);
    fftw_destroy_plan(plan);
    /* rescale by size of array
       to compensate for scaling in fft/ifft pair */
    for (Cnt=0;Cnt < (Len);Cnt++) {
      FFTBuf[Cnt*BufNumChannels+ChanCnt].re/=Len*2;
      /*FFTBuf[Cnt*BufNumChannels+ChanCnt].re=(double)Buf[Cnt*BufNumChannels+ChanCnt];*/
    }
    

  }
  
}

void FDQuant(int Ceil,int Floor,int LowMask)
{
  int CCnt,Cnt;
  int StartPos;
  int Len;
  static int notfirsttime;


  if (!FDBuf || FDBufLen != BlkSize*3 || FDBufNumChannels != NumChannels) {
    if (FDBuf) free(FDBuf);
    FDBufLen=BlkSize*3;
    FDBufNumChannels=NumChannels;
    FDBuf=calloc(sizeof(int)*FDBufLen*FDBufNumChannels,1);
    if (FDStore) free(FDStore);
    FDStore=NULL;
    

  }

  /* Shift FDBuf to left */
  for (Cnt=0;Cnt < BlkSize*2 ;Cnt++) {
    for (CCnt=0;CCnt < NumChannels;CCnt++) {
      FDBuf[(Cnt)*FDBufNumChannels+CCnt]=FDBuf[(Cnt+BlkSize)*FDBufNumChannels+CCnt];
    }
  }
  

  /* load in new values */
  for (Cnt=0;Cnt < BlkSize;Cnt++) {
    for (CCnt=0;CCnt < NumChannels;CCnt++) {
      FDBuf[(Cnt+2*BlkSize)*NumChannels+CCnt]=SBufEntry(Buffer,Cnt,CCnt);
    }
  }
  
  if ((BlkSize*FFTLenFactorNumerator)%FFTLenFactorDenominator != 0) {
    fprintf(stderr,"Warning: block size %d is not a multiple of %d\n",(int)BlkSize,(int)FFTLenFactorDenominator);
  }
  
  StartPos=(int)((double)(BlkSize)*(1- (((double)FFTLenFactorNumerator)/((double)FFTLenFactorDenominator)-1.0)/2.0));

  Len=BlkSize*FFTLenFactorNumerator/FFTLenFactorDenominator;

  if (!notfirsttime) {
    fprintf(stderr,"bits \"%s\": Frequency domain quantization delay=%d samples\n",Title,BlkSize*3-StartPos-Len);
    notfirsttime=1;
  }
  FDQuantBuf(FDBuf+FDBufNumChannels*StartPos,FDBufNumChannels,Len,Ceil,Floor,LowMask);
  
  if (FDStoreLen != Len-BlkSize || !FDStore) {
    if (FDStore) free(FDStore);
    FDStoreLen=Len-BlkSize;
    FDStore=calloc(sizeof(int)*FDStoreLen*FDBufNumChannels,1);
  }
  
  /* add stored data to FFTBuf */
  for (Cnt=0;Cnt < FDStoreLen;Cnt++) {
    for (CCnt=0;CCnt < FDBufNumChannels;CCnt++) {
      FFTBuf[CCnt+Cnt*FDBufNumChannels].re=FFTBuf[CCnt+Cnt*FDBufNumChannels].re*(.5-.5*cos(((double)Cnt)/((double)FDStoreLen)*M_PI)) + ((double)FDStore[CCnt+Cnt*FDBufNumChannels])*(.5+.5*cos(((double)Cnt)/((double)FDStoreLen)*M_PI));
    }
  }

  /* extract new stored data from FFTBuf */
  for (Cnt=BlkSize;Cnt < Len;Cnt++) {
    for (CCnt=0;CCnt < FDBufNumChannels;CCnt++) {
      FDStore[CCnt+(Cnt-BlkSize)*FDBufNumChannels]=FFTBuf[CCnt+Cnt*FDBufNumChannels].re;
    }
  }
  
  /* Extract quantized data from FFTBuf into data buffer */

  for (Cnt=0;Cnt < (BlkSize);Cnt++) {
    for (CCnt=0;CCnt < FDBufNumChannels;CCnt++) {
      SBufEntry(Buffer,Cnt,CCnt)=(int)(FFTBuf[Cnt*FDBufNumChannels+CCnt].re);
    }
  }
  
	      
}



int main(int argc, char *argv[])
{
  int Cnt;

  int Ceil;
  int Floor;
  int LowMask;
  double ScaleFactor;

  int UseHighBits,UseLowBits;

  register int Data;
  int MaxVal;
  

  for (Cnt=1; argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    if (!strncmp(argv[Cnt],"-lowbits=",9)) {
      LowBits=strtol(argv[Cnt]+9,NULL,10);
    }
    if (!strncmp(argv[Cnt],"-highbits=",10)) {
      HighBits=strtol(argv[Cnt]+10,NULL,10);
    }
    if (!strcmp(argv[Cnt],"-lowdisabled")) {
      LowDisabled=1;
    }
    if (!strcmp(argv[Cnt],"-highdisabled")) {
      HighDisabled=1;
    }
    if (!strcmp(argv[Cnt],"-muteoption")) {
      MuteOption=1;
    }
    if (!strcmp(argv[Cnt],"-nohighbits")) {
      NoHighBits=1;
    }
    if (!strcmp(argv[Cnt],"-autoscale")) {
      AutoScale=1;
    }
    if (!strcmp(argv[Cnt],"-freqdomain")) {
      FreqDomain=1;
    }
    if (!strcmp(argv[Cnt],"-muted")) {
      Muted=1;
    }
    if (!strcmp(argv[Cnt],"-nopanel")) {
      NoPanel=1;
    }

    if (!strncmp(argv[Cnt],"-title=",7)) {
      Title=strdup(argv[Cnt]+7);
    }
    if (!strncmp(argv[Cnt],"-panelx=",8)) {
      PanelX=strtol(argv[Cnt]+8,NULL,10);
    }
    if (!strncmp(argv[Cnt],"-panely=",8)) {
      PanelY=strtol(argv[Cnt]+8,NULL,10);
    }
  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   input stream\n"
	    "        #1   output stream\n"
	    "        #2   error messages (do not connect)\n"
	    "\n"
	    "Usage:\n"
	    "   %s -magic -highbits=%d -lowbits=%d -lowdisabled -highdisabled -nopanel -title=%s\n"
	    "        -muteoption -nohighbits -autoscale -freqdomain -muted\n"
	    "        -panelx=%d -panely=%d\n"
	    "\n"
	    " -highbits specifies initial number of high order bits to remove.\n"
	    " -lowbits specified initial number of low bits to remove\n"
	    " -lowdisabled initially disables low bit removal\n"
	    " -highdisabled initially disables high bit removal\n"
	    " -nopanel eliminates the control panel altogether\n"
	    " -title=string sets the window title\n"
	    " -muteoption enables a mute button on the panel\n"
	    " -nohighbits disables slider for high bit removal\n"
	    " -autoscale automatically scales max frame value to 32767\n"
	    " -freqdomain performs quantization in freq domain (DCT)\n"
	    " -muted mutes output\n"
	    " -panelx sets horizontal position of panel (in pixels)\n"
	    " -panely sets vertical position of panel (in pixels)\n"
	    
	    ,argv[0],(int)HighBits,(int)LowBits,Title,(int)PanelX,(int)PanelY);
    exit(1);
  }
  
  if (!NoPanel) {
    bpifd=initbpi(argv[0],&bpibuf);
    
  }

  Buffer=CreateSBuf();
  

  for (;;) {
    SBufPrepareHeader(Buffer);

    /* Read data in, eliminate bits as desired, and output it on stdout */
    ReadSBuf(Buffer,InputPipe); 
    assert(SBufCheckHeader(Buffer));
    SBufReallocateIfNeeded(Buffer);
    SBufPrepareFinish(Buffer);
    ReadSBuf(Buffer,InputPipe);
    
    /* copy out needed data */
    BlkSize=Buffer->Header->BlockSize;
    NumChannels=Buffer->Header->NumChannels;
    
    /* send/recv data for the panel */
    if (!NoPanel) {
      bpi_readchars();
      write_queued_outbuffer(bpibuf);
    }
    /* Process data here */
    UseHighBits=HighBits;
    UseLowBits=LowBits;
    if (LowDisabled) UseLowBits=0;
    if (HighDisabled) UseHighBits=0;
    
    if (UseHighBits>=16) { 
      Ceil=0;
      Floor=0;
    } else {
      Ceil=(1<< (15-UseHighBits))-1;
      Floor=-(1<< (15-UseHighBits));
    }
    LowMask=(1<<UseLowBits)-1;
    if (UseHighBits+UseLowBits >= 16) LowMask=~0;
    
    
    if (!FreqDomain) {
      if (AutoScale) {
	MaxVal=0;
	for (Cnt=0;Cnt < (BlkSize*NumChannels);Cnt++) {
	  Data=SBufEntry(Buffer,0,Cnt);
	  if (Data < 0) Data=-Data;
	  if (Data > MaxVal) {
	    MaxVal=Data;
	  }
	}
	ScaleFactor=32767.0/MaxVal;
	for (Cnt=0;Cnt < (BlkSize*NumChannels);Cnt++) {
	 SBufEntry(Buffer,0,Cnt)*=ScaleFactor;
	}
	
      }
      
      for (Cnt=0;Cnt < (BlkSize*NumChannels);Cnt++) {
	Data=SBufEntry(Buffer,0,Cnt);
	Data&=~LowMask; 
	
	if (Data > Ceil) Data=Ceil&(~LowMask);
	else if (Data < Floor) Data=Floor&(~LowMask);
	if (!Muted) 
	  SBufEntry(Buffer,0,Cnt)=Data;
	else /* muted */
	  SBufEntry(Buffer,0,Cnt)=0;
	
      }
      
      if (AutoScale) {
	for (Cnt=0;Cnt < (BlkSize*NumChannels);Cnt++) {
	  SBufEntry(Buffer,0,Cnt)/=ScaleFactor;
	}
      }
    } else {
      FDQuant(Ceil,Floor,LowMask);
    }
    
    SBufPrepare(Buffer);
    while(!WriteSBuf(Buffer,OutputPipe));
  }
}
