/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "io_spec.h"
#include "buffer.h"
#include "linklist.h"
#include "util.h"

int GotMagic=0;
int Delay=100; /* samples */

int InputPipe=0;
int OutputPipe=1;

int BlkSize=0;
int NumChannels=0;


struct SBuf *Buffer;
struct SBuf *OutBuf;

INT16 *DelayBuf=NULL;
int DBLen=0;
int DBNumChannels=0;


int main(int argc, char *argv[])
{
  int Cnt;

  int Ceil;
  int Floor;
  int LowMask;
  int CCnt;
  double ScaleFactor;

  int UseHighBits,UseLowBits;

  register int Data;
  int MaxVal;
  

  for (Cnt=1; argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    if (!strncmp(argv[Cnt],"-delay=",7)) {
      Delay=strtol(argv[Cnt]+7,NULL,10);
    }
  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   input stream\n"
	    "        #1   output stream\n"
	    "        #2   error messages (do not connect)\n"
	    "\n"
	    "Usage:\n"
	    "   %s -magic -delay=%d\n"
	    "\n"
	    "This program delays the audio stream by the specified\n"
	    "number of samples\n"
	    ,argv[0],(int)Delay);
    exit(1);
  }
  

  Buffer=CreateSBuf();
  OutBuf=CreateSBuf();
  for (;;) {
    
    SBufPrepareHeader(Buffer);
    ReadSBuf(Buffer,InputPipe); 
    assert(SBufCheckHeader(Buffer));
    SBufReallocateIfNeeded(Buffer);
    SBufPrepareFinish(Buffer);
    
    ReadSBuf(Buffer,InputPipe); 
    
    
    
    /* copy out needed data */
    BlkSize=Buffer->Header->BlockSize;
    NumChannels=Buffer->Header->NumChannels;
    SBufSetHeader(OutBuf,Buffer->Header->NumChannels,Buffer->Header->SampleRate,Buffer->Header->FrameNumber,Buffer->Header->BlockSize);
    SBufReallocateIfNeeded(OutBuf);
    
    /* Process data here */
    if (!DelayBuf || DBLen != Delay || DBNumChannels != NumChannels) {
      if (DelayBuf) free(DelayBuf);
      DBLen=Delay;
      DBNumChannels=NumChannels;
      DelayBuf=calloc(1,DBNumChannels*DBLen*sizeof(INT16));
    }
    
    
    /* Copy data out of delay */
    if (Delay > BlkSize) {
      for (Cnt=0;Cnt < BlkSize;Cnt++) {
	for (CCnt=0;CCnt < NumChannels;CCnt++) {
	  SBufEntry(OutBuf,Cnt,CCnt)=DelayBuf[(Cnt)*NumChannels+CCnt];
	}
      }

    } else {
      for (Cnt=0;Cnt < Delay;Cnt++) {
	for (CCnt=0;CCnt < NumChannels;CCnt++) {
	  SBufEntry(OutBuf,Cnt,CCnt)=DelayBuf[(Cnt)*NumChannels+CCnt];
	}
      }
    }
    
    /* Shift delay */
    
    for (Cnt=BlkSize; Cnt < Delay ;Cnt++) {
      for (CCnt=0;CCnt < NumChannels;CCnt++) {
	DelayBuf[(Cnt-BlkSize)*NumChannels+CCnt]=DelayBuf[(Cnt)*NumChannels+CCnt];
      }
    }
    
    /* Copy data out of new block */
    for (Cnt=0;Cnt + Delay < BlkSize;Cnt++) {
      for (CCnt=0;CCnt < NumChannels;CCnt++) {

	SBufEntry(OutBuf,Cnt+Delay,CCnt)=SBufEntry(Buffer,Cnt,CCnt);
      }
    }

    for (; Cnt < BlkSize;Cnt++) {
      for (CCnt=0;CCnt < NumChannels;CCnt++) {
	DelayBuf[(Cnt+Delay-BlkSize)*NumChannels+CCnt]=SBufEntry(Buffer,Cnt,CCnt);
	
      }
    }
    
    
    SBufPrepare(OutBuf);
    
    WriteSBuf(OutBuf,OutputPipe);

    
  }
}
