/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; 
     version 2.1 of the License.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#ifndef AGLAOPHONE_BUFFER_H
#define AGLAOPHONE_BUFFER_H /* This is used to tell oldbuffer.h 
			       not to do the macro substitutions */
#endif


struct SBuf {
  struct AglBlk *Header; /* NOTE: Header and data are part of same allocation */
  INT16 *Data;
  int Size; /* number of 16 bit (stereo) samples for which there
	       is space in buffer */
  int NumChannels;
  int Bytes; /* Number of bytes actually used in buffer */
  int Pos; /* Position at which we are reading/writing */
  
};

#define SBufEntry(sbuf,elem,chan) ((sbuf)->Data[(sbuf)->NumChannels*elem + chan ])



struct SBuf *CreateSBuf(void); /* Reserves a big enough data buffer for the header only */
void DeleteSBuf(struct SBuf *Buf); /* Free resources associated with an SBuf */
int WriteSBuf(struct SBuf *Buf,int Pipe); /* returns 1 when the SBuf has been 
					     fully written, 0 otherwise */
int ReadSBuf(struct SBuf *Buf,int Pipe); /* returns 1 when the SBuf header has been 
						  fully read, 0 otherwise */


int SBufIsDone(struct SBuf *Buf);
int SBufIsHeaderDone(struct SBuf *Buf); /* returns non-zero if we were reading
					   the header and that is now 
					   finished. otherwise return zero */
int SBufIsDataDone(struct SBuf *Buf); /* returns non-zero if were reading data
					 (not just the header) and that is
					 now finished. otherwise return zero */

void SBufNewSize(struct SBuf *Buf,int NewSize); /* reallocates Buf->Data, clears buffer. The new buffer has sufficient space for NewSize 16 bit (stereo) samples. Note that the NumChannels must be set correctly before this call. */

int SBufIsReady(struct SBuf *Buf);
void SBufPrepare(struct SBuf *Buf);      /* Prepare to send/receive a block */
     /* This can be used to send a full block. It should generally not
	be used to receive, because it assumes that the block size will 
	not change. The proper sequence for receiving is
	SBufPrepareHeader();
	ReadSBuf();
	assert(SBufCheckHeader())
	SBufReallocateIfNeeded();
	SBufPrepareFinish();
	ReadSBuf() */


void SBufPrepareHeader(struct SBuf *Buf); /* Prepare to send/receive a header only */


/* Prepare to send/receive data, given that the header has
   already been received/sent*/
void SBufPrepareFinish(struct SBuf *Buf);


void SBufSetHeader(struct SBuf *Buf,int NumChannels,int SampleRate,int FrameNumber, int BlockSize);

     /* For SBufGetDataHeader: 
	The returned pointer is into the buffer data. If this 
	gets reallocated or overwritten the returned pointer is
	no longer valid! */
struct AglBlk *SBufGetHeader(struct SBuf *Buf);

/* Copy the data header from Source to Dest. */
     /* Returns 1 if the header data requires a newly 
	reallocated block for the next data block. i.e., if Size*NumChannels has changed */
int SBufCopyHeader(struct SBuf *Source, struct SBuf *Dest);

/* SBufReallocateIfNeeded() looks at the Header in Buf,
   determines if reallocation of the data area is necessary,
   and does so. Be warned that the header will no longer
   be at the same location after this call. Only the header 
   (not the corresponding data) will be copied to the new 
   location. Returns 1 if it
   reallocated, zero otherwise. */
int SBufReallocateIfNeeded(struct SBuf *Buf);



     /* do a sanity check on the header block. Return 
	non-zero if OK, 0 otherwise */
int SBufCheckHeader(struct SBuf *Buf);

int SBufIsEmpty(struct SBuf *Buf);

void SBufEmpty(struct SBuf *Buf);
