/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-2000 Stephen D. Holland
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; 
     version 2.1 of the License.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

*/

/* These typedefs are OK for most current platforms */
typedef unsigned char UINT8;
typedef unsigned short UINT16;
typedef unsigned int UINT32;
typedef signed char INT8;
typedef signed short INT16;
typedef signed int INT32;



struct AglBlk {
  UINT8 MagicByte[6]; /* 6 0x7f bytes signal the start of a block */
  UINT8 zero; /* one zero byte specifies the termination of the 0x7f's */
  UINT8 NumChannels; /* 1 for mono, 2 for stereo. Some modules may support 
			more than 2 channels. */
  UINT32 SampleRate; /* a 4 byte machine-endian sample rate, in samples/sec */
  UINT16 FrameNumber; /* serial number of this block */
  UINT16 BlockSize; /* in (stereo) samples (minimum blocksize is 6).
		       Does not include the header */
  UINT8 MoreMagicByte[3]; /* 2 0x7f bytes signal the start of data */
  UINT8 One; /* should be equal to 0x01 */
  UINT32 Reserved; /* Reserved for future expansion. Set to zero.
		      This pads the block length to 24, which is a 
		      multiple of 8. This should help compatibility
		      with 64 bit architectures */
  
  /* INT16 Data[BlockSize*NumChannels]; */ /* The data is present at the 
     end of the AglBlk. The number of bytes is 
     sizeof(INT16)*BlockSize*NumChannels. The data is host-endian */
  
};

 /* Total length: 6+1+1+4+2+2+3+1+ 2*BlockSize*NumChannels */


