/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; 
     version 2.1 of the License.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

*/


struct OldSBuf {
  INT16 *Data;
  int Size; /* number of 16 bit (stereo) samples for which there
	       is space in buffer */
  int NumChannels;
  int Bytes; /* Number of bytes actually used in buffer */
  int Pos; /* Position at which we are reading/writing */
  
};


#define OB_SBufEntry(sbuf,elem,chan) ((sbuf)->Data[(sbuf)->NumChannels*elem + chan +sizeof(struct DataBlkHeader)/sizeof(INT16)])



struct OldSBuf *OB_CreateSBuf(void); /* Reserves a big enough data buffer for a StatusBlk */
void OB_DeleteSBuf(struct OldSBuf *Buf); /* Free resources associated with an SBuf */
int OB_WriteSBuf(struct OldSBuf *Buf,int Pipe); /* returns 1 when the SBuf has been 
					     fully written, 0 otherwise */
int OB_ReadSBuf(struct OldSBuf *Buf,int Pipe); /* returns 1 when the SBuf has been 
					    fully read, 0 otherwise */


int OB_SBufIsDone(struct OldSBuf *Buf);
void OB_SBufNewSize(struct OldSBuf *Buf,int NewSize); /* reallocates Buf->Data, clears buffer. The new buffer has sufficient space for NewSize 16 bit (stereo) samples. Note that the NumChannels must be set correctly before this call. */
int OB_SBufIsReady(struct OldSBuf *Buf);
int OB_SBufIsStatus(struct OldSBuf *Buf); /* returns 1 if this is a status block */
void OB_SBufPrepare(struct OldSBuf *Buf);      /* Prepare to send/receive a full block */
void OB_SBufPrepareStatus(struct OldSBuf *Buf); /* Prepare to send/receive a status block */
void OB_SBufSetDataHeader(struct OldSBuf *Buf,int NextIsStatusBlk);

     /* For SBufGetDataHeader: 
	The returned pointer is into the buffer data. If this 
	gets reallocated or overwritten the returned pointer is
	no longer valid! */
struct DataBlkHeader *OB_SBufGetDataHeader(struct OldSBuf *Buf);

     /* Copy the data header from Source to Dest. Return 
	NextBlockIsStatusBlk field */
int OB_SBufCopyDataHeader(struct OldSBuf *Source, struct OldSBuf *Dest);


void OB_SBufSetStatusBlk(struct OldSBuf *Buf,int NumChannels,int SampleRate,int FrameNumber,int BlockSize);

struct StatusBlk *OB_SBufGetStatusBlk(struct OldSBuf *Buf);

     /* Returns 1 if the status block data requires a newly 
	reallocated block for the next data block. i.e., if Size*NumChannels has changed */
int OB_SBufCopyStatusBlk(struct OldSBuf *Src,struct OldSBuf *Dest);

/* SBufReallocateIfNeeded() looks at the StatusBlk in Buf,
   determines if reallocation of the data area is necessary,
   and does so. Be warned that the status block will no longer
   be at the same location after this call. Returns 1 if it
   reallocated, zero otherwise. */
int OB_SBufReallocateIfNeeded(struct OldSBuf *Buf);

struct StatusBlk *OB_SBufGetStatus(struct OldSBuf *Buf);



     /* do a sanity check on the status block. Return 
	non-zero if OK, 0 otherwise */
int OB_SBufCheckStatus(struct OldSBuf *Buf);

int OB_SBufIsEmpty(struct OldSBuf *Buf);

#ifndef AGLAOPHONE_BUFFER_H
/* if the new version has not been included, use macros
   to make the old ones take the place of the new ones */

#define SBuf OldSBuf

#define SBufEntry OB_SBufEntry
#define CreateSBuf OB_CreateSBuf 
#define DeleteSBuf OB_DeleteSBuf
#define WriteSBuf OB_WriteSBuf
#define ReadSBuf OB_ReadSBuf
#define SBufIsDone OB_SBufIsDone
#define SBufNewSize OB_SBufNewSize
#define SBufIsReady OB_SBufIsReady
#define SBufIsStatus OB_SBufIsStatus
#define SBufPrepare OB_SBufPrepare
#define SBufPrepareStatus OB_SBufPrepareStatus
#define SBufSetDataHeader OB_SBufSetDataHeader
#define SBufGetDataHeader OB_SBufGetDataHeader
#define SBufCopyDataHeader OB_SBufCopyDataHeader
#define SBufSetStatusBlk OB_SBufSetStatusBlk
#define SBufGetStatusBlk OB_SBufGetStatusBlk
#define SBufCopyStatusBlk OB_SBufCopyStatusBlk
#define SBufReallocateIfNeeded OB_SBufReallocateIfNeeded
#define SBufGetStatus OB_SBufGetStatus
#define SBufCheckStatus OB_SBufCheckStatus
#define SBufIsEmpty OB_SBufIsEmpty



#endif
