/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "io_spec_old.h"
#include "oldbuffer.h"
#include "linklist.h"
#include "util.h"


/* #defines for encoder param style */

#define ENC_E102A 0 /* Encoder-1.02A */
#define ENC_BLADEENC 1 /* bladeenc */
#define ENC_LAME 2 /* Lame ain't an MP3 encoder (not implemented) */
#define ENC_FHG 3 /* Fraunhofer (not implemented) */
int Encoder=ENC_E102A;

/* We don't need USE_NAMED_PIPE if the encoder has been patched to 
   support "-" for output
#define USE_NAMED_PIPE
*/
int InputPipe=0;
int OutputPipe=1;

int qualitylevel=9; /* FhG Quality level */

int BlockSize=0;
int NumChannels=0;
int SampleRate=0;
int OutputFrame=0;
int OutputStatusPeriod=1;  /*100*/ /* Value of 1 gives compatibility
			     with new io_spec */

int PromisedStatus=0;

struct SBuf *InBuf;
int InBufWriting=0;  /* set to 1 if we are currently writing, 0 if reading */

struct SBuf *OutBuf;
int OutBufWriting=0; /* set to 1 if we are currently writing, 0 if reading */
struct StatusBlk *Status;

int bitrate=128;
int layer=3;
int JointStereo=0;

char *encodername="../mp3/encoder-1.02a/encode";
char *decodername="../mp3/mpg123-0.59r/mpg123";
char *rshname=NULL;

int main(int argc, char *argv[])
{
  int Cnt;
  int Ceil;
  int Floor;
  int LowMask;
  register int Data;
  int UseHighBits,UseLowBits;
  int encoderpipe[2];
  int decoderpipe[2];
  char *pipename;
  int encoderpid;
  int decoderpid;
  char *encoderpath;
  char *decoderpath;
  char layerstring[100];
  char sampfreqstring[100];
  char decsampfreqstring[100];
  char bitratestring[100];
  char qualitystring[100];
  int LittleEndian=0;
  int EndianTest=1;
  int ToEncoder;
  int FromDecoder;
  fd_set readfd,writefd;
  int maxfd;
  int NextIsStatus;
  int GotMagic=0;
  int mp3pipe[2];

  char *execpath;
  
  for (Cnt=1; argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    else if (!strcmp(argv[Cnt],"-joint")) {
      JointStereo=1;
    }
    else if (!strncmp(argv[Cnt],"-bitrate=",9)) {
      bitrate=strtol(argv[Cnt]+9,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-quality=",9)) {
      qualitylevel=strtol(argv[Cnt]+9,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-layer=",7)) {
      layer=strtol(argv[Cnt]+7,NULL,10);
    }
    else if (!strncmp(argv[Cnt],"-encodername=",13)) {
      encodername=argv[Cnt]+13;
    }
    else if (!strncmp(argv[Cnt],"-decodername=",13)) {
      decodername=argv[Cnt]+13;
    }
    else if (!strncmp(argv[Cnt],"-encoderhost=",13)) {
      rshname=argv[Cnt]+13;
    } 
    else if (!strcmp(argv[Cnt],"-e102a")) {
      Encoder=ENC_E102A;
    }
    else if (!strcmp(argv[Cnt],"-bladeenc")) {
      Encoder=ENC_BLADEENC;
    }
    else if (!strcmp(argv[Cnt],"-lame")) {
      Encoder=ENC_LAME;
    }
    else if (!strcmp(argv[Cnt],"-fhg")) {
      Encoder=ENC_FHG;
    }
  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   input stream\n"
	    "        #1   output stream\n"
	    "        #2   error messages (do not connect)\n"
	    "\n"
	    "Usage:\n"
	    "   %s -magic -joint -quality=%d -bitrate=%d -layer=%d -encodername=%s -decodername=%s [-encoderhost=%s] -e102a -bladeenc -lame\n"
	    "\n"
	    " -joint   -- enables joint stereo mode (e102a and lame only)\n"
	    " -quality= -- specify quality setting (0-9) for FhG encoder\n"
	    " -encodername=  -- specify file name of encoder, relative to mp3gate directory\n"
	    " -decodername=  -- specify file name of decoder, relative to mp3gate directory\n"
	    " -encoderhost= -- specify remote host for encoding (via rsh)\n"
	    " -e102a -- Use Encoder 1.02A parameter format\n"
	    " -bladeenc -- Use BladeEnc parameter format\n"
	    " -lame -- use LAME (Lame Ain't an MP3 Encoder) parameter format\n"
	    " -fhg -- use Fraunhofer IIS encoder parameter format\n"
	    ,argv[0],(int)qualitylevel,(int)bitrate,(int)layer,encodername,decodername,(rshname==NULL) ? "(null)" : rshname);
    exit(1);
  }
  

  InBuf=CreateSBuf();
  SBufPrepareStatus(InBuf);
  
  /* read in initial status block */
  while (!ReadSBuf(InBuf,InputPipe));

  Status=SBufGetStatusBlk(InBuf);
  BlockSize=Status->BlockSize;
  NumChannels=Status->NumChannels;
  SampleRate=Status->SampleRate;
  OutputFrame=Status->FrameNumber;

  SBufReallocateIfNeeded(InBuf);

  /* find our executable */
  execpath=find_argv0(argv[0]); /* determine path to our executable */
  
  if (!execpath || !strlen(execpath)) {
    execpath=malloc(3);
    strcpy(execpath,"./");
  }
  
  /* strip filename off execpath */
  while (strlen(execpath) && execpath[strlen(execpath)-1]!='/')
    execpath[strlen(execpath)-1]=0;
  if (!strlen(execpath)) strcpy(execpath,"./");
  
  encoderpath=malloc(strlen(encodername)+strlen(execpath)+5);
  decoderpath=malloc(strlen(decodername)+strlen(execpath)+5);
  
  encoderpath[0]=0;
  decoderpath[0]=0;

  if (encodername[0] != '/' && !rshname) {
    strcat(encoderpath,execpath);
    strcat(encoderpath,"/");
  }
  strcat(encoderpath,encodername);

  if (decodername[0] != '/') {
    strcat(decoderpath,execpath);
    strcat(decoderpath,"/");
  }
  strcat(decoderpath,decodername);

#ifdef USE_NAMED_PIPE  
  /* Start up mp3 programs */
  pipename=tmpnam(NULL);
  
  /* create fifo between encoder and player */
  /*** BUG *** we just leave our old pipes lying around in /tmp
       because there's no good way to get rid of them when we exit.
       Maybe we should trap SIGPIPE and remove it there? */
  mkfifo(pipename,0666);
#else
  pipe(mp3pipe);
  pipename="-";
#endif

  /* start encoder */
  pipe(encoderpipe);

  sprintf(layerstring,"%d",(int)layer);
  sprintf(decsampfreqstring,"%d",(int)SampleRate);

  if (Encoder==ENC_BLADEENC) {
    sprintf(sampfreqstring,"-rawfreq=%d",(int)SampleRate);
  } else if (Encoder==ENC_FHG) {
    sprintf(sampfreqstring,"%d",(int)SampleRate);
    
  } else {
    sprintf(sampfreqstring,"%.1f",(float)(((float)SampleRate)/1000.0));
  }
  sprintf(qualitystring,"%d",(int)qualitylevel);
  if (Encoder==ENC_FHG) {
    sprintf(bitratestring,"%d",(int)(bitrate*1000));
  } else {
    sprintf(bitratestring,"%d",(int)bitrate);
  }
  LittleEndian=*((char *)&EndianTest);
  /*  fprintf(stderr,"LittleEndian=%d\n",(int)LittleEndian);*/

  if ((encoderpid=fork()) == 0) {
    /* child process */
    close(encoderpipe[1]);
    dup2(encoderpipe[0],0);

#ifndef USE_NAMED_PIPE  
    close(mp3pipe[0]);
    dup2(mp3pipe[1],1);
#endif

    if (Encoder==ENC_E102A) {
      if (rshname) {
	execlp("rsh",rshname,encoderpath,"-l",layerstring,"-m",(NumChannels > 1) ? ((JointStereo!=0) ? "j":"s"):"m","-s",sampfreqstring,"-b",bitratestring,(LittleEndian==1) ? "-L":"-c","-",pipename,NULL);
      } else {
	execl(encoderpath,encoderpath,"-l",layerstring,"-m",(NumChannels > 1) ? ((JointStereo!=0) ? "j":"s"):"m","-s",sampfreqstring,"-b",bitratestring,(LittleEndian==1) ? "-L":"-c","-",pipename,NULL);
      }
    } 
    else if (Encoder==ENC_LAME) {
      if (rshname) {
	execlp("rsh",rshname,encoderpath,"-m",(NumChannels > 1) ? ((JointStereo!=0) ? "j":"s"):"m","-s",sampfreqstring,"-b",bitratestring,(LittleEndian==1) ? "-x":"--nohist","-",pipename,NULL);
      } else {
	execl(encoderpath,encoderpath,"-m",(NumChannels > 1) ? ((JointStereo!=0) ? "j":"s"):"m","-s",sampfreqstring,"-b",bitratestring,(LittleEndian==1) ? "-x":"--nohist","-",pipename,NULL);
      }
    } 
    else if (Encoder==ENC_BLADEENC) {
      if (rshname) {
	execlp("rsh",rshname,encoderpath,(NumChannels > 1) ? "-rawstereo" : "-rawmono",sampfreqstring,"-br",bitratestring,(LittleEndian==1) ? "-rawbyteorder=LITTLE" : "-rawbyteorder=BIG","STDIN",(strcmp(pipename,"-")==0) ? "STDOUT" : pipename,NULL);
      } else {
	execl(encoderpath,encoderpath,(NumChannels > 1) ? "-rawstereo" : "-rawmono",sampfreqstring,"-br",bitratestring,(LittleEndian==1) ? "-rawbyteorder=LITTLE" : "-rawbyteorder=BIG","STDIN",(strcmp(pipename,"-")==0) ? "STDOUT" : pipename,NULL);

      }
    }
    else if (Encoder==ENC_FHG) {
      if (rshname) {
	execlp("rsh",rshname,encoderpath,"-nc",(NumChannels > 1) ? "2":"1","-sr",sampfreqstring,"-br",bitratestring,(LittleEndian==1) ? "-le":"-be","-sti",(strcmp(pipename,"-")==0) ? "-sto": "-of",(strcmp(pipename,"-")==0) ? "-crc": pipename,"-qual",qualitystring,NULL);
      } else {
	execl(encoderpath,encoderpath,"-nc",(NumChannels > 1) ? "2":"1","-sr",sampfreqstring,"-br",bitratestring,(LittleEndian==1) ? "-le":"-be","-sti",(strcmp(pipename,"-")==0) ? "-sto": "-of",(strcmp(pipename,"-")==0) ? "-crc": pipename,"-qual",qualitystring,NULL);
      }
    } 

    fprintf(stderr,"%s: error running %s\n",argv[0],encoderpath);
    exit(1);
  }
  close(encoderpipe[0]);
  ToEncoder=encoderpipe[1];
  
  if (encoderpid < 0) {
    fprintf(stderr,"%s: error forking process\n",argv[0]);
    exit(1);
  }


  /* start decoder */
  pipe(decoderpipe);
  
  if ((decoderpid=fork()) == 0) {
    /* child process */
    close(encoderpipe[1]);

    close(decoderpipe[0]);
    dup2(decoderpipe[1],1);

#ifndef USE_NAMED_PIPE  
    close(mp3pipe[1]);
    dup2(mp3pipe[0],0);
    
#endif

    execl(decoderpath,decoderpath,"-r",decsampfreqstring,/*(NumChannels > 1) ? "--stereo":"--mono"*/ "--stereo","-s",pipename,NULL);
    fprintf(stderr,"%s: error running %s\n",argv[0],decoderpath);
    exit(1);
  }
  close(decoderpipe[1]);
  FromDecoder=decoderpipe[0];
  if (decoderpid < 0) {
    fprintf(stderr,"%s: error forking process\n",argv[0]);
    exit(1);
  }

  setnonblocking(InputPipe);
  setnonblocking(OutputPipe);
  setnonblocking(ToEncoder);
  setnonblocking(FromDecoder);

  OutBuf=CreateSBuf();
  OutBuf->NumChannels=NumChannels;
  if (NumChannels==1) {
    SBufNewSize(OutBuf,BlockSize*2); /* mpg123 always outputs stereo samples, even in mono mode */
  } else {
    SBufNewSize(OutBuf,BlockSize);
  }
  /* send initial status block */
  OutBufWriting=1;
  SBufSetStatusBlk(OutBuf,NumChannels,SampleRate,OutputFrame,BlockSize);
  SBufPrepareStatus(OutBuf);

  SBufPrepare(InBuf);
  InBufWriting=0;

  /*
  SBufPrepare(OutBuf);
  SBufSetDataHeader(OutBuf,0);
  OutBuf->Pos+=sizeof(struct DataBlkHeader);
  */ 
  
  for (;;) {
    FD_ZERO(&readfd);
    FD_ZERO(&writefd);

    maxfd=maxva(InputPipe,OutputPipe,ToEncoder,FromDecoder,-2);
    
    if (OutBufWriting) {
      FD_SET(OutputPipe,&writefd);
    } else {
      FD_SET(FromDecoder,&readfd);
    }
    
    if (InBufWriting) {
      FD_SET(ToEncoder,&writefd);
    } else {
      FD_SET(InputPipe,&readfd);
    }

    select(maxfd+1,&readfd,&writefd,NULL,NULL);
    if (FD_ISSET(OutputPipe,&writefd)) {
      assert(OutBufWriting);
      if (WriteSBuf(OutBuf,OutputPipe)) {
	/* if finished */

	if (PromisedStatus) { /* Send status block */
	  SBufSetStatusBlk(OutBuf,NumChannels,SampleRate,OutputFrame,BlockSize);
	  SBufPrepareStatus(OutBuf);
	  
	  PromisedStatus=0;
	} else {
	  OutBufWriting=0;
	  /* prepare to read */
	  SBufPrepare(OutBuf);
	  OutBuf->Pos=sizeof(struct DataBlkHeader);
	}
      }
    }
    if (FD_ISSET(FromDecoder,&readfd)) {
      assert(!OutBufWriting);
      /*     fprintf(stderr,"ReadSBuf Pos=%d\n",(int)OutBuf->Pos);*/
      if (ReadSBuf(OutBuf,FromDecoder)) {
	/* finished reading block from the decoder */

	OutputFrame++;
	
	if (!(OutputFrame % OutputStatusPeriod)) {
	  NextIsStatus=1;
	  PromisedStatus=1;
	} else {
	  NextIsStatus=0;
	  PromisedStatus=0;
	}
	SBufSetDataHeader(OutBuf,NextIsStatus);
	if (NumChannels==1) {
	  /* mpg123 always outputs stereo, even in the mono case -- pull 
	     out every other sample, then manually prepare to send
	     half of capacity */
	  for (Cnt=0;Cnt < BlockSize;Cnt++) {
	    OutBuf->Data[sizeof(struct DataBlkHeader)/sizeof(INT16)+Cnt]=OutBuf->Data[sizeof(struct DataBlkHeader)/sizeof(INT16)+Cnt*2];
	  }
	  OutBuf->Bytes=BlockSize*sizeof(INT16)+sizeof(struct DataBlkHeader);
	  OutBuf->Pos=0;

	} else {
	  SBufPrepare(OutBuf);
	} 
	OutBufWriting=1;
      }
    }
    if (FD_ISSET(ToEncoder,&writefd)) {
      assert(InBufWriting);
      if (WriteSBuf(InBuf,ToEncoder)) {
	/* finished sending block to encoder */

	/* Prepare for next block */
	if ((SBufGetDataHeader(InBuf))->NextBlockIsStatusBlk)
	  SBufPrepareStatus(InBuf);
	else SBufPrepare(InBuf);
	
	InBufWriting=0;
      }
    }
    if (FD_ISSET(InputPipe,&readfd)) {
      assert(!InBufWriting);
      if (ReadSBuf(InBuf,InputPipe)) {
	/* finished reading data from stdin */

	if (SBufIsStatus(InBuf)) {
	  Status=SBufGetStatusBlk(InBuf);
	  if (Status->NumChannels != NumChannels || Status->SampleRate != SampleRate || Status->BlockSize != BlockSize) {
	    fprintf(stderr,"%s: Warning: dynamic parameter changing not supported\n",argv[0]);
	  }
	  /* Go ahead and read next block */
	  SBufPrepare(InBuf);
	} else {
	  /* Switch to write mode */
	  SBufPrepare(InBuf);
	  InBuf->Pos=sizeof(struct DataBlkHeader);
	  InBufWriting=1;
	}
      }
    }
  }
}
