/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "io_spec_old.h"
#include "oldbuffer.h"

int GotMagic=0;
int Signed=1; /* 1 = signed numbers, 0=unsigned numbers on input */
int ByteSwap=0;
int BlockSize=1024;
int NumChannels=1;
int SampleRate=44100;

int HostIsBigEndian=0;
/*int StatusSpacing=100;*/
int StatusSpacing=1; /* StatusSpacing=1 gives compatibility with new io_spec */
int FrameNumber=0;

int readpipe=0;
int writepipe=1;

static char **global_argv;

void SwapBytes(char *Buf,int NumBytes)
{
  int Cnt;
  char Tmp;
  
  for (Cnt=0;Cnt < NumBytes;Cnt+=2) {
    Tmp=Buf[Cnt];
    Buf[Cnt]=Buf[Cnt+1];
    Buf[Cnt+1]=Tmp;    
  }
}

void SignShift(INT16 *Buf,int NumWords)
{
  int Cnt;
  
  for (Cnt=0;Cnt < NumWords;Cnt++) {
    Buf[Cnt]+=32768;
  }
}

void InputRead(int pipe,char *Buffer,int nbytes)
{
  int Pos=0;
  int Ret;
  
  while (Pos < nbytes) {
    Ret=read(pipe,Buffer+Pos,nbytes-Pos);
    if (Ret < 0) {
      if (errno != EAGAIN && errno != EINTR) {
	perror(global_argv[0]);
	exit(1);
      }
      
      Ret=0;
    } else if (!Ret) {
      /* EOF */
      exit(0);
      
    }
    Pos+=Ret;
    
  }
  
}

int main(int argc,char *argv[])
{
  int Cnt;
  int EndianCheck=1;
  char *ECPtr;
  struct SBuf *Buffer;
  struct StatusBlk *Status;


  ECPtr=(char *)&EndianCheck;
  if (!EndianCheck) HostIsBigEndian=1;

  global_argv=argv;

  for (Cnt=1;argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    if (!strcmp(argv[Cnt],"-signed")) {
      Signed=1;
    }
    if (!strcmp(argv[Cnt],"-unsigned")) {
      Signed=0;
    }
    if (!strcmp(argv[Cnt],"-little-endian")) {
      if (HostIsBigEndian) ByteSwap=1;
      else ByteSwap=0;
    }
    if (!strcmp(argv[Cnt],"-big-endian")) {
      if (HostIsBigEndian) ByteSwap=0;
      else ByteSwap=1;
    }
    if (!strncmp(argv[Cnt],"-blocksize=",11)) {
      BlockSize=strtol(argv[Cnt]+11,NULL,10);
    }
    if (!strcmp(argv[Cnt],"-stereo")) {
      NumChannels=2;
    }
    if (!strcmp(argv[Cnt],"-mono")) {
      NumChannels=1;
    }
    if (!strncmp(argv[Cnt],"-samplerate=",12)) {
      SampleRate=strtol(argv[Cnt]+12,NULL,10);
    }
  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   raw 16 bit PCM input\n"
	    "        #1   aglaophone format output\n"
	    "        #2   error messages (do not connect)\n"
	    "\n"
	    "IMPORTANT: The output from this module is not rate-limited.\n"
	    "           In order to work correctly it must be passed\n"
	    "           throught the pacer module\n"
	    "\n"
	    "Usage:\n"
	    "   %s -magic -signed -unsigned -little-endian -big-endian -blocksize=%d -stereo -mono -samplerate=%d\n"
	    "\n"
	    "   Defaults are unsigned, native-processor-endian, mono\n"
	    ,argv[0],(int)BlockSize,(int)SampleRate);
    exit(1);
  }
  
  Buffer=CreateSBuf();

  Buffer->NumChannels=NumChannels;
  SBufNewSize(Buffer,BlockSize);
  
  for (;;) {
    /* Read data in from stdin, process it, and push it out stdout in
       spectr-format */
    
    if (!(FrameNumber % StatusSpacing)) {
      /* send status block first */
      SBufSetStatusBlk(Buffer,NumChannels,SampleRate,FrameNumber,BlockSize);
      while (!WriteSBuf(Buffer,writepipe));
    }
    /* Read in data from input */
    InputRead(readpipe,((char *)Buffer->Data)+sizeof(struct DataBlkHeader),BlockSize*sizeof(INT16)*NumChannels);

    if (ByteSwap) {
      SwapBytes(((char *)Buffer->Data)+sizeof(struct DataBlkHeader),BlockSize*sizeof(INT16)*NumChannels);
    }
    
    if (!Signed) {
      SignShift(Buffer->Data + sizeof(struct DataBlkHeader)/sizeof(INT16),BlockSize*NumChannels);
    }

    /* Write data block */
    SBufPrepare(Buffer);
    SBufSetDataHeader(Buffer,!((FrameNumber+1)%StatusSpacing));
    WriteSBuf(Buffer,writepipe);
    
    FrameNumber++;
  }

  
}

