/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "io_spec.h"
#include "buffer.h"


int HostIsBigEndian=0;


int GotMagic=0;
int Signed=1; /* 1 = signed numbers, 0=unsigned numbers on output */
int ByteSwap=0;

int BlockSize=0;
int NumChannels=0;
int SampleRate=0;

int readpipe=0;
int writepipe=1;

static char **global_argv;

void SwapBytes(char *Buf,int NumBytes)
{
  int Cnt;
  char Tmp;
  
  for (Cnt=0;Cnt < NumBytes;Cnt+=2) {
    Tmp=Buf[Cnt];
    Buf[Cnt]=Buf[Cnt+1];
    Buf[Cnt+1]=Tmp;    
  }
}

void SignShift(INT16 *Buf,int NumWords)
{
  int Cnt;
  
  for (Cnt=0;Cnt < NumWords;Cnt++) {
    Buf[Cnt]+=32768;
  }
}

void OutputWrite(int pipe,char *Buffer,int nbytes)
{
  int Pos=0;
  int Ret;
  
  while (Pos < nbytes) {
    Ret=write(pipe,Buffer+Pos,nbytes-Pos);
    if (Ret < 0) {
      if (errno != EAGAIN && errno != EINTR) {
	perror(global_argv[0]);
	exit(1);
      }
      
      Ret=0;
    } else if (!Ret) {
      /* EOF */
      exit(0);
      
    }
    Pos+=Ret;
    
  }
  
}



int main(int argc,char *argv[])
{
  int Cnt;
  int EndianCheck=1;
  char *ECPtr;
  struct SBuf *Buffer;
  struct StatusBlk *Status;


  ECPtr=(char *)&EndianCheck;
  if (!EndianCheck) HostIsBigEndian=1;
  
  global_argv=argv;

  for (Cnt=1;argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    if (!strcmp(argv[Cnt],"-signed")) {
      Signed=1;
    }
    if (!strcmp(argv[Cnt],"-unsigned")) {
      Signed=0;
    }
    if (!strcmp(argv[Cnt],"-little-endian")) {
      if (HostIsBigEndian) ByteSwap=1;
      else ByteSwap=0;
    }
    if (!strcmp(argv[Cnt],"-big-endian")) {
      if (HostIsBigEndian) ByteSwap=0;
      else ByteSwap=1;
    }
  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   aglaophone format input\n"
	    "        #1   raw 16 bit PCM output\n"
	    "        #2   error messages (do not connect)\n"
      	    "\n"
	    "Usage:\n"
	    "   %s -magic -signed -unsigned -little-endian -big-endian\n"
	    "\n"
	    "   Defaults are unsigned, native-processor-endian\n"
	    ,argv[0]);
    exit(1);
  }
  
  Buffer=CreateSBuf();

  for (;;) {
    /* Read data in from stdin, process it, and push it out stdout in
       raw format */
    

    /* Read in data from input */
    SBufPrepareHeader(Buffer);
    while(!ReadSBuf(Buffer,readpipe));
    assert(SBufCheckHeader(Buffer));
    SBufReallocateIfNeeded(Buffer);
    SBufPrepareFinish(Buffer);
    while(!ReadSBuf(Buffer,readpipe));
    
    if (Buffer->Header->BlockSize != BlockSize) {
      BlockSize=Buffer->Header->BlockSize;
      /*fprintf(stderr,"%s: BlockSize=%d\n",global_argv[0],(int)BlockSize);*/
    }
    if (Buffer->Header->NumChannels != NumChannels) {
      NumChannels=Buffer->Header->NumChannels;
      fprintf(stderr,"%s: NumChannels=%d\n",global_argv[0],(int)NumChannels);
    }

    if (Buffer->Header->SampleRate != SampleRate) {
      SampleRate=Buffer->Header->SampleRate;
      fprintf(stderr,"%s: SampleRate=%d\n",global_argv[0],(int)SampleRate);
    }


    if (ByteSwap) {
      SwapBytes(((char *)Buffer->Data),Buffer->Header->BlockSize*sizeof(INT16)*Buffer->Header->NumChannels);
    }
    
    if (!Signed) {
      SignShift(Buffer->Data,Buffer->Header->BlockSize*Buffer->Header->NumChannels);
    }

    /* Write data block */
    OutputWrite(writepipe,((char *)Buffer->Data),Buffer->Header->BlockSize*sizeof(INT16)*Buffer->Header->NumChannels);
    
  }

  
}

