%
%     Aglaophone -- a system for real time audio processing
%     Copyright (C) 1998-1999 Stephen D. Holland
% 
%     This program is free software; you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation; version 2 of the License.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with this program; if not, write to the Free Software
%     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%


% buildfilterdat.m:  build the down/upsampling filter data

% this MATLAB script generates the C source file 'filterdata.c' which contains the
% filter data. 


% We provide down/upsampling filters for downsample factors of 2:6

downsamplefactor=2:12;
upsamplefactor=downsamplefactor;

upsamplefilterwidth=7; % in downsampled samples
downsamplefilterwidth=10;

qa=1; % quality assurance -- do we want to view the frequency response plots?
writeoutput=1; % do we want to write the filters to 'filterdata.c'?

upsamplefilterlengths=zeros(size(upsamplefactor,2),1);
downsamplefilterlengths=zeros(size(downsamplefactor,2),1);

downfilterdata=zeros(downsamplefilterwidth*max(downsamplefactor)+1,size(downsamplefactor,2));
%filterdata=zeros((filterwidth+1)*max(downsamplefactor)-1,size(downsamplefactor,2));
upfilterdata=zeros(2*(upsamplefilterwidth)*max(upsamplefactor)-1,size(upsamplefactor,2));
linearupfilterdata=zeros(2*max(upsamplefactor)-1,size(upsamplefactor,2));

Cnt=1;

for i=downsamplefactor

  downfilter=remez(downsamplefilterwidth*i,[0 (1/i  * .85) (1/i * 1.25) 1]' , [ 1 1 0 0]');
  upfilter=intfilt(i,upsamplefilterwidth,0.8);

  linearupfilter=zeros(1,i*2-1);
  linearupfilter(1:i)=(1:i)/i;
  linearupfilter(i:(2*i-1))=(i:-1:1)/i;


 if qa ~= 0
    figure(Cnt*3-2);
    downfft=abs(fft(downfilter));
    downfb=(0:(size(downfilter,2)-1))/size(downfilter,2);
    plot((downfb),20*log10(downfft));
    grid on
    axis([0 .5 -90 10]);
    xlabel 'Frequency relative to sampling frequency'
    ylabel 'Filter response (dB)'
    title(sprintf('Frequency response of filter for %ix oversampling ADC',i));

    figure(Cnt*3-1);
    upfft=abs(fft(upfilter));
    upfb=(0:(size(upfilter,2)-1))/size(upfilter,2);
    plot((upfb),20*log10(upfft));
    grid on
    axis([0 .5 -90 10]);
    xlabel 'Frequency relative to sampling frequency'
    ylabel 'Filter response (dB)'
    title(sprintf('Frequency response of interpolation filter for %ix oversampling DAC',i));

    figure(Cnt*3);
    linearupfft=abs(fft(linearupfilter));
    linearupfb=(0:(size(linearupfilter,2)-1))/size(linearupfilter,2);
    plot((linearupfb),20*log10(linearupfft));
    grid on
    axis([0 .5 -90 10]);
    xlabel 'Frequency relative to sampling frequency'
    ylabel 'Filter response (dB)'
    title(sprintf('Frequency response of linear interpolation for %ix oversampling DAC',i));

  end
  
  linearupsamplefilterlengths(Cnt,1)=size(linearupfilter,2);
  upsamplefilterlengths(Cnt,1)=size(upfilter,2);
  downsamplefilterlengths(Cnt,1)=size(downfilter,2);

  linearupfilterdata(:,Cnt)=[linearupfilter' ; zeros(size(linearupfilterdata,1)-size(linearupfilter,2),1)];
  upfilterdata(:,Cnt)=[upfilter' ; zeros(size(upfilterdata,1)-size(upfilter,2),1)];
  downfilterdata(:,Cnt)=[downfilter' ; zeros(size(downfilterdata,1)-size(downfilter,2),1)];


  Cnt=Cnt+1;  
end

if writeoutput ~= 0
  datafile=fopen('filterdata.c','w');
  fprintf(datafile,'/* WARNING: do not modify. This file is automatically generated\n');
  fprintf(datafile,'   by the MATLAB script buildfilterdat.m */\n\n');
  fprintf(datafile,'#include "filterdata.h"\n\n');
  fprintf(datafile,'int numfilters=%i;\n',size(downsamplefactor,2));
  fprintf(datafile,'int linearupfilterlengths[]={\n');
  fprintf(datafile,'%i,\n',linearupsamplefilterlengths);
  fprintf(datafile,'};\n\n');
  fprintf(datafile,'int upfilterlengths[]={\n');
  fprintf(datafile,'%i,\n',upsamplefilterlengths);
  fprintf(datafile,'};\n\n');
  fprintf(datafile,'int downfilterlengths[]={\n');
  fprintf(datafile,'%i,\n',downsamplefilterlengths);
  fprintf(datafile,'};\n\n');
  fprintf(datafile,'float linearupfilterdata[]={\n');
  fprintf(datafile,'%f,\n',linearupfilterdata);
  fprintf(datafile,'};\n\n');
  fprintf(datafile,'float upfilterdata[]={\n');
  fprintf(datafile,'%f,\n',upfilterdata);
  fprintf(datafile,'};\n\n');
  fprintf(datafile,'float downfilterdata[]={\n');
  fprintf(datafile,'%f,\n',downfilterdata);
  fprintf(datafile,'};\n\n');

  fclose(datafile);
end

