/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-2000 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef LINUX_ALSA /* only compile this module with ALSA drivers selected */

/*#define LINUX_ALSA_DEBUG */

/* This module communicates with the Linux 'Advanced Linux Sound Architecture'
   (ALSA) audio API.  */

/* notes: disabled MMAP in line 259 of pcm_plugin.c */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/asoundlib.h>
/*#include <linux/asound.h>
  #include <linux/asoundid.h>*/
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <math.h>

#include "io_spec_old.h"
#include "soundspec.h"
#include "linklist.h"
#include "util.h"
#include "soundcardio.h"

extern char **gargv;


#define max_mac(a,b) (((a) > (b)) ? (a) : (b)) /* warning: this expands to 
						  multiple references! */

#define min_mac(a,b) (((a) < (b)) ? (a) : (b)) /* warning: this expands to 
						  multiple references! */


#define do_params snd_pcm_plugin_params
#define do_setup snd_pcm_plugin_setup
#define do_status snd_pcm_plugin_status
#define do_flush snd_pcm_plugin_flush
#define do_write snd_pcm_plugin_write
/*#define do_write snd_pcm_write*/
#define do_read snd_pcm_plugin_read

/*
#define do_params snd_pcm_channel_params
#define do_setup snd_pcm_channel_setup
#define do_flush snd_pcm_channel_flush
#define do_write snd_pcm_write
#define do_read snd_pcm_read
#define do_status snd_pcm_channel_status*/

struct soundspec *sound_init(int CardID,int CardDevice, int Mode,int log2fragsize,int log2numfrags)
     /* fragment size specification not currently supported */
{
  struct soundspec *spec;
  int OpenMode;

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"sound_init()\n");
#endif
  
  spec=calloc(sizeof(struct soundspec),1);
  spec->CardID=CardID;
  spec->CardDevice=CardDevice;
  spec->Mode=Mode;
  spec->log2fragsize=log2fragsize;
  spec->log2numfrags=log2numfrags;
  spec->PlayRate=spec->RecordRate=0;
  spec->NumChannels=0;
  spec->alsahandle=NULL;
  spec->recbuffer=NULL;
  spec->playbuffer=NULL;

  if (spec->Mode==SIM_RECORD) OpenMode=SND_PCM_OPEN_CAPTURE;
  else if (spec->Mode==SIM_PLAY) OpenMode=SND_PCM_OPEN_PLAYBACK;
  else OpenMode=SND_PCM_OPEN_DUPLEX;
  
  
  if (snd_pcm_open(&spec->alsahandle,spec->CardID,spec->CardDevice,OpenMode)) {
    /* Error opening soundcard */
    fprintf(stderr,"Error opening soundcard %d device %d mode %d\n",spec->CardID,spec->CardDevice,OpenMode);
    
    exit(1);
  }
  assert(spec->alsahandle);

  if (spec->Mode==SIM_PLAY || spec->Mode==SIM_DUPLEX) {
    if ((spec->playsoundfh=snd_pcm_file_descriptor(spec->alsahandle,SND_PCM_CHANNEL_PLAYBACK)) < 0) {
      fprintf(stderr,"Error obtaining soundcard file descriptor\n");
      exit(1);
    }
  } else spec->playsoundfh=-1;

  if (spec->Mode==SIM_RECORD || spec->Mode==SIM_DUPLEX) {
    if ((spec->recordsoundfh=snd_pcm_file_descriptor(spec->alsahandle,SND_PCM_CHANNEL_CAPTURE)) < 0) {
      fprintf(stderr,"Error obtaining soundcard file descriptor\n");
      exit(1);
    }
  } else spec->recordsoundfh=-1;
  
  snd_pcm_nonblock_mode(spec->alsahandle,1);
  
  snd_pcm_info(spec->alsahandle,&spec->info); /* fill out info struct */

  return spec;
  
}

void do_format(struct soundspec *spec,int channel)
     /* spec->fmt should be already filled out */
     /* returns actual parameters in spec->fmt, also */
{
  int bps;
  int size;
  int retval;
  
  /*  fprintf(stderr,"do_format()\n");*/

  snd_pcm_channel_setup_t setup;
  snd_pcm_channel_params_t params;

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"do_format()\n");
#endif

  snd_pcm_capture_flush(spec->alsahandle);

  
  bps=spec->fmt.rate*spec->fmt.voices*2;

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"bps=%d\n",bps);
#endif

  size=1 << ((int)(log(bps/40)/log(2))); /* 1/40 second fragment size */

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"size=%d\n",size);
#endif

  if (size < 16) size=16;
  
  if (do_flush(spec->alsahandle,channel)< 0) {
    /*  fprintf(stderr, "%s: could not flush channel\n",gargv[0]);
	exit(1);*/
  }
  memset(&params, 0, sizeof(params));
  params.mode=SND_PCM_MODE_BLOCK;
  /*params.mode=SND_PCM_MODE_STREAM;*/
  params.channel=channel;
  params.format=spec->fmt;

  if (channel == SND_PCM_CHANNEL_PLAYBACK) {
    params.start_mode=SND_PCM_START_FULL;
  } else {
    params.start_mode=SND_PCM_START_DATA;
  }
  params.stop_mode=SND_PCM_STOP_STOP;

  /*  These would be used for SND_PCM_MODE_BLOCK */
  params.buf.block.frag_size=size;
  /* params.buf.block.frags_max=10;*/
  params.buf.block.frags_max=10;
  params.buf.block.frags_min=1;
  
  /* These are use for SND_PCM_MODE_STREAM 
  params.buf.stream.queue_size=size*5;

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"queue_size=%d\n",params.buf.stream.queue_size);
#endif

  params.buf.stream.fill = SND_PCM_FILL_SILENCE;
  params.buf.stream.max_fill = size;*/
  
    
  if ((retval=do_params(spec->alsahandle,&params)) < 0) {

    if (retval==-EINVAL) {
      /* some ALSA drivers only support a discrete set of rate settings:
	 8000 11025 16000 22050 32000 44100 48000 
	 find closest freq */
      if (params.format.rate < 9512) params.format.rate=8000;
      else if (params.format.rate < 13512) params.format.rate=11025;
      else if (params.format.rate < 19025) params.format.rate=16000;
      else if (params.format.rate < 27025) params.format.rate=22050;
      else if (params.format.rate < 38050) params.format.rate=32000;
      else if (params.format.rate < 46050) params.format.rate=44100;
      else params.format.rate=48000;
				       
      retval=do_params(spec->alsahandle,&params);
    }
    if (retval < 0) {
    fprintf(stderr, "%s: could not set channel %d parameters: %s\n",gargv[0],channel,snd_strerror(retval));
      exit(1);
    }
  } 
  if (snd_pcm_plugin_prepare(spec->alsahandle, channel) < 0) {
    fprintf(stderr, "%s: could not prepare channel\n",gargv[0]);
    exit(1);
  }
 

  memset(&setup,0,sizeof(setup));
  setup.mode=SND_PCM_MODE_BLOCK;
  /*setup.mode=SND_PCM_MODE_STREAM;*/
 
  setup.channel=channel;
  
  if (do_setup(spec->alsahandle,&setup) < 0) {
    fprintf(stderr, "%s: could not extract setup\n",gargv[0]);
    exit(1);
  }

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"Requested %d Hz, got %d Hz\n",spec->fmt.rate,setup.format.rate);
#endif

  spec->fmt=setup.format;

  /*
  if (snd_pcm_channel_go(spec->alsahandle,channel) < 0) {
    fprintf(stderr, "%s: could not start channel %d\n",gargv[0],channel);
    exit(1);    
  }
  */


  if (spec->Mode==SIM_DUPLEX && (spec->info.flags & SND_PCM_INFO_DUPLEX_RATE)) {
    /* if we're in duplex mode and this soundcard requires play and 
       record rates to match, then we have to allocate the record buffer
       as well as the play buffer, since this function is only called once. */
    assert(channel != SND_PCM_CHANNEL_CAPTURE);
    
    if (spec->recbuffer) free(spec->recbuffer);
    spec->recbufferlen=setup.buf.block.frag_size;
    spec->recbuffer=calloc(1,spec->recbufferlen);
    spec->recbufferpos=spec->recbufferlen;

  }

  if (channel==SND_PCM_CHANNEL_CAPTURE) {
    if (spec->recbuffer) free(spec->recbuffer);
    spec->recbufferlen=setup.buf.block.frag_size;
    spec->recbuffer=calloc(1,spec->recbufferlen);
    spec->recbufferpos=spec->recbufferlen;
  }
  
  if (channel==SND_PCM_CHANNEL_PLAYBACK) {
    if (spec->playbuffer) free(spec->playbuffer);
    spec->playbufferlen=setup.buf.block.frag_size;
    spec->playbuffer=calloc(1,spec->playbufferlen);
    spec->playbufferpos=0;
    snd_pcm_playback_go(spec->alsahandle);

  }

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"frag_size=%d frags=%d frags_min=%d frags_max=%d\n",setup.buf.block.frag_size,setup.buf.block.frags,setup.buf.block.frags_min,setup.buf.block.frags_max);
#endif

}


int sound_setrecordrate(struct soundspec *spec,int *NumChannels,int DesiredRate)
{
  int format;
  char *testptr;

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"sound_setrecordrate()\n");
#endif

  /* set sample format  (16 bit, unsigned hardware-endian) */

  if (spec->Mode==SIM_DUPLEX && (spec->info.flags & SND_PCM_INFO_DUPLEX_MONO) && *NumChannels > 1) {
    /* Card can't handle stereo in duplex mode, drop down to mono */
    /*fprintf(stderr,"SND_PCM_INFO_DUPLEX_RATE");*/
    *NumChannels=1;
  }

  if (spec->Mode==SIM_DUPLEX && (spec->info.flags & SND_PCM_INFO_DUPLEX_RATE)) {
    /* if we're in duplex mode and this soundcard requires play and 
       record rates to match, play rate gets priority */
    /* fprintf(stderr,"SND_PCM_INFO_DUPLEX_RATE");*/
    spec->RecordRate=spec->PlayRate;
    *NumChannels=spec->NumChannels;
    return spec->RecordRate;
  }


  spec->RecordRate=DesiredRate;
  spec->NumChannels=*NumChannels;

  memset(&spec->fmt, 0, sizeof(spec->fmt));
  

  format=1;
  testptr=(char *) &format;
  if (*testptr)
    spec->fmt.format=SND_PCM_SFMT_S16_LE;
  else spec->fmt.format=SND_PCM_SFMT_S16_BE;

  spec->fmt.interleave=1;
  
  spec->fmt.voices=spec->NumChannels;
  spec->fmt.rate=spec->RecordRate;
  

  do_format(spec,SND_PCM_CHANNEL_CAPTURE);

  spec->NumChannels=spec->fmt.voices;
  spec->RecordRate=spec->fmt.rate;

  *NumChannels=spec->NumChannels;
  
  return spec->RecordRate;
}

int sound_setplayrate(struct soundspec *spec,int *NumChannels,int DesiredRate)
{
  int format;
  char *testptr;
  /* set sample format  (16 bit, unsigned hardware-endian) */

#ifdef LINUX_ALSA_DEBUG
  fprintf(stderr,"sound_setplayrate()\n");
#endif


  if (spec->Mode==SIM_DUPLEX && (spec->info.flags & SND_PCM_INFO_DUPLEX_MONO) && *NumChannels > 1) {
    /* Card can't handle stereo in duplex mode, drop down to mono */
    *NumChannels=1;
  }

  spec->PlayRate=DesiredRate;
  spec->NumChannels=*NumChannels;

  memset(&spec->fmt, 0, sizeof(spec->fmt));
  
  format=1;
  testptr=(char *) &format;
  if (*testptr)
    spec->fmt.format=SND_PCM_SFMT_S16_LE;
  else spec->fmt.format=SND_PCM_SFMT_S16_BE;

  spec->fmt.interleave=1;

  
  spec->fmt.voices=spec->NumChannels;
  spec->fmt.rate=spec->PlayRate;
  
  do_format(spec,SND_PCM_CHANNEL_PLAYBACK);

  spec->NumChannels=spec->fmt.voices;
  spec->PlayRate=spec->fmt.rate;

  *NumChannels=spec->NumChannels;


  if (spec->Mode==SIM_DUPLEX && (spec->info.flags & SND_PCM_INFO_DUPLEX_RATE)) {
    /* if we're in duplex mode and this soundcard requires play and 
       record rates to match, play rate gets priority */
    spec->RecordRate=spec->PlayRate;
    
  }


  return spec->PlayRate;

}


void handle_record_read_error(snd_pcm_t *handle)
{
  snd_pcm_channel_status_t status;
  
  memset(&status,0,sizeof(status));
  status.channel = SND_PCM_CHANNEL_CAPTURE;
  if (do_status(handle, &status)<0) {
    fprintf(stderr, "%s: record status error\n",gargv[0]);
    exit(1);
  }
  if (status.status == SND_PCM_STATUS_OVERRUN) {
    fprintf(stderr,"%s: overrun at position %u!!!\n",gargv[0], status.scount);

    if (snd_pcm_plugin_prepare(handle, SND_PCM_CHANNEL_CAPTURE)<0) {
      fprintf(stderr, "%s overrun: prepare error\n",gargv[0]);
      exit(1);
    }
    return; /* ok, ignore error */
  }
  fprintf(stderr, "%s: read error\n",gargv[0]);
  exit(1);
}

void handle_play_write_error(snd_pcm_t *handle)
{
  snd_pcm_channel_status_t status;
  
  memset(&status,0,sizeof(status));
  status.channel = SND_PCM_CHANNEL_PLAYBACK;
  if (do_status(handle, &status)<0) {
    fprintf(stderr, "%s: play status error\n",gargv[0]);
    exit(1);
  }
  if (status.status == SND_PCM_STATUS_UNDERRUN) {
    fprintf(stderr,"%s: underrun at position %u!!!\n",gargv[0], status.scount);
    
    if (snd_pcm_plugin_prepare(handle, SND_PCM_CHANNEL_PLAYBACK)<0) {
      fprintf(stderr, "%s overrun: prepare error\n",gargv[0]);
      exit(1);
    }
    return; /* ok, ignore error */
  }
  fprintf(stderr, "%s: write error\n",gargv[0]);
  exit(1);
}


int sound_readbuf(struct soundspec *spec,INT16 *CurBlock,int Size,int CurPos,int *moreavail,int DebugFlag)
{
  int GotBytes;
  int xferamnt;


  *moreavail=0;
  
  while (CurPos < Size) {

    xferamnt=min_mac(Size-CurPos,spec->recbufferlen-spec->recbufferpos);
    
    if (xferamnt > 0) {
      memcpy(((char *)CurBlock)+CurPos,spec->recbuffer+spec->recbufferpos,xferamnt);
      CurPos+=xferamnt;
      spec->recbufferpos+=xferamnt;
    }
    
    if (spec->recbufferpos==spec->recbufferlen) {
      GotBytes=do_read(spec->alsahandle,spec->recbuffer,spec->recbufferlen);
      if (GotBytes==-EAGAIN) GotBytes=0;
      if (GotBytes < 0) {
	/* Error! */
	fprintf(stderr,"do_read() error %d: %s. recbufferlen=%d\n",-GotBytes,snd_strerror(GotBytes),spec->recbufferlen);
	handle_record_read_error(spec->alsahandle);
	GotBytes=0;
      } else if (GotBytes > 0) {
	assert(GotBytes==spec->recbufferlen); /* should get exactly one frame */
	spec->recbufferpos=0; /* mark record buffer as not yet copied out */
      } else if (GotBytes==0) {
	return CurPos; /* no data available, return to caller */
      }
      
    }
  }
  return CurPos;

  /*while (CurPos < Size) {
    GotBytes=do_read(spec->alsahandle,(void *)(((char *)CurBlock)+CurPos),Size-CurPos);
    if (GotBytes < 0) {
      
      handle_record_read_error(spec->alsahandle,SND_PCM_CHANNEL_CAPTURE);
      GotBytes=0;
    } else if (GotBytes==0) {
      return CurPos;
    }
    CurPos+=GotBytes;
  
    
  return CurPos;
  */
}

int sound_writebuf(struct soundspec *spec,INT16 *CurBlock,int Size,int CurPos,int *morespace,int DebugFlag)
{
  int GotBytes;
  int xferamnt;
  double debugtime;
  snd_pcm_channel_status_t status;

  *morespace=0 ;

  /*  write(2,((char *)CurBlock)+CurPos,Size-CurPos);*/

  /*  fprintf(stderr,"spec->playbufferpos=%d\n",spec->playbufferpos);
      fprintf(stderr,"CurPos=%d Size=%d\n",CurPos,Size);*/


  while (CurPos < Size) {
    xferamnt=min_mac(Size-CurPos,spec->playbufferlen-spec->playbufferpos);
    if (xferamnt > 0) {
      memcpy(spec->playbuffer+spec->playbufferpos,((char *)CurBlock)+CurPos,xferamnt);

      CurPos+=xferamnt;
      spec->playbufferpos+=xferamnt;
    }
    
    if (spec->playbufferpos == spec->playbufferlen) {
      /*fprintf(stderr,"writing %d bytes\n",spec->playbufferlen);*/

      memset(&status,0,sizeof(status));
      status.channel = SND_PCM_CHANNEL_PLAYBACK;
      if (do_status(spec->alsahandle, &status)<0) {
	fprintf(stderr,"%s: status error\n",gargv[0]);
	exit(1);
      }
      if (status.free < spec->playbufferlen) {
	/* no frags available -- return! */
	return CurPos;
      }
      /*      fprintf(stderr,"free bytes: %d, count=%d\n",status.free,status.count);*/
      startdebugtimer();
      GotBytes=do_write(spec->alsahandle,spec->playbuffer,spec->playbufferlen);
      if (GotBytes==-EAGAIN) GotBytes=0;
      debugtime=resultdebugtimer();
      if (debugtime > 5e-3) fprintf(stderr,"Slow write: %f ms. AskBytes=%ld \n",(float)(debugtime*1000),(long)spec->playbufferlen);
      
      if (GotBytes < 0) {
	/* error! */
	fprintf(stderr,"do_write() error: %s\n",snd_strerror(GotBytes));
	handle_play_write_error(spec->alsahandle);
	GotBytes=0;
      } else if (GotBytes==0) {
	/* no space left -- exit! */
	return CurPos;
      } else {
	/* GotBytes > 0 */
	assert(GotBytes==spec->playbufferlen); /* in block mode, sizes should match */
	spec->playbufferpos=0;
	return CurPos;
      }
      
    }
  }
  return CurPos;

  /*
  while (CurPos < Size) {
    
    GotBytes=do_write(spec->alsahandle,(void *)(((char *)CurBlock)+CurPos),Size-CurPos);
    if (GotBytes < 0) {
      fprintf(stderr,"Error writing to soundcard: %s\n",snd_strerror(GotBytes));
      exit(1);
    }
    if (GotBytes==0) {
      return CurPos;
    }
    CurPos+=GotBytes;
  
  }
  return CurPos;
  */
}

#endif /* LINUX_ALSA */

