/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>

#include "soundcardio.h"
#include "linklist.h"
#include "util.h"
#include "panelinterface_h.h"

int writetogui=-1;
struct queued_outbuffer *writegui=NULL;

int readfromgui=-1;

int GUIReady=0;




/*

 AnaGram Parsing Engine
 Copyright (c) 1993, 1996, Parsifal Software.
 All Rights Reserved.

*/



#ifndef PANELINTERFACE_H
#include "panelinterface.h"
#endif

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define RULE_CONTEXT (&((PCB).cs[(PCB).ssx]))
#define ERROR_CONTEXT ((PCB).cs[(PCB).error_frame_ssx])
#define CONTEXT ((PCB).cs[(PCB).ssx])



panelinterface_pcb_type panelinterface_pcb;
#define PCB panelinterface_pcb

/*  Line 107, /home3/sdh4/aglaophone/soundcardio/panelinterface.syn */
int initpanelinterface(char *argv0,struct queued_outbuffer **buf)
  /* initialize the interface to the GUI panel. 
     returns the socket on which data can be read from the
     panel (for select()) */
  /* NOTE: We do asynchronous I/O on writes to the GUI panel. This means
     that we will not block waiting for it to process. For more information 
     see information about the queued_outbuffer in util.c */
  
{

  
  init_panelinterface(); /* call the AnaGram generated initialization routine */

  
  spawnjava(argv0,"sio_panel",&writetogui,&readfromgui);

  /* set non-blocking i/o */
  
  
  writegui=create_queued_outbuffer(writetogui);
  *buf=writegui;
  
  
  
  /* send initialization data */
  
  qob_printf(writegui,1,"RecordSampleRate=%d\n",(int)RecordSampleRate);
  qob_printf(writegui,1,"RecordNumChannels=%d\n",(int)RecordNumChannels);
  qob_printf(writegui,1,"RecordBufferSize=%d\n",(int)RecordBufferSize);
  qob_printf(writegui,1,"ADCMode=%d\n",(int)ADCMode);
  qob_printf(writegui,1,"DACMode=%d\n",(int)DACMode);
  qob_printf(writegui,1,"PlayBufferSize=%d\n",(int)(PlayBufferSize*NumPlayBuffers));
  qob_printf(writegui,1,"PlayBufPos=%d\n",(int)((NextPlayBuffer-NextAvailBuffer+NumPlayBuffers) % NumPlayBuffers));
  qob_printf(writegui,1,"PlaySampleRate=%d\n",(int)PlaySampleRate);
  qob_printf(writegui,1,"WindowTitle=\"%s\"\n",Title);
  qob_printf(writegui,1,"PanelMode=%d\n",(int)Mode);
  qob_printf(writegui,1,"BeginGUI\n");

  /* Wait for GUI to come up before continuing */
  while (queued_outbuffer_data_available(writegui)) write_queued_outbuffer(writegui);

  setnonblocking(readfromgui);
  while(!GUIReady) {
    fd_set rfg_fds;
    
    FD_ZERO(&rfg_fds);
    FD_SET(readfromgui,&rfg_fds);
    
    pi_readchars();
    if (!GUIReady) select(readfromgui+1,&rfg_fds,NULL,NULL,NULL);
  }
  /*fprintf(stderr,"Got GUIReady\n");*/

  setnonblocking(writetogui);

  return readfromgui;
}


int pi_parsechar(int Char)
{
  (PCB).input_code=Char;
  /*fprintf(stderr,"ParseChar: %c\n",(char)Char);*/
  panelinterface();
  
  return ((PCB).exit_flag==AG_RUNNING_CODE);
}

void pi_readchars(void)
{
  parserreadchars(readfromgui,&pi_parsechar,&init_panelinterface);
}



#ifndef CONVERT_CASE
#define CONVERT_CASE(c) (c)
#endif
#ifndef TAB_SPACING
#define TAB_SPACING 8
#endif

#define ag_rp_1() (GUIReady=1)

static void ag_rp_2(int i) {
/* Line 57, /home3/sdh4/aglaophone/soundcardio/panelinterface.syn */
    if (i != RecordSampleRate) {
      if (Mode==SIM_RECORD || Mode==SIM_DUPLEX) {
	QueueNewRecordRate(i);
      }
    }
  
}

static void ag_rp_3(int i) {
/* Line 64, /home3/sdh4/aglaophone/soundcardio/panelinterface.syn */
    if (i != RecordNumChannels) {
      if (Mode==SIM_RECORD || Mode==SIM_DUPLEX) {
	QueueNewRecordNumChannels(i);
      }
    }
  
}

static void ag_rp_4(int i) {
/* Line 71, /home3/sdh4/aglaophone/soundcardio/panelinterface.syn */
    if (i != ADCMode) {
      SetADCMode(i);
    }
  
}

static void ag_rp_5(int i) {
/* Line 76, /home3/sdh4/aglaophone/soundcardio/panelinterface.syn */
    if (i!=RecordBufferSize) {
      if (Mode==SIM_RECORD || Mode==SIM_DUPLEX) {
	QueueNewRecordBufSize(i);
      }
    }
  
}

static void ag_rp_6(int i) {
/* Line 83, /home3/sdh4/aglaophone/soundcardio/panelinterface.syn */
    if (i != DACMode) {
      SetDACMode(i);
    }
  
}

static void ag_rp_7(int i) {
/* Line 88, /home3/sdh4/aglaophone/soundcardio/panelinterface.syn */
    if (i < (PlayBufferSize*3)) i=PlayBufferSize*3; /* Lower limit on play buffer size */

    if (i!=(PlayBufferSize*NumPlayBuffers)) {
      QueueNewPlayNumBuffers(i/PlayBufferSize);
    }
    
  
}

#define ag_rp_8(d) ((unsigned)(d-'0'))

#define ag_rp_9(i, d) ((i*10)+(unsigned)(d-'0'))


#define READ_COUNTS 
#define WRITE_COUNTS 
static int ag_null_value;
#define V(i,t) (*(t *) (&(PCB).vs[(PCB).ssx + i]))
#define VS(i) (PCB).vs[(PCB).ssx + i]

#ifndef GET_CONTEXT
#define GET_CONTEXT CONTEXT = (PCB).input_context
#endif

typedef enum {
  ag_action_1,
  ag_action_2,
  ag_action_3,
  ag_action_4,
  ag_action_5,
  ag_action_6,
  ag_action_7,
  ag_action_8,
  ag_action_9,
  ag_action_10,
  ag_action_11,
  ag_action_12
} ag_parser_action;

static int ag_ap;



static const unsigned char ag_rpx[] = {
    0,  0,  0,  1,  2,  3,  4,  5,  6,  7,  0,  0,  8,  9
};

static const unsigned char ag_key_itt[] = {
 0
};

static const unsigned short ag_key_pt[] = {
0
};

static const unsigned char ag_key_ch[] = {
    0, 80,102,255,102,255,117,255, 66, 83,255,121,255, 97,255,108,255, 66,
   78, 83,255,100,255,114,255,111,255, 99,255,101,255, 65, 68, 71, 80, 82,
  255
};

static const unsigned char ag_key_act[] = {
  0,3,3,4,2,4,2,4,2,3,4,2,4,2,4,2,4,3,3,3,4,2,4,2,4,2,4,2,4,2,4,3,3,3,2,
  2,4
};

static const unsigned char ag_key_parm[] = {
    0, 12, 11,  0,  0,  0,  0,  0,  0, 13,  0,  0,  0,  0,  0,  0,  0,  9,
    7,  5,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  8, 10,  4,  0,  0,
    0
};

static const unsigned char ag_key_jmp[] = {
    0, 24, 28,  0,  1,  0,  4,  0,  6, 36,  0,  8,  0, 11,  0, 13,  0, 47,
   58, 70,  0, 17,  0, 21,  0, 23,  0, 25,  0, 27,  0,  0,  8, 16, 15, 29,
    0
};

static const unsigned char ag_key_index[] = {
   31,  0,  0,  0,  0,  0,  0,  0,  0,  0, 31,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0
};

static const unsigned char ag_key_ends[] = {
68,67,77,111,100,101,61,0, 65,67,77,111,100,101,61,0, 
85,73,82,101,97,100,121,0, 111,115,61,0, 101,114,83,105,122,101,61,0, 
97,109,112,108,101,82,97,116,101,61,0, 
117,102,102,101,114,83,105,122,101,61,0, 
117,109,67,104,97,110,110,101,108,115,61,0, 
97,109,112,108,101,82,97,116,101,61,0, 
};
#define AG_TCV(x) (((int)(x) >= 0 && (int)(x) <= 255) ? ag_tcv[(x)] : 0)

static const unsigned char ag_tcv[] = {
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  3,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 14, 14, 14, 14, 14, 14,
   14, 14, 14, 14,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0
};

#ifndef SYNTAX_ERROR
#define SYNTAX_ERROR fprintf(stderr,"%s, line %d, column %d\n", \
  (PCB).error_message, (PCB).line, (PCB).column)
#endif

#ifndef FIRST_LINE
#define FIRST_LINE 1
#endif

#ifndef FIRST_COLUMN
#define FIRST_COLUMN 1
#endif


#ifndef PARSER_STACK_OVERFLOW
#define PARSER_STACK_OVERFLOW {fprintf(stderr, \
   "\nParser stack overflow, line %d, column %d\n",\
   (PCB).line, (PCB).column);}
#endif

#ifndef REDUCTION_TOKEN_ERROR
#define REDUCTION_TOKEN_ERROR {fprintf(stderr, \
    "\nReduction token error, line %d, column %d\n", \
    (PCB).line, (PCB).column);}
#endif


typedef enum
  {ag_accept_key, ag_set_key, ag_jmp_key, ag_end_key, ag_no_match_key,
   ag_cf_accept_key, ag_cf_set_key, ag_cf_end_key} key_words;


static void ag_track(void) {
  int ag_k = 0;
  while (ag_k < (PCB).rx) {
    int ag_ch = (PCB).lab[ag_k++];
    switch (ag_ch) {
    case '\n':
      (PCB).column = 1, (PCB).line++;
    case '\r':
    case '\f':
      break;
    case '\t':
      (PCB).column += (TAB_SPACING) - ((PCB).column - 1) % (TAB_SPACING);
      break;
    default:
      (PCB).column++;
    }
  }
  ag_k = 0;
  while ((PCB).rx < (PCB).fx) (PCB).lab[ag_k++] = (PCB).lab[(PCB).rx++];
  (PCB).fx = ag_k;
  (PCB).rx = 0;
}


static void ag_prot(void) {
  int ag_k = 32 - ++(PCB).btsx;
  if (ag_k <= (PCB).ssx) {
    (PCB).exit_flag = AG_STACK_ERROR_CODE;
    PARSER_STACK_OVERFLOW;
    return;
  }
  (PCB).bts[(PCB).btsx] = (PCB).sn;
  (PCB).bts[ag_k] = (PCB).ssx;
  (PCB).vs[ag_k] = (PCB).vs[(PCB).ssx];
  (PCB).ss[ag_k] = (PCB).ss[(PCB).ssx];
}

static void ag_undo(void) {
  if ((PCB).drt == -1) return;
  while ((PCB).btsx) {
    int ag_k = 32 - (PCB).btsx;
    (PCB).sn = (PCB).bts[(PCB).btsx--];
    (PCB).ssx = (PCB).bts[ag_k];
    (PCB).vs[(PCB).ssx] = (PCB).vs[ag_k];
    (PCB).ss[(PCB).ssx] = (PCB).ss[ag_k];
  }
  (PCB).token_number = (panelinterface_token_type) (PCB).drt;
  (PCB).ssx = (PCB).dssx;
  (PCB).sn = (PCB).dsn;
  (PCB).drt = -1;
}


static const unsigned char ag_tstt[] = {
13,12,11,10,9,8,7,5,4,0,1,2,
14,0,6,
14,0,6,
14,0,6,
14,0,6,
14,0,6,
14,0,6,
14,0,6,
14,0,6,
3,0,
13,12,11,10,9,8,7,5,4,0,2,
14,0,
14,0,
14,0,
14,0,
14,0,
14,0,
14,0,
14,0,
3,0,
  0
};


static unsigned const char ag_astt[68] = {
  1,1,1,1,1,1,1,1,2,7,1,1,2,7,1,2,7,1,2,7,1,2,7,1,2,7,1,2,7,1,2,7,1,2,7,1,3,
  7,1,1,1,1,1,1,1,1,2,6,1,10,5,10,5,10,4,10,4,10,4,10,4,10,4,10,4,3,7,11
};


static const unsigned char ag_pstt[] = {
1,2,3,4,5,6,7,8,3,0,10,9,
12,1,11,
12,2,12,
12,3,13,
12,4,14,
12,5,15,
12,6,16,
12,7,17,
12,8,18,
1,9,
1,2,3,4,5,6,7,8,3,0,19,
13,11,
13,10,
13,9,
13,8,
13,7,
13,6,
13,5,
13,4,
2,19,
  0
};


static const unsigned char ag_sbt[] = {
     0,  12,  15,  18,  21,  24,  27,  30,  33,  36,  38,  49,  51,  53,
    55,  57,  59,  61,  63,  65,  67
};


static const unsigned char ag_sbe[] = {
     9,  13,  16,  19,  22,  25,  28,  31,  34,  37,  47,  50,  52,  54,
    56,  58,  60,  62,  64,  66,  67
};


static const unsigned char ag_fl[] = {
  1,2,3,1,2,2,2,2,2,2,2,2,1,2
};

static const unsigned char ag_ptt[] = {
    0,  1,  1,  2,  2,  2,  2,  2,  2,  2,  2,  2,  6,  6
};




static void ag_ra(void)
{
  switch(ag_rpx[ag_ap]) {
  case   1: ag_rp_1(); break;
  case   2: ag_rp_2(VS(1)); break;
  case   3: ag_rp_3(VS(1)); break;
  case   4: ag_rp_4(VS(1)); break;
  case   5: ag_rp_5(VS(1)); break;
  case   6: ag_rp_6(VS(1)); break;
  case   7: ag_rp_7(VS(1)); break;
  case   8: VS(0) = ag_rp_8(VS(0)); break;
  case   9: VS(0) = ag_rp_9(VS(0), VS(1)); break;
  }
}

#define TOKEN_NAMES panelinterface_token_names
const char *const panelinterface_token_names[15] = {
  "panelinterface",
  "panelinterface",
  "line",
  "",
  "",
  "",
  "unsigned integer",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "DIGIT",

};

static char ag_msg[82];

static void ag_diagnose(void) {
  int ag_snd = (PCB).sn, ag_k;

  ag_k = ag_sbt[ag_snd];
  if (*TOKEN_NAMES[ag_tstt[ag_k]] && ag_astt[ag_k + 1] == ag_action_8) {
    sprintf(ag_msg, "Missing %s", TOKEN_NAMES[ag_tstt[ag_k]]);
  }
  else if ((PCB).token_number && *TOKEN_NAMES[(PCB).token_number]) {
    sprintf(ag_msg, "Unexpected %s", TOKEN_NAMES[(PCB).token_number]);
  }
  else if (isprint((*(PCB).lab)) && (*(PCB).lab) != '\\') {
    sprintf(ag_msg, "Unexpected \'%c\'", (char) (*(PCB).lab));
  }
  else sprintf(ag_msg, "Unexpected input");
  (PCB).error_message = ag_msg;


}
static int ag_action_1_r_proc(void);
static int ag_action_2_r_proc(void);
static int ag_action_3_r_proc(void);
static int ag_action_4_r_proc(void);
static int ag_action_1_s_proc(void);
static int ag_action_3_s_proc(void);
static int ag_action_1_proc(void);
static int ag_action_2_proc(void);
static int ag_action_3_proc(void);
static int ag_action_4_proc(void);
static int ag_action_5_proc(void);
static int ag_action_6_proc(void);
static int ag_action_7_proc(void);
static int ag_action_8_proc(void);
static int ag_action_9_proc(void);
static int ag_action_10_proc(void);
static int ag_action_11_proc(void);
static int ag_action_8_proc(void);


static int (*const  ag_r_procs_scan[])(void) = {
  ag_action_1_r_proc,
  ag_action_2_r_proc,
  ag_action_3_r_proc,
  ag_action_4_r_proc
};

static int (*const  ag_s_procs_scan[])(void) = {
  ag_action_1_s_proc,
  ag_action_2_r_proc,
  ag_action_3_s_proc,
  ag_action_4_r_proc
};

static int (*const  ag_gt_procs_scan[])(void) = {
  ag_action_1_proc,
  ag_action_2_proc,
  ag_action_3_proc,
  ag_action_4_proc,
  ag_action_5_proc,
  ag_action_6_proc,
  ag_action_7_proc,
  ag_action_8_proc,
  ag_action_9_proc,
  ag_action_10_proc,
  ag_action_11_proc,
  ag_action_8_proc
};


static int ag_action_10_proc(void) {
  (PCB).btsx = 0, (PCB).drt = -1;
  ag_track();
  return 0;
}

static int ag_action_11_proc(void) {
  (PCB).btsx = 0, (PCB).drt = -1;
  (PCB).vs[(PCB).ssx] = *(PCB).lab;
  (PCB).ssx--;
  ag_ra();
  (PCB).ssx++;
  ag_track();
  return 0;
}

static int ag_action_3_r_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  (PCB).btsx = 0, (PCB).drt = -1;
  (PCB).reduction_token = (panelinterface_token_type) ag_ptt[ag_ap];
  ag_ra();
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_3_s_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  (PCB).btsx = 0, (PCB).drt = -1;
  (PCB).reduction_token = (panelinterface_token_type) ag_ptt[ag_ap];
  ag_ra();
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_4_r_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  (PCB).reduction_token = (panelinterface_token_type) ag_ptt[ag_ap];
  return 1;
}

static int ag_action_2_proc(void) {
  (PCB).btsx = 0, (PCB).drt = -1;
  if ((PCB).ssx >= 32) {
    (PCB).exit_flag = AG_STACK_ERROR_CODE;
    PARSER_STACK_OVERFLOW;
  }
  (PCB).vs[(PCB).ssx] = *(PCB).lab;
  (PCB).ss[(PCB).ssx] = (PCB).sn;
  (PCB).ssx++;
  (PCB).sn = ag_ap;
  ag_track();
  return 0;
}

static int ag_action_9_proc(void) {
  if((PCB).drt == -1) {
    (PCB).drt=(PCB).token_number;
    (PCB).dssx=(PCB).ssx;
    (PCB).dsn=(PCB).sn;
  }
  ag_prot();
  (PCB).ss[(PCB).ssx] = (PCB).sn;
  (PCB).ssx++;
  (PCB).sn = ag_ap;
  (PCB).rx = 0;
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_2_r_proc(void) {
  (PCB).ssx++;
  (PCB).sn = ag_ap;
  return 0;
}

static int ag_action_7_proc(void) {
  --(PCB).ssx;
  (PCB).exit_flag = AG_SUCCESS_CODE;
  (PCB).rx = 0;
  return 0;
}

static int ag_action_1_proc(void) {
  (PCB).exit_flag = AG_SUCCESS_CODE;
  ag_track();
  return 0;
}

static int ag_action_1_r_proc(void) {
  (PCB).exit_flag = AG_SUCCESS_CODE;
  return 0;
}

static int ag_action_1_s_proc(void) {
  (PCB).exit_flag = AG_SUCCESS_CODE;
  return 0;
}

static int ag_action_4_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  (PCB).reduction_token = (panelinterface_token_type) ag_ptt[ag_ap];
  (PCB).btsx = 0, (PCB).drt = -1;
  (PCB).vs[(PCB).ssx] = *(PCB).lab;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  else (PCB).ss[(PCB).ssx] = (PCB).sn;
  ag_track();
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbe[(PCB).sn] + 1;
    unsigned ag_t2 = ag_sbt[(PCB).sn+1] - 1;
    do {
      unsigned ag_tx = (ag_t1 + ag_t2)/2;
      if (ag_tstt[ag_tx] < (const unsigned char)(PCB).reduction_token) ag_t1 = ag_tx + 1;
      else ag_t2 = ag_tx;
    } while (ag_t1 < ag_t2);
    ag_ap = ag_pstt[ag_t1];
    if ((ag_s_procs_scan[ag_astt[ag_t1]])() == 0) break;
  }
  return 0;
}

static int ag_action_3_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  (PCB).btsx = 0, (PCB).drt = -1;
  (PCB).vs[(PCB).ssx] = *(PCB).lab;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  else (PCB).ss[(PCB).ssx] = (PCB).sn;
  ag_track();
  (PCB).reduction_token = (panelinterface_token_type) ag_ptt[ag_ap];
  ag_ra();
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbe[(PCB).sn] + 1;
    unsigned ag_t2 = ag_sbt[(PCB).sn+1] - 1;
    do {
      unsigned ag_tx = (ag_t1 + ag_t2)/2;
      if (ag_tstt[ag_tx] < (const unsigned char)(PCB).reduction_token) ag_t1 = ag_tx + 1;
      else ag_t2 = ag_tx;
    } while (ag_t1 < ag_t2);
    ag_ap = ag_pstt[ag_t1];
    if ((ag_s_procs_scan[ag_astt[ag_t1]])() == 0) break;
  }
  return 0;
}

static int ag_action_8_proc(void) {
  ag_undo();
  (PCB).rx = 0;
  (PCB).exit_flag = AG_SYNTAX_ERROR_CODE;
  ag_diagnose();
  SYNTAX_ERROR;
  {(PCB).rx = 1; ag_track();}
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_5_proc(void) {
  int ag_sd = ag_fl[ag_ap];
  if((PCB).drt == -1) {
    (PCB).drt=(PCB).token_number;
    (PCB).dssx=(PCB).ssx;
    (PCB).dsn=(PCB).sn;
  }
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  else {
    ag_prot();
    (PCB).ss[(PCB).ssx] = (PCB).sn;
  }
  (PCB).rx = 0;
  (PCB).reduction_token = (panelinterface_token_type) ag_ptt[ag_ap];
  ag_ra();
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbe[(PCB).sn] + 1;
    unsigned ag_t2 = ag_sbt[(PCB).sn+1] - 1;
    do {
      unsigned ag_tx = (ag_t1 + ag_t2)/2;
      if (ag_tstt[ag_tx] < (const unsigned char)(PCB).reduction_token) ag_t1 = ag_tx + 1;
      else ag_t2 = ag_tx;
    } while (ag_t1 < ag_t2);
    ag_ap = ag_pstt[ag_t1];
    if ((ag_r_procs_scan[ag_astt[ag_t1]])() == 0) break;
  }
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_6_proc(void) {
  int ag_sd = ag_fl[ag_ap];
  (PCB).reduction_token = (panelinterface_token_type) ag_ptt[ag_ap];
  if((PCB).drt == -1) {
    (PCB).drt=(PCB).token_number;
    (PCB).dssx=(PCB).ssx;
    (PCB).dsn=(PCB).sn;
  }
  if (ag_sd) {
    (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  }
  else {
    ag_prot();
    (PCB).vs[(PCB).ssx] = ag_null_value;
    (PCB).ss[(PCB).ssx] = (PCB).sn;
  }
  (PCB).rx = 0;
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbe[(PCB).sn] + 1;
    unsigned ag_t2 = ag_sbt[(PCB).sn+1] - 1;
    do {
      unsigned ag_tx = (ag_t1 + ag_t2)/2;
      if (ag_tstt[ag_tx] < (const unsigned char)(PCB).reduction_token) ag_t1 = ag_tx + 1;
      else ag_t2 = ag_tx;
    } while (ag_t1 < ag_t2);
    ag_ap = ag_pstt[ag_t1];
    if ((ag_r_procs_scan[ag_astt[ag_t1]])() == 0) break;
  }
  return (PCB).exit_flag == AG_RUNNING_CODE;
}


void init_panelinterface(void) {
  unsigned ag_t1 = 0;
  (PCB).rx = (PCB).fx = 0;
  (PCB).ss[0] = (PCB).sn = (PCB).ssx = 0;
  (PCB).exit_flag = AG_RUNNING_CODE;
  (PCB).key_sp = NULL;
  (PCB).key_state = 0;
  (PCB).line = FIRST_LINE;
   (PCB).column = FIRST_COLUMN;
  (PCB).btsx = 0, (PCB).drt = -1;
  while (ag_tstt[ag_t1] == 0) {
    ag_ap = ag_pstt[ag_t1];
    (ag_gt_procs_scan[ag_astt[ag_t1]])();
    ag_t1 = ag_sbt[(PCB).sn];
  }
}

void panelinterface(void) {
  (PCB).lab[(PCB).fx++] = (PCB).input_code;
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    while (1) {
      const  unsigned char *ag_p;
      unsigned ag_ch;
      if ((PCB).rx >= (PCB).fx) return;
      ag_ch = (unsigned) CONVERT_CASE((PCB).lab[(PCB).rx++]);
      if ((PCB).key_sp) {
        if (ag_ch != *(PCB).key_sp++) {
          (PCB).rx = (PCB).save_index;
          (PCB).key_sp = NULL;
          (PCB).key_state = 0;
          break;
        } else if (*(PCB).key_sp) continue;
        if (ag_key_act[(PCB).key_state] == ag_cf_end_key) {
          int ag_k1;
          int ag_k2;
          if ((PCB).rx >= (PCB).fx) {
            (PCB).rx--;
            (PCB).key_sp--;
            return;
          }
          (PCB).key_sp = NULL;
          ag_k1 = ag_key_parm[(PCB).key_state];
          ag_k2 = ag_key_pt[ag_k1];
          if (ag_key_itt[ag_k2 + CONVERT_CASE((PCB).lab[(PCB).rx])])
            (PCB).rx = (PCB).save_index;
          else {
            (PCB).token_number =  (panelinterface_token_type) ag_key_pt[ag_k1+1];
            (PCB).key_state = 0;
          }
          break;
        }
        else {
          (PCB).token_number = (panelinterface_token_type) ag_key_parm[(PCB).key_state];
          (PCB).key_state = 0;
          (PCB).key_sp = NULL;
        }
        break;
      }
      if ((PCB).key_state == 0) {
        (PCB).token_number = (panelinterface_token_type) AG_TCV(ag_ch);
        if (((PCB).key_state = ag_key_index[(PCB).sn]) == 0) break;
        (PCB).save_index = 1;
      }
      ag_p = &ag_key_ch[(PCB).key_state];
      while (*ag_p < ag_ch) ag_p++;
      if (*ag_p == ag_ch) {
        (PCB).key_state = (int)(ag_p - ag_key_ch);
        switch (ag_key_act[(PCB).key_state]) {
        case ag_cf_set_key: {
          int ag_k1;
          int ag_k2;
          if ((PCB).rx >= (PCB).fx) {
            (PCB).rx--;
            return;
          }
          ag_k1 = ag_key_parm[(PCB).key_state];
          ag_k2 = ag_key_pt[ag_k1];
          (PCB).key_state = ag_key_jmp[(PCB).key_state];
          if (ag_key_itt[ag_k2 + CONVERT_CASE((PCB).lab[(PCB).rx])]) continue;
          (PCB).save_index = (PCB).rx;
          (PCB).token_number = (panelinterface_token_type) ag_key_pt[ag_k1+1];
          continue;
        }
        case ag_set_key:
          (PCB).save_index = (PCB).rx;
          (PCB).token_number = (panelinterface_token_type) ag_key_parm[(PCB).key_state];
        case ag_jmp_key:
          (PCB).key_state = ag_key_jmp[(PCB).key_state];
          continue;
        case ag_cf_end_key:
        case ag_end_key:
          (PCB).key_sp = ag_key_ends + ag_key_jmp[(PCB).key_state];
          continue;
        case ag_accept_key:
          (PCB).token_number = (panelinterface_token_type) ag_key_parm[(PCB).key_state];
          (PCB).key_state = 0;
          break;
        case ag_cf_accept_key: {
          int ag_k1;
          int ag_k2;
          if ((PCB).rx >= (PCB).fx) {
            (PCB).rx--;
            return;
          }
          ag_k1 = ag_key_parm[(PCB).key_state];
          ag_k2 = ag_key_pt[ag_k1];
          if (ag_key_itt[ag_k2 + CONVERT_CASE((PCB).lab[(PCB).rx])])
            (PCB).rx = (PCB).save_index;
          else {
            (PCB).token_number = (panelinterface_token_type) ag_key_pt[ag_k1+1];
            (PCB).key_state = 0;
          }
          break;
        }
        }
        break;
      } else {
        (PCB).rx = (PCB).save_index;
        (PCB).key_state = 0;
        break;
      }
    }

    {
      unsigned ag_t1 = ag_sbt[(PCB).sn];
      unsigned ag_t2 = ag_sbe[(PCB).sn] - 1;
      do {
        unsigned ag_tx = (ag_t1 + ag_t2)/2;
        if (ag_tstt[ag_tx] > (const unsigned char)(PCB).token_number)
          ag_t1 = ag_tx + 1;
        else ag_t2 = ag_tx;
      } while (ag_t1 < ag_t2);
      if (ag_tstt[ag_t1] != (PCB).token_number)  ag_t1 = ag_sbe[(PCB).sn];
      ag_ap = ag_pstt[ag_t1];
      (ag_gt_procs_scan[ag_astt[ag_t1]])();
    }
  }

}


