/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

import javax.swing.*;
import javax.swing.filechooser.*;
import javax.swing.border.*;
import javax.swing.event.*;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

public class sio_panel extends JPanel implements ActionListener {
  final static int SIM_RECORD=0; /* these must be the same as those in soundcardio.h */
  final static int SIM_PLAY=1;
  final static int SIM_DUPLEX=2;
  static int PanelMode=SIM_DUPLEX;

  static JFrame frame;
  static sio_panel panel;
  
  static String Title="soundcardio/soundcardio"; /* Must be in GUI thread to access/change this */

  
  static String metal= "Metal";
  static String metalClassName = "javax.swing.plaf.metal.MetalLookAndFeel";

  static String motif = "Motif";
  static String motifClassName = "com.sun.java.swing.plaf.motif.MotifLookAndFeel";

  static String windows = "Windows";
  static String windowsClassName = "com.sun.java.swing.plaf.windows.WindowsLookAndFeel";

  public final static Border emptyBorder10 = new EmptyBorder(10,10,10,10);
  
  RecordPane recordPane;
  PlayPane playPane;

  static Thread consoleInputThread;

  // JButton button;

  
  public final static Dimension hpad10 = new Dimension(10,1);
  public final static Dimension vpad10 = new Dimension(1,10);

  //  FilePreviewer previewer;
  // JFileChooser chooser;

  class RecordPane extends JPanel implements ActionListener, ChangeListener, ItemListener {
    
    JPanel RatePanel;
    JSlider SRateSlider;    
    boolean SRateLastIsAdjusting=false; /* for jdk 1.2.2 bug workaround */   
    Timer SRateBugHack;
    JLabel SRateReadout;
    int SampleRate=44100;

    JPanel CheckboxPanel;
    JCheckBox StereoButton;
    boolean StereoFlag=true;
    JCheckBox FilterStatus;
    boolean FilterStatusFlag=true;
    
    int BufferSamples=1024;
    JPanel BufferSizePanel;
    JComboBox BufferSize;
    JLabel BufferSizeLabel;

    
    void UpdateSampleRate(int newrate) {
      
      if (SampleRate != newrate) {
	SampleRate=newrate;
	SRateSlider.setValue(SampleRate/10);
	SRateReadout.setText("" + newrate+" Hz");
	System.out.println("RecordSampleRate="+newrate);
      }
    }

    void UpdateStereoFlag(boolean newval) {
      if (newval!=StereoFlag) {
	StereoFlag=newval;
	StereoButton.setSelected(StereoFlag);
	System.out.println("RecordNumChannels="+ ((StereoFlag) ? 2 : 1));
      }
    }
    void UpdateFilterStatusFlag(boolean newval) {
      if (newval!=FilterStatusFlag) {
	FilterStatusFlag=newval;
	FilterStatus.setSelected(FilterStatusFlag);
	System.out.println("ADCMode="+ ((FilterStatusFlag) ? 1 : 0));
      }
    }
    
    void UpdateBufferSize(int NewSize) {
      if (NewSize != BufferSamples) {
	BufferSamples=NewSize;
	BufferSize.setSelectedItem(new String("" + NewSize + " samples"));
	System.out.println("RecordBufferSize="+NewSize);
      }
    }

    public RecordPane() {
      this.setLayout(new BoxLayout(this,BoxLayout.Y_AXIS));
      this.setBorder(BorderFactory.createTitledBorder("Record from soundcard"));

      
      RatePanel=new JPanel();
      RatePanel.setLayout(new BoxLayout(RatePanel,BoxLayout.X_AXIS));
      RatePanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEmptyBorder(),"Sample rate"));
      
      
      SRateSlider=new JSlider(JSlider.HORIZONTAL,4000/10,44100/10,SampleRate/10);
      SRateSlider.addChangeListener(this);
      
      SRateBugHack=new Timer(100,this);
      SRateBugHack.setRepeats(true);
      SRateBugHack.setCoalesce(true);
      SRateBugHack.start();

      SRateReadout=new JLabel(""+SampleRate+" Hz");
      
      /* Minimum size of rate readout is 10 times the width of a '5' */
      /* For some reason, specifying the minimum size of a label doesn't
	 work, so we put the label in it's own little panel, and give
	 that panel a minimum size. */
      JPanel SRateReadoutPanel=new JPanel();
      SRateReadoutPanel.add(SRateReadout);

      int minwidth=10*SRateReadout.getFontMetrics(SRateReadout.getFont()).charWidth('5');
      SRateReadoutPanel.setMinimumSize(new Dimension(minwidth,SRateReadoutPanel.getMinimumSize().height));
      /* Push that size on it VERY hard */
      SRateReadoutPanel.setPreferredSize(new Dimension(minwidth,SRateReadoutPanel.getPreferredSize().height));
      SRateReadoutPanel.setMaximumSize(new Dimension(minwidth,SRateReadoutPanel.getMaximumSize().height));
      
      RatePanel.add(SRateSlider);
      RatePanel.add(SRateReadoutPanel);

      CheckboxPanel=new JPanel();
      CheckboxPanel.setLayout(new BoxLayout(CheckboxPanel,BoxLayout.X_AXIS));
      StereoButton=new JCheckBox("Stereo",StereoFlag);
      StereoButton.addItemListener(this);
      FilterStatus=new JCheckBox("Anti alias filter",FilterStatusFlag);
      FilterStatus.addItemListener(this);
      CheckboxPanel.add(StereoButton);
      CheckboxPanel.add(Box.createRigidArea(hpad10));
      CheckboxPanel.add(FilterStatus);

      BufferSizePanel=new JPanel();
      BufferSizePanel.setLayout(new BoxLayout(BufferSizePanel,BoxLayout.X_AXIS));
      
      BufferSizeLabel=new JLabel("Buffer size");
      
      BufferSize=new JComboBox();
      BufferSize.setEditable(true);
      BufferSize.addItem("256 samples");
      BufferSize.addItem("512 samples");
      BufferSize.addItem("1024 samples");
      BufferSize.addItem("2048 samples");
      BufferSize.addItem("4096 samples");
      BufferSize.addItem("8192 samples");
      BufferSize.addItem("16384 samples");
      BufferSize.addItem("32768 samples");
      BufferSize.setSelectedItem("" + BufferSamples + " samples");
      BufferSize.addActionListener(this);

      BufferSizePanel.add(BufferSizeLabel);
      BufferSizePanel.add(Box.createRigidArea(hpad10));
      BufferSizePanel.add(BufferSize);


      this.add(RatePanel);
      this.add(CheckboxPanel);
      this.add(BufferSizePanel);
      // add(Box.createRigidArea(vpad10));
    };
   void CheckJDK122SliderHack() {
       /* if the most recent ChangeEvent had getValueIsAdjusting set,
	  and it is no longer set, then we assume the user dropped
	  the knob, so we fire off a fake ChangeEvent to ourselves */
       
       if ((SRateLastIsAdjusting) && (!SRateSlider.getValueIsAdjusting())) {
	   SRateLastIsAdjusting=false;
	   stateChanged(new ChangeEvent(SRateSlider));
       }
   
   }
 
    public void stateChanged(ChangeEvent event) {
      JComponent c = (JComponent) event.getSource();
      CheckJDK122SliderHack();
      if(c == SRateSlider) {
	if (SRateSlider.getValueIsAdjusting()) {
	  /* Update only the display, don't actually change anything */
	  SRateReadout.setText("" + SRateSlider.getValue()*10 + " Hz");
	  SRateLastIsAdjusting=true;
	} else {
	    SRateLastIsAdjusting=false;
	  UpdateSampleRate(SRateSlider.getValue()*10);
	}
	
      }
      
    }


    public void itemStateChanged(ItemEvent event) {
      JComponent c = (JComponent) event.getSource();
      CheckJDK122SliderHack();
      if(c == StereoButton) {
	UpdateStereoFlag(StereoButton.isSelected());
	
      }
      if(c == FilterStatus) {
	UpdateFilterStatusFlag(FilterStatus.isSelected());
	
      }

    }

    public void actionPerformed(ActionEvent event) {
      Object c = (Object) event.getSource();
      CheckJDK122SliderHack();
      if(c == BufferSize) {
	String Selection=(String)BufferSize.getSelectedItem();

	int WSindex=Selection.indexOf(' ');
	if (WSindex > 0)
	  Selection=Selection.substring(0,WSindex);
	//System.err.println(Selection);
	
	/* Convert selection to integer buffer size */
	int NewVal;
	try {
	  NewVal=Integer.valueOf(Selection).intValue();
	  UpdateBufferSize(NewVal);
	}
	catch(NumberFormatException e) {
	  System.err.println("Bad number in RecordBufferSize ComboBox: " + e);
	  BufferSize.setSelectedItem(new String("" + BufferSamples + " samples"));
	}
      }
      
    }
  };

  class PlayPane extends JPanel implements ActionListener {
    final int  DAC_IMPULSE=0;
    final int DAC_STEP=1;
    final int DAC_LINEAR=2;
    final int DAC_FILTERED=3;
    

    JPanel DACPanel;
    JRadioButton filteredButton;
    JRadioButton impulseButton;
    JRadioButton stepButton;
    JRadioButton linearButton;
    ButtonGroup DACGroup;
    int DACMode=DAC_FILTERED;
    
    JLabel BufferSizeLabel;
    JPanel BufferSizePanel;
    JComboBox BufferSize;
    int BufferSamples=8192;
    
    JPanel BufPosPanel;
    JProgressBar bufferPosition;
    int BufPos=50; /* in percent */

    JPanel SRatePanel;
    JProgressBar SampleRate;
    int PlaySamples=22050;


    void UpdateDACMode(int Value) {
      
      if (DACMode != Value) {
	DACMode=Value;
	
	switch (DACMode) {
	case DAC_IMPULSE:
	  impulseButton.setSelected(true);
	  break;	
  
	case DAC_STEP:
	  stepButton.setSelected(true);
	  break;	

	case DAC_LINEAR:
	  linearButton.setSelected(true);
	  break;	

	case DAC_FILTERED:
	  filteredButton.setSelected(true);
	  break;	
	  
	}
	System.out.println("DACMode="+DACMode);
      }
      
    }


    void UpdateBufferSize(int NewSize) {
      if (NewSize != BufferSamples) {
	BufferSamples=NewSize;
	BufferSize.setSelectedItem(new String("" + NewSize + " samples"));
	System.out.println("PlayBufferSize="+NewSize);
      }
    }
    
    void UpdateBufPos(int NewPos) {
      if (NewPos != BufPos) {
	BufPos=NewPos;
	bufferPosition.setValue(BufPos);
	System.out.println("PlayBufPos="+BufPos); /* This is kinda redundant,
						 but we have it for 
						 consistency */
      }
    }
    void UpdateSampleRate(int NewRate) {
      if (NewRate != PlaySamples) {
	PlaySamples=NewRate;
	SampleRate.setValue(PlaySamples);
	SampleRate.setString(""+PlaySamples + " Hz");
	System.out.println("PlaySampleRate="+PlaySamples); 
      }
    }
    
    
    public PlayPane() {
      this.setLayout(new BoxLayout(this,BoxLayout.Y_AXIS));
      this.setBorder(BorderFactory.createTitledBorder("Play to soundcard"));

      filteredButton= new JRadioButton("Filtered");
      filteredButton.setSelected(true);
      filteredButton.addActionListener(this);

      impulseButton = new JRadioButton("Impulse");
      impulseButton.addActionListener(this);
      
      stepButton = new JRadioButton("Step");
      stepButton.addActionListener(this);
      
      linearButton = new JRadioButton("Linear");
      linearButton.addActionListener(this);
      
      DACGroup=new ButtonGroup();
      DACGroup.add(filteredButton);
      DACGroup.add(impulseButton);
      DACGroup.add(stepButton);
      DACGroup.add(linearButton);

      DACPanel=new JPanel();
      DACPanel.setLayout(new BoxLayout(DACPanel,BoxLayout.X_AXIS));
      DACPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEmptyBorder(),"DAC mode"));

      //DACPanel.add(new JLabel("DAC mode"));
      //DACPanel.add(Box.createRigidArea(hpad10));
      DACPanel.add(filteredButton);
      DACPanel.add(impulseButton);
      DACPanel.add(stepButton);
      DACPanel.add(linearButton);

      this.add(DACPanel);
      


      BufferSizePanel=new JPanel();
      BufferSizePanel.setLayout(new BoxLayout(BufferSizePanel,BoxLayout.X_AXIS));
      
      BufferSizeLabel=new JLabel("Buffer size");
      
      BufferSize=new JComboBox();
      BufferSize.setEditable(true);
      BufferSize.addItem("2048 samples");
      BufferSize.addItem("4096 samples");
      BufferSize.addItem("8192 samples");
      BufferSize.addItem("16384 samples");
      BufferSize.addItem("32768 samples");
      BufferSize.addItem("65536 samples");
      BufferSize.addItem("131072 samples");
      BufferSize.addItem("262144 samples");
      BufferSize.setSelectedItem("" + BufferSamples + " samples");
      BufferSize.addActionListener(this);

      BufferSizePanel.add(BufferSizeLabel);
      BufferSizePanel.add(Box.createRigidArea(hpad10));
      BufferSizePanel.add(BufferSize);
      this.add(BufferSizePanel); 
      this.add(Box.createRigidArea(vpad10));
     
      


      bufferPosition=new JProgressBar(0,100);
      bufferPosition.setValue(BufPos);
      bufferPosition.setStringPainted(true);
      BufPosPanel=new JPanel();
      BufPosPanel.setLayout(new BoxLayout(BufPosPanel,BoxLayout.X_AXIS));
      
      BufPosPanel.add(new JLabel("Buffer Position"));
      BufPosPanel.add(Box.createRigidArea(hpad10));
      BufPosPanel.add(bufferPosition);
      
      this.add(BufPosPanel);
      this.add(Box.createRigidArea(vpad10));

      SRatePanel=new JPanel();
      SRatePanel.setLayout(new BoxLayout(SRatePanel,BoxLayout.X_AXIS));
      SRatePanel.add(new JLabel("Sample rate"));
      SRatePanel.add(Box.createRigidArea(hpad10));
      SampleRate=new JProgressBar(0,44100);
      SampleRate.setValue(PlaySamples);
      SampleRate.setString(""+PlaySamples + " Hz");
      SampleRate.setStringPainted(true);
      SRatePanel.add(SampleRate);

      this.add(SRatePanel);
      

    };
    public void actionPerformed(ActionEvent event) {
      JComponent c = (JComponent) event.getSource();
      if(c == impulseButton) {
	UpdateDACMode(DAC_IMPULSE);
      }
      if(c == stepButton) {
	UpdateDACMode(DAC_STEP);
      }
      if(c == linearButton) {
	UpdateDACMode(DAC_LINEAR);
      }
      if(c == filteredButton) {
	UpdateDACMode(DAC_FILTERED);
      }

      if(c == BufferSize) {
	String Selection=(String)BufferSize.getSelectedItem();
	
	int WSindex=Selection.indexOf(' ');
	if (WSindex > 0)
	  Selection=Selection.substring(0,WSindex);
	//System.err.println(Selection);
	
	/* Convert selection to integer buffer size */
	int NewVal;
	try {
	  NewVal=Integer.valueOf(Selection).intValue();
	  UpdateBufferSize(NewVal);
	}
	catch(NumberFormatException e) {
	  System.err.println("Bad number in PlayBufferSize ComboBox: " + e);
	  BufferSize.setSelectedItem(new String("" + BufferSamples + " samples"));
	}
      }
      
      

    }
  };
  

  public sio_panel() {
    
    setBorder(emptyBorder10);
    setLayout(new BoxLayout(this,BoxLayout.Y_AXIS));
    /*setLayout(new BorderLayout());*/

    recordPane=new RecordPane();
    playPane=new PlayPane();
    
  }
  
  public void buildpanel(int Mode) {
    if (Mode==SIM_RECORD || Mode==SIM_DUPLEX) {
      add(recordPane);
    }
    
    if (Mode==SIM_PLAY || Mode==SIM_DUPLEX) {
      add(playPane);
    }

  }

    
  public void actionPerformed(ActionEvent e) {
  }
  
  
  static class ConsoleInput implements Runnable {
    sio_panel panel;
    ConsoleInput(sio_panel parentpanel) {
      panel=parentpanel;
    }

    public void run() {
      /* console input thread begins here */
      /**** WARNING **** Practically everything in this thread
	    (except for reading from stdin) must be synchronized
	    with invokeAndWait(). See
http://java.sun.com/products/jfc/tsc/archive/tech_topics_arch/threads/threads.html
for details. */
      
      try {
	Reader r = new BufferedReader(new InputStreamReader(System.in));
	StreamTokenizer st = new StreamTokenizer(r);
	int ttype;
	
	/* Example of invokeandwait(): 
	
	SwingUtilities.invokeAndWait(new Runnable() {public void run() {
	  System.out.println("Doing invokeAndWait");
	}});
	*/
	
	
	st.resetSyntax();
	st.wordChars('a','z');
	st.wordChars('A','Z');
	st.wordChars('\u00A0','\u00ff');
	st.whitespaceChars('\u0000','\u0020');
	st.quoteChar('"');
	st.parseNumbers();
	st.eolIsSignificant(true);
	
	/* read in, and process, tokens */
	
	while ((ttype=st.nextToken()) != st.TT_EOF) {
	  switch(ttype) {
	  case st.TT_EOL:
	    break;
	    
	  case st.TT_WORD:
	    /* System.err.println("Got "+st.sval); */
	    if (st.sval.compareTo("RecordSampleRate")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {	
		final int number=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.recordPane.UpdateSampleRate(number);
		}});
	      }
	    }
	    else if (st.sval.compareTo("RecordNumChannels")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {	
		
		switch ((int)st.nval) {
		case 1:
		  SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		    panel.recordPane.UpdateStereoFlag(false);
		  }});
		  break;
		  
		case 2:
		  SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		    panel.recordPane.UpdateStereoFlag(true);
		  }});
		  break;
		  
		default:
		  System.err.println("Bad number of channels: " + (int) st.nval);
		  
		}
	      }
	    } 
	    else if (st.sval.compareTo("RecordBufferSize")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {
		final int newsize=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		  panel.recordPane.UpdateBufferSize(newsize);
		}});
	      }
	    }
	    else if (st.sval.compareTo("ADCMode")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {
		switch ((int)st.nval) {
		case 0:
		  SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		    panel.recordPane.UpdateFilterStatusFlag(false);
		  }});
		  break;
		  
		case 1:
		  SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		    panel.recordPane.UpdateFilterStatusFlag(true);
		  }});
		  break;
		  
		default:
		  System.err.println("Bad ADCMode: " + (int) st.nval);
		  
		}
		
	      }
	    }
	    else if (st.sval.compareTo("DACMode")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {
		final int number=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.playPane.UpdateDACMode(number);
		}});		
	      }
	    }
	    else if (st.sval.compareTo("PlayBufferSize")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {
		final int newsize=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		  panel.playPane.UpdateBufferSize(newsize);
		}});
	      }
	    }
	    else if (st.sval.compareTo("PlayBufPos")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {
		final int newpos=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		  panel.playPane.UpdateBufPos(newpos);
		}});
	      }
	    }
	    else if (st.sval.compareTo("PlaySampleRate")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {
		final int newrate=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		  panel.playPane.UpdateSampleRate(newrate);
		}});
	      }
	    }
	    else if (st.sval.compareTo("BeginGUI")==0) {
	      SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		panel.buildpanel(PanelMode);
		frame = new JFrame(Title);
		frame.addWindowListener(new WindowAdapter() {
		  public void windowClosing(WindowEvent e) {System.exit(0);}
		});
		frame.getContentPane().add("Center", panel);
		frame.pack();
		/* System.err.println("ready to set visible "); */
		frame.setVisible(true); /* No more GUI stuff allowed after this call, 
					   unless thread-synchronized */
		System.out.println("GUIReady");
	      }});
	    } else if (st.sval.compareTo("WindowTitle")==0) {
	      if (st.nextToken()=='=' && st.nextToken()=='"') {
		final String GotTitle=new String(st.sval);
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  Title=new String(GotTitle);
		}});
	      }
	    } else if (st.sval.compareTo("PanelMode")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {
		final int newmode=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() { 
		  PanelMode=newmode;
		}});
		
	      }
	    }
 	    else if (st.sval.compareTo("PlayNumChannels")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {	
		/* Do nothing -- we don't have a PlayNumChannels readout */
	      }
	    }
	    break;
	    
	    
	  default:
	    System.err.println("Got unknown input token: " + ttype);
	    
	  }
	  
	}
	System.exit(0);
      }
      catch(Exception except) {
	System.err.println("Error in input processing thread: "+except);
	
      }
    }
  }
  
  
  public static void main(String s[]) {
    ConsoleInput consoleInput;
    
    
    /* System.err.println("main started "); */
    
    try {
      //UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
      UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
      //UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
    } catch (Exception exc) {
      System.err.println("Error loading Look&Feel: " + exc);
    }
    
    
    panel = new sio_panel();
    /* System.err.println("sio_panel created "); */
    

    /* start up the control-input thread */
    
    consoleInput = new ConsoleInput(panel);
    consoleInputThread=new Thread(consoleInput);
    consoleInputThread.start();
  }
}
