/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <string.h>

#include <GL/glut.h>

#include "fftw.h"

#include "io_spec_old.h"
#include "oldbuffer.h"
#include "linklist.h"
#include "util.h"


/* Bug: Line 270:   glVertex2i(CIX+BlackX,CIY+CIHeight *-1*);
   should have the -1, but that doesn't erase what it should (???) */


char *Title; /* Window title */

int Win=0; /* GLUT window id */

struct SBuf *InputBuf;
int NextIsStatus=1;

int NumChannels;
int SampleRate;
int BlockSize;

int paused=0; /* is display paused */

int PassThru=0;

int UseChannel=0; /* 0=left, 1=right */

int SelectedFreq=-1;

int drawoscope=1;
int drawspectrogram=1;

void *TextFont=GLUT_BITMAP_9_BY_15;
int TFHeight=15; /* Height, in pixels, of the above text font */
int TFTopMargin=6;
int TFBottomMargin=4;

int statuslinex=0;
int statusliney=0;
int statuslinewidth=0;
int statuslineheight=0;


float gain=+0/* -10 */; /* in dB */
int oscopeperiod=100; /* period of oscilloscope traces, in ms */
int oscopemax; /* calculated with calcoscopemax() */

float *oscopebuf=NULL; /* buffer for storing data to be displayed with the
		     oscilloscope */
float *oldoscopebuf=NULL; /* buffer for storing data that is on the screen
			     (so we know how to erase it) */

int oscopebufsize=0;
int oscopebufpos=0;

/* oscilloscope display position */
int oscopex,oscopey;
int oscopewidth,oscopeheight;


struct oglpixel {
  GLfloat R,G,B,Alpha;
};

struct oglpixel *CurImage=NULL;
int CIX,CIY; /* (X,Y) of upper left corner of image */
int CIWidth,CIHeight;


int CurX=0; /* X coordinate of next fft column to draw, relative to CIX */
struct oglpixel *CurCol=NULL;
struct oglpixel BlackPixel={
  0.0,0.0,0.0,1.0,
};

float ScaleFactor;


int WindowSizeX=0;
int WindowSizeY=0;

float PixelSizeX=0.0,PixelSizeY;

int FFTSize=2048;
int FFTPeriod=50; /* in ms -- 50 ms between FFT's */
int FFTSamplePeriod; /* FFTPeriod, in samples instead of ms */

int FFTPos=0; /* Number of samples which have been written into the
		 FFT buffer */
FFTW_COMPLEX *FFTBuf;
FFTW_COMPLEX *FFTWindowedBuf;
FFTW_COMPLEX *FFTOutBuf;

int doingreshape=1; /* this is non-zero if we've just done a reshape, but
		       haven't yet gotten the display callback. It disables
		       other operations such as reading from standard input
		       during this period */




void calcoscopemax(void)
     /* recalculate the oscopemax global.
	Needs to be called if oscopeperiod or SampleRate changes */
{
  oscopemax=oscopeperiod*SampleRate/1000.0; 
}

void ProcessOscopebuf(void)
     /* process and display the oscilloscope buffer */
{
  float CenterPos;
  int Cnt;

  float YScale;
  int YPos;
  float *swap;



  glBegin(GL_POINTS);
  
  
  YScale=oscopeheight/2.0;
  YPos=oscopey+oscopeheight/2;

  /* first, erase the previous data */
  if (oldoscopebuf) {
    glColor3f(0.0,0.0,0.0); /* black */
    for (Cnt=0;Cnt < oscopebufsize;Cnt++) {
      glVertex2i(Cnt,YPos-(int)(YScale*oldoscopebuf[Cnt]));
    }
  }
  
  
  glColor3f(1.0,1.0,1.0); /* white */
  
  for (Cnt=0;Cnt < oscopebufsize;Cnt++) {
    glVertex2i(Cnt,YPos-(int)(YScale*oscopebuf[Cnt]));
  }
  glEnd();
  /* now swap the old and new oscopebufs */

  swap=oldoscopebuf;
  oldoscopebuf=oscopebuf;
  oscopebuf=swap;
  if (!oscopebuf) oscopebuf=malloc(oscopebufsize*sizeof(float));
  
}

void CreateFFTBuf(void)
{
  int Cnt;

  FFTBuf=malloc(FFTSize*sizeof(FFTW_COMPLEX));
  FFTOutBuf=malloc(FFTSize*sizeof(FFTW_COMPLEX));
  FFTWindowedBuf=malloc(FFTSize*sizeof(FFTW_COMPLEX));

  for (Cnt=0;Cnt < FFTSize;Cnt++) {
    FFTBuf[Cnt].im=0.0;
  }
  for (Cnt=0;Cnt < FFTSize;Cnt++) {
    FFTWindowedBuf[Cnt].im=0.0;
  }


  FFTPos=0;
}


void ProcessFFTBuf(void)
{
  fftw_plan plan;
  int Cnt;
  float Value;
  float re,im;

  float h,s,v,r,g,b;
  int k;
  float f,n,t,p,m1;

  int BlackX;



  /* perform windowing with raised cosine function */
  
  
  
  for (Cnt=0;Cnt < FFTSize;Cnt++) {
    /*    FFTBuf[Cnt].re=FFTBuf[Cnt].re*cos(((float)(FFTSize-Cnt))/(.15*FFTSize)*(M_PI/2));*/
    FFTWindowedBuf[Cnt].re=FFTBuf[Cnt].re*(0.5+0.5*cos(((float)(Cnt))/(FFTSize)*2*M_PI + M_PI));
    
  }


  /* Perform the FFT */

  plan = fftw_create_plan(FFTSize, FFTW_FORWARD, FFTW_ESTIMATE); /* use FFTW_MEASURE instead of FFTW_ESTIMATE in order to use the wisdom functionality (slightly faster) See FFTW documentation. This might be desirable since the ESTIMATE optimization is for a RISC machine with lots of registers -- probably bad for the i386 */
  
  
  
  fftw(plan, 1, FFTBuf, 1, 0, FFTOutBuf, 1, 0);
  fftw_destroy_plan(plan);
  /* Now take the FFT'd data and stick it into CurCol */
  
  for (Cnt=0;Cnt < CIHeight;Cnt++) {
    re=FFTOutBuf[Cnt*FFTSize/(2*CIHeight)].re;
    im=FFTOutBuf[Cnt*FFTSize/(2*CIHeight)].im;
    Value=log(sqrt(re*re+im*im));
    /* Map the Value to a color (hsv) */

    /* Color scheme:
        0 < Value < .2     black
        .2 < Value < .3    color brightness increases to full. Changing hue
	.3 < Value < .8    color brightness max, changing hue
        .8 < Value < .9    saturation goes to zero, changing hue
        .9 < Value         white 
	*/
    if (Value < .05) {
      v=0; /* black below .05 */
    } else if (Value < .1) {
      v=(Value-.05)*20.0; /* intensity increases to full brightness at Value=.1 */
    } else {
      v=1.0;
    }

    
    if (Value < .075) {
      h=0.0; /* color doesn't matter when it's black */
    } else if (Value < 2.25) {
      h=(Value-.075)/(2.25-.075); /* Have the hue change throughout the entire
			  active region */
    } else {
      h=1.0; /* hue doesn't matter when it's white */
    }
    
    if (Value < 2.2) {
      s=1.0; /* Full saturation until the white transition at .86 */
    } else if (Value < 2.3) {
      s=1.0-(Value-2.2)/(2.3-1.2); /* saturation fades to zero by .9 */
    } else {
      s=0.0;
    }

    
    /* Map the (hsv) to a color (rgb) */
    /* This algorithm is based on the one from MATLAB, which comes from:
       Alvy Ray Smith, Color Gamut Transform Pairs, SIGGRAPH '78.
       C. B. Moler, 8-17-86, 5-10-91, 2-2-92. */
    
    /*
    if (h > 1.0) h=1.0;
    if (h < 0.0) h=0.0;
    if (s > 1.0) s=1.0;
    if (s < 0.0) s=0.0;
    if (v > 1.0) v=1.0;
    if (v < 0.0) v=0.0;
    */
    

    h*=6.0;
    k=(int)(h-1.e-4);
    if (k < 0) k=0;
    
    f=h-k;
    n=1-s*f;
    t=1-s;
    p=1-(s*(1-f));
    r=(k==0)+n*(k==1)+t*(k==2)+t*(k==3)+p*(k==4)+(k==5);
    g=p*(k==0)+(k==1)+(k==2)+n*(k==3)+t*(k==4)+t*(k==5);
    b=t*(k==0)+t*(k==1)+p*(k==2)+(k==3)+(k==4)+n*(k==5);
    m1= (r>g) ? r:g;
    f=v/((m1 > b) ? m1:b);
    
    r*=f;
    g*=f;
    b*=f;
    
    CurCol[Cnt].R=r;
    CurCol[Cnt].G=g;
    CurCol[Cnt].B=b;
    CurCol[Cnt].Alpha=1.0;
    
    
  }
  
  /* Write the buffer to the screen */
  /*  fprintf(stderr,"doing write X=%f Y=%f\n",(float)(((float)(CIX+CurX)*2.0)/WindowSizeX),(float)(((float)CIY*2.0)/WindowSizeY));*/
      

  glRasterPos2i(CIX+CurX,CIY+CIHeight-1);
  glDrawPixels(1,CIHeight,GL_RGBA,GL_FLOAT,CurCol);
  
  /* Copy the buffer to our record of what's on the screen */
  
  for (Cnt=0; Cnt < CIHeight;Cnt++) {
    CurImage[Cnt*CIWidth + CurX]=CurCol[Cnt];
  }
  
  /* Blank out the column 20 pixels ahead */

  BlackX=(CurX+20)%CIWidth; /* X coordinate of the column we will blank out */

  glBegin(GL_LINES);
  glColor3f(0.0,0.0,0.0); /* Black */
  glVertex2i(CIX+BlackX,CIY);
  glVertex2i(CIX+BlackX,CIY+CIHeight/*-1*/);
  glEnd();
  

  
  /* Blank it out in the CurImage */
  for (Cnt=0; Cnt < CIHeight;Cnt++) {
    CurImage[Cnt*CIWidth + BlackX]=BlackPixel;
  }
  
  
  /* Update the X counter */
  CurX++;
  CurX %= CIWidth;
  
}

void CopyInputBufToFFTBuf(void)
     /* this also copies it to the o-scope buffer */
{
  int Cnt;
  float Amnt;
  int Channel;

  if (paused) {
    /* Throw away buffers and do nothing */
    FFTPos=0;
    oscopebufpos=0;
    return;
    
  }
  
  for (Cnt=0; Cnt < BlockSize;Cnt++) {
    Amnt=0.0;
    /*    for (Channel=0;Channel < NumChannels;Channel++) {
      Amnt+=  ((float)((INT16) SBufEntry(InputBuf,Cnt,Channel)))/32768;
      }*/
    Amnt=  ((float)((INT16) SBufEntry(InputBuf,Cnt,UseChannel)))/32768;
    
    if (drawspectrogram) {
      if (FFTPos < FFTSize) {
	FFTBuf[FFTPos].re=Amnt*pow(10.0,gain/20.0);
	/*    FFTBuf[FFTPos].im=0.0;  (redundant -- unnecessary) */
      }
      FFTPos++;
      
      if (FFTPos==FFTSize) {
	/* We've filled up the buffer -- FFT it and write it to the display */
	
	ProcessFFTBuf();
	if (FFTSize > FFTSamplePeriod) {
	  /* Our FFT's are overlapping. Let's move the data down in 
	     memory and continue recording */
	  memmove(FFTBuf,FFTBuf+FFTSamplePeriod,(FFTSize-FFTSamplePeriod)*sizeof(FFTW_COMPLEX));
	  FFTPos=FFTSize-FFTSamplePeriod;
	}
      }
      
      if (FFTPos >= FFTSize && FFTPos>=FFTSamplePeriod) {
	/* Time to start saving data for another FFT */
	FFTPos=0;
	
      }
      
    }
    if (drawoscope) {
      if (oscopebufpos < oscopebufsize) {
	oscopebuf[oscopebufpos]=Amnt;
	oscopebufpos++;
	if (oscopebufpos==oscopebufsize) {
	  /* now we can flush it out */
	  ProcessOscopebuf();
	}
      } else if (oscopebufpos < oscopemax) {
	oscopebufpos++;
      } else {
	oscopebufpos=0;
      }
    }
    

  }
}

void DrawString(char *String)
     /* Starting position should be set before this call with glRasterPos() */
     /* The color should already be set before this call with glColor() */
{
  int Cnt;
  
  for (Cnt=0;String[Cnt];Cnt++) {
    glutBitmapCharacter(TextFont,String[Cnt]);
  }
}


void DrawStatusLine(void)
{
  /* erase any previous contents by drawing a black rectangle */
  char StatusStringBuf[200];
  
  glBegin(GL_QUADS);
  glColor3f(0.0,0.0,0.0);
  glVertex2i(statuslinex,statusliney);
  glVertex2i(statuslinex+statuslinewidth,statusliney);
  glVertex2i(statuslinex+statuslinewidth,statusliney+statuslineheight);
  glVertex2i(statuslinex,statusliney+statuslineheight);
  glEnd();
  glColor3f(1.0,1.0,1.0);

  /* Write text */
  snprintf(StatusStringBuf,199,"Nyquist: %ld Hz",SampleRate/2);
  glRasterPos2i(statuslinex+/*statuslinewidth/5*/statuslinewidth/12,statusliney+TFTopMargin+TFHeight);
  DrawString(StatusStringBuf);
  
  if (SelectedFreq >= 0)
    snprintf(StatusStringBuf,199,"Selected: %ld Hz",SelectedFreq);
  else
    snprintf(StatusStringBuf,199,"Selected:");
  glRasterPos2i(statuslinex+statuslinewidth*5/12,statusliney+TFTopMargin+TFHeight);
  DrawString(StatusStringBuf);
  

  snprintf(StatusStringBuf,199,"gain: %1.1f dB",(float)gain);
  glRasterPos2i(statuslinex+statuslinewidth*9/12/*statuslinewidth*3/5*/,statusliney+TFTopMargin+TFHeight);
  DrawString(StatusStringBuf);
  

}

void ReadStdin(void)
{
  /* This is the GLUT idle-function. We select() on stdin, waiting up to
     100ms for input to become available. */

  fd_set readfd;
  struct timeval timeout;
  int BufFinished;
  struct StatusBlk *Status;
  int MajorChange;
  int Cnt=0;


  if (doingreshape) return; /* let the display callback come first */

  FD_ZERO(&readfd);
  
  FD_SET(0,&readfd);
  timeout.tv_sec=0;
  timeout.tv_usec=100000;
  
  if (!CurImage) return; /* wait for window size to be set up */
  
  select(1,&readfd,NULL,NULL,&timeout);
  do {
    BufFinished=ReadSBuf(InputBuf,0);
    if (BufFinished) {
      if (SBufIsStatus(InputBuf)) {
	assert(NextIsStatus);
	NextIsStatus=0;

	if (PassThru) {
	  /* Write it out to the passthru */
	  SBufPrepareStatus(InputBuf);
	  while (!WriteSBuf(InputBuf,1)); /* Write it to stdout */
	}
	
	MajorChange=0;
	Status=SBufGetStatusBlk(InputBuf);
	if (NumChannels != Status->NumChannels) {
	  MajorChange=1;
	}
	NumChannels=Status->NumChannels;
	/* fprintf(stderr,"spectr: got status block: NumChannels=%d, SampleRate=%d,FrameNumber=%d,BlockSize=%d\n",(int)Status->NumChannels,(int)Status->SampleRate,(int)Status->FrameNumber,(int)Status->BlockSize); */
	
	if (SampleRate != Status->SampleRate) {
	  MajorChange=1;
	}
	SampleRate=Status->SampleRate;
	
	BlockSize=Status->BlockSize;
	SBufReallocateIfNeeded(InputBuf);

	if (MajorChange) {
	  FFTSamplePeriod=FFTPeriod*SampleRate/1000.0;
	  calcoscopemax(); /* recalculate the oscopemax global */
	  DrawStatusLine();
	}
	
	SBufPrepare(InputBuf);
      } else {
	/* We just got a non-status block */
	assert(!NextIsStatus);
	
	CopyInputBufToFFTBuf();
	if (PassThru) {
	  /* Write it out to pass-thru */
	  SBufPrepare(InputBuf);
	  while (!WriteSBuf(InputBuf,1)); /* Write it to stdout */
	}
	if ((SBufGetDataHeader(InputBuf))->NextBlockIsStatusBlk) {
	  NextIsStatus=1;
	  SBufPrepareStatus(InputBuf);
	} else	
	  SBufPrepare(InputBuf);
	
      }
    }
    
    Cnt++;
  } while (BufFinished && Cnt < 1);
  
}

void WinMouseFunc(int button,int state,int x, int y)
{
  if (button==GLUT_LEFT_BUTTON && state==GLUT_DOWN) {
    if (y > CIY && y < (CIY+CIHeight)) {
      SelectedFreq=(CIY-y+CIHeight)*SampleRate/2/CIHeight;
    } else 
      SelectedFreq=-1;
  }
  DrawStatusLine();
}


void WinKeyboardFunc(unsigned char key, int x, int y) 
{
  /* keyboard call for 'regular' keys. see also WinSpecialFunc() */

  if (paused) {
    paused=0;
    return;
  }

  switch (key) {
  case ' ': /* space pauses output */
    paused=1;
    break;
    

  }
}

void WinSpecialFunc(int key, int x, int y)
     /* callback for processing 'special' keys -- cursor keys, function keys, etc. */
{
  switch (key) {
  case GLUT_KEY_UP:
    gain+=1.5;
    DrawStatusLine();
    break;

  case GLUT_KEY_DOWN:
    gain-=1.5;
    DrawStatusLine();
    break;
    
  }
}

void WinDisplayFunc(void)
{
  /* Redraw our display */


  
  glClear(GL_COLOR_BUFFER_BIT);

  if (!doingreshape) {
    /* for a resize, there's nothing in the buffer, so no point in
       redrawing */
    glRasterPos2f(0.0,CIHeight-1);
    glDrawPixels(CIWidth,CIHeight,GL_RGBA,GL_FLOAT,CurImage);
  }
  else doingreshape=0;

  DrawStatusLine();
}

void WinReshapeFunc(int Width, int Height)
     /* called on initialization and when the window is resized */
{
  int XCnt,YCnt;
  int TFTotalHeight;

  TFTotalHeight=TFHeight+TFTopMargin+TFBottomMargin; /* total height of text area */
  
  /*  fprintf(stderr,"Got Reshape()\n");*/
  
  doingreshape=1;

  glViewport(0,0,Width,Height);

  glClear(GL_COLOR_BUFFER_BIT);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(-0.0001,Width,Height,-0.0001); /* the epsilons make the 
						 rounding come out right,
						 at least on Mesa */

  WindowSizeX=Width;
  WindowSizeY=Height;
  
    
  /* Lay out window */
  if (drawoscope && drawspectrogram) {
    CIX=0;
    CIY=0;
    
    CIWidth=Width;
    CIHeight=Height*2/3 - TFTotalHeight/2;

    statuslinex=0;
    statusliney=CIY+CIHeight;
    statuslineheight=TFTotalHeight;
    statuslinewidth=Width;
    

    oscopex=0;
    oscopey=statusliney+statuslineheight;    
    oscopewidth=Width;
    oscopeheight=Height-oscopey;
    
  } else if (!drawoscope && drawspectrogram) {
    CIX=0;
    CIY=0;
    
    CIWidth=Width;
    CIHeight=Height-TFTotalHeight;

    statuslinex=0;
    statusliney=CIY+CIHeight;
    statuslinewidth=Width;
    statuslineheight=TFTotalHeight;

    oscopex=0;
    oscopey=0;
    oscopewidth=0;
    oscopeheight=0;
  } else if (drawoscope && !drawspectrogram) {
    CIX=0;
    CIY=0;
    CIWidth=0;
    CIHeight=0;
    
    statuslinex=0;
    statusliney=0;
    statuslinewidth=Width;
    statuslineheight=TFTotalHeight;
    
    oscopex=0;
    oscopey=statusliney+statuslineheight;
    oscopewidth=Width;
    oscopeheight=Height-oscopey;
  } else {
    fprintf(stderr,"spectr: Must draw either oscilloscope or spectrogram (or both)\n");
    exit(1);
  }
  
  
  oscopebufsize=Width;
  if (oscopebuf) free(oscopebuf);
  oscopebuf=malloc(oscopebufsize*sizeof(float));
  oscopebufpos=0;
  if (oldoscopebuf) {
    free(oldoscopebuf);
    oldoscopebuf=NULL; /* this will be reallocated in ProcessOscopebuf() */
  }
  
  
  
  /* Reallocate image buffers */
  
  CurX=0;
  if (CurImage) free(CurImage);
  CurImage=malloc(sizeof(struct oglpixel)*CIWidth*CIHeight);
  /* Clear the image */
  for (YCnt=0;YCnt < CIHeight;YCnt++) {
    for (XCnt=0;XCnt < CIWidth;XCnt++) {
      CurImage[YCnt*CIWidth + XCnt].R=0.0;
      CurImage[YCnt*CIWidth + XCnt].G=0.0;
      CurImage[YCnt*CIWidth + XCnt].B=0.0;
      CurImage[YCnt*CIWidth + XCnt].Alpha=1.0;
    }
  }

  if (CurCol) free(CurCol);
  CurCol=malloc(sizeof(struct oglpixel)*CIHeight);
  for (YCnt=0;YCnt < CIHeight;YCnt++) {
    CurImage[YCnt].R=0.0;
    CurImage[YCnt].G=0.0;
    CurImage[YCnt].B=0.0;
    CurImage[YCnt].Alpha=1.0;
    
  }

  
}

int main(int argc, char *argv[])
{
  int Cnt;
  int GotMagic=0;
  int windowwidth=700;
  int windowheight=500;
  int windowx=50;
  int windowy=50;

  Title=argv[0];
  
  /* Process arguments */
  
  for (Cnt=1;argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    else if (!strcmp(argv[Cnt],"-passthru")) {
      PassThru=1;
    } else if (!strncmp(argv[Cnt],"-fftsize=",9)) {
      FFTSize=strtol(argv[Cnt]+9,NULL,10);
    } else if (!strncmp(argv[Cnt],"-gain=",6)) {
      gain=strtol(argv[Cnt]+6,NULL,10);
    } else if (!strncmp(argv[Cnt],"-fftperiod=",11)) {
      FFTPeriod=strtol(argv[Cnt]+11,NULL,10);
    } else if (!strncmp(argv[Cnt],"-oscopeperiod=",14)) {
      oscopeperiod=strtol(argv[Cnt]+14,NULL,10);
    } else if (!strcmp(argv[Cnt],"-left")) {
      UseChannel=0;
    } else if (!strcmp(argv[Cnt],"-right")) {
      UseChannel=1;
    } else if (!strcmp(argv[Cnt],"-nospectrogram")) {
      drawspectrogram=0;
    } else if (!strcmp(argv[Cnt],"-nooscope")) {
      drawoscope=0;
    } else if (!strncmp(argv[Cnt],"-title=",7)) {
      Title=argv[Cnt]+7;
    } else if (!strncmp(argv[Cnt],"-windowwidth=",13)) {
      windowwidth=strtol(argv[Cnt]+13,NULL,10);
    } else if (!strncmp(argv[Cnt],"-windowheight=",14)) {
      windowheight=strtol(argv[Cnt]+14,NULL,10);
    } else if (!strncmp(argv[Cnt],"-windowx=",9)) {
      windowx=strtol(argv[Cnt]+9,NULL,10);
    } else if (!strncmp(argv[Cnt],"-windowy=",9)) {
      windowy=strtol(argv[Cnt]+9,NULL,10);
    }
    else {
      fprintf(stderr,"Unknown argument: %s\n",argv[Cnt]);
      exit(1);
    }
  }


  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   signal input\n"
	    "        #1   optional pass-thru of input\n"
	    "        #2   error messages (do not connect)\n"
	    "\n"
	    "Usage:\n"
	    "   %s -magic -passthru -fftsize=%ld -gain=%f -fftperiod=%ld \n"
	    "       -oscopeperiod=%ld -left/-right\n -nospectrogram -nooscope\n"
	    "       -windowwidth=%d -windowheight=%d windowx=%d windowy=%d\n"
	    "\n"
	    "   fftsize is the number of samples corresponding to each\n"
	    "   vertical line drawn in the spectrogram. CPU usage will be\n"
	    "   decreased when you increase fftsize.\n"
	    "   gain applies to spectrogram only, and is in dB.\n"
	    "   -oscopeperiod specifies the time between oscilloscope traces,\n"
	    "   in ms.\n"
	    "   -left and -right tell the program to take input from the\n"
	    "   left or right channel of a stereo signal. The default is \n"
	    "   -left\n"
	    ,argv[0],(long)FFTSize,(float)gain,(long)FFTPeriod,(long)oscopeperiod,windowwidth,windowheight,windowx,windowy);
    exit(1);
  }
  


  setnonblocking(0); /* set stdin to non-blocking i/o */
  
  InputBuf=CreateSBuf();
  SBufPrepareStatus(InputBuf);

  CreateFFTBuf();
  
  glutInit(&argc,argv);
  glutInitWindowSize(windowwidth,windowheight);
  glutInitWindowPosition(windowx,windowy);
  glutInitDisplayMode(GLUT_RGBA|GLUT_SINGLE);
  Win=glutCreateWindow(Title);
  
  glutReshapeFunc(WinReshapeFunc);
  glutDisplayFunc(WinDisplayFunc);
  glutSpecialFunc(WinSpecialFunc);
  glutKeyboardFunc(WinKeyboardFunc);

  glutMouseFunc(WinMouseFunc);

  glutIdleFunc(ReadStdin);
  glutMainLoop();
  
  return 0;
}
