/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "io_spec_old.h"
#include "oldbuffer.h"
#include "linklist.h"
#include "util.h"

struct SBuf *Buffer;

int BlockSize;
int NumChannels;

int InputPipe=0;
int OutputPipeA=1;
int OutputPipeB=3;

int main(int argc, char *argv[])
{
  int GotMagic=0;
  struct StatusBlk *Status;
  int Cnt;

  for (Cnt=1; argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
  }
  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   input stream\n"
	    "        #1   output stream A\n"
	    "        #2   error messages (do not connect)\n"
	    "        #3   output stream B\n"
	    "\n"
	    "Usage:\n"
	    "   %s -magic\n"
	    ,argv[0]);
    exit(1);
  }
  
  Buffer=CreateSBuf();
  SBufPrepareStatus(Buffer);
  
  for (;;) {
    /* Read data in, then write it to both output pipes */
    while (!ReadSBuf(Buffer,InputPipe)); 
 
   if (SBufIsStatus(Buffer)) {
     Status=SBufGetStatus(Buffer);
     
     /* Collect data */
     BlockSize=Status->BlockSize;
     NumChannels=Status->NumChannels;

     /* write it out to channel A */
     SBufPrepareStatus(Buffer);
     while(!WriteSBuf(Buffer,OutputPipeA));
     
     /* and B */
     SBufPrepareStatus(Buffer);
     while(!WriteSBuf(Buffer,OutputPipeB));
     
     SBufReallocateIfNeeded(Buffer);
     
     /* Now get the next data block */
     SBufPrepare(Buffer);
     while(!ReadSBuf(Buffer,InputPipe));
     
   }
   /* OK. We just got a data buffer */

   /* Send it out pipe A */
   SBufPrepare(Buffer);
   while(!WriteSBuf(Buffer,OutputPipeA));

   /* and pipe B */
   SBufPrepare(Buffer);
   while(!WriteSBuf(Buffer,OutputPipeB));
   
   /* go on and read the next buffer */
    if ( (SBufGetDataHeader(Buffer))->NextBlockIsStatusBlk) 
      SBufPrepareStatus(Buffer);
    else
      SBufPrepare(Buffer);
  }

  return 0;
}
