/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

import javax.swing.*;
import javax.swing.filechooser.*;
import javax.swing.border.*;
import javax.swing.event.*;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

public class syncswitch_panel extends JPanel implements ActionListener,ItemListener {
  static JFrame frame;
  static syncswitch_panel panel;
    
  
  static String Title="Synchronizing Switch";

  public final static Border emptyBorder10 = new EmptyBorder(10,10,10,10);
  static Thread consoleInputThread;
  public final static Dimension hpad10 = new Dimension(10,1);
  public final static Dimension vpad10 = new Dimension(1,10);

  public final static int STATE_VISIBLE=0;
  public final static int STATE_BLIND=1;

  int map[]; /* length[2];  map[external_id] = internal button number */
  int imap[]; /* length[2];  inverse map --  imap[internal button number]=external_id */
  int State; /* either STATE_VISIBLE or STATE_BLIND */

  int internalselected; /* either 0 or 1 */
  boolean jumpback; /* is jumpback mode on or off? */

  JRadioButton InputButton[]; /* length 2 */
  JButton RevealExitButton;
  JTextField StatusLine;
  JCheckBox JumpBackBox;
  
  String BlindLabel[]; /* length 2, indexed by internal button number */
  String InputLabel[]; /* length 2, indexed by external button number */

  String Message;
  String RevealMessage[]; /* length 2, indexed by external button number
			     Message to print when revealing if corresponding
			     button is selected */

  

  void externalsetbutton(int external_id)
  {
    int internal_id;
    internal_id=map[external_id];
    
    internalsetbutton(internal_id);
  }
  
  void internalsetbutton(int internal_id)
  {
    if (internal_id != internalselected) {
      internalselected=internal_id;
      InputButton[internal_id].setSelected(true);
      System.out.println("SelectedInput="+imap[internal_id]);
    }
  }

  void printselected() 
  {
    System.out.println("SelectedInput="+imap[internalselected]);    
  }

  void setjumpback(boolean newval)
  {
    if (newval != jumpback) {
      jumpback=newval;
      JumpBackBox.setSelected(jumpback);
      System.out.println("JumpBack="+((jumpback) ? 1:0));
    }
  }

  void printjumpback()
  {
      System.out.println("JumpBack="+((jumpback) ? 1:0));    
  }
  
  void SetVisibleLabels()
  {
    /* Set all labels with the values they should have 
       in STATUS_VISIBLE */
    StatusLine.setText(Message);
    RevealExitButton.setText("Exit");
    InputButton[0].setText(InputLabel[imap[0]]);
    InputButton[1].setText(InputLabel[imap[1]]);
    
  }
  
  void SetBlindLabels()
  {
    StatusLine.setText(Message);
    RevealExitButton.setText("Reveal");
    InputButton[0].setText(BlindLabel[0]);
    InputButton[1].setText(BlindLabel[1]);
    
  }

  
  public syncswitch_panel() {
    setBorder(emptyBorder10);
    setLayout(new BorderLayout());

    map=new int[2];
    imap=new int[2];

    BlindLabel=new String[2];
    InputLabel=new String[2];
    RevealMessage=new String[2];



    JLabel SelectLabel=new JLabel("Select Input");
    add(SelectLabel,BorderLayout.NORTH);

    InputButton=new JRadioButton[2];

    JPanel RadioPane=new JPanel();
    ButtonGroup InputSel=new ButtonGroup();
    RadioPane.setLayout(new BoxLayout(RadioPane,BoxLayout.Y_AXIS));
    InputButton[0]=new JRadioButton("Input 1");
    InputButton[1]=new JRadioButton("Input 2");
    InputSel.add(InputButton[0]);
    InputSel.add(InputButton[1]);
    InputButton[0].addActionListener(this);
    InputButton[1].addActionListener(this);

    RadioPane.add(Box.createRigidArea(vpad10));
    RadioPane.add(Box.createVerticalGlue());
    RadioPane.add(InputButton[0]);
    RadioPane.add(Box.createVerticalGlue());
    RadioPane.add(InputButton[1]);
    RadioPane.add(Box.createVerticalGlue());
    RadioPane.add(Box.createRigidArea(vpad10));
    add(RadioPane,BorderLayout.CENTER);
    
    JPanel RightPanel=new JPanel();
    RightPanel.setLayout(new BoxLayout(RightPanel,BoxLayout.Y_AXIS));
    RightPanel.add(Box.createVerticalGlue());
    JumpBackBox=new JCheckBox("Jump back");
    JumpBackBox.addItemListener(this);
    RightPanel.add(JumpBackBox);
    RightPanel.add(Box.createVerticalGlue());
    RevealExitButton=new JButton("Reveal");
    RevealExitButton.addActionListener(this);
    RightPanel.add(RevealExitButton);
    RightPanel.add(Box.createVerticalGlue());
    add(RightPanel,BorderLayout.EAST);
    
    
    StatusLine=new JTextField("This is the status line",20);
    StatusLine.setEditable(false);
    add(StatusLine,BorderLayout.SOUTH);
    

    /* The following initialization should be un-done later
       when parameters are set by the calling program. If parameters
       are not set, these are the defaults: */
    internalselected=0;
    jumpback=true;

    map[0]=0;
    map[1]=1;
    imap[0]=0;
    imap[1]=1;
    State=STATE_VISIBLE;
    BlindLabel[0]="First input";
    BlindLabel[1]="Second input";
    
    InputLabel[0]="Input A";
    InputLabel[1]="Input B";
    Message="Please select between inputs";
    RevealMessage[0]="You selected input A";
    RevealMessage[1]="You selected input B";
    SetVisibleLabels();
    

  }

  public void actionPerformed(ActionEvent event) {
    JComponent c = (JComponent) event.getSource();
    if(c == InputButton[0]) {
      internalsetbutton(0);
    }
    if (c==InputButton[1]) {
      internalsetbutton(1);
    }
    if (c==RevealExitButton) {
      if (State==STATE_VISIBLE) {
	/* User pressed "Exit" button */
	System.exit(0);
      }
      else {
	/* user pressed "Reveal button -- change to STATE_VISIBLE */
	State=STATE_VISIBLE;
	Message=RevealMessage[imap[internalselected]];
	SetVisibleLabels();
      }
    }
  }


  public void itemStateChanged(ItemEvent event) {
    
    JComponent c = (JComponent) event.getSource();
    if (c==JumpBackBox) {
      setjumpback(JumpBackBox.isSelected());
    }
    
  }

  static class ConsoleInput implements Runnable {
    
    syncswitch_panel panel;
    ConsoleInput(syncswitch_panel parentpanel) {
      panel=parentpanel;
    }

    public void run() {
      /* console input thread begins here */
      /**** WARNING **** Practically everything in this thread
	    (except for reading from stdin) must be synchronized
	    with invokeAndWait(). See
http://java.sun.com/products/jfc/tsc/archive/tech_topics_arch/threads/threads.html
for details. */
      
      try {
	Reader r = new BufferedReader(new InputStreamReader(System.in));
	StreamTokenizer st = new StreamTokenizer(r);
	int ttype;
	
	/* Example of invokeandwait(): 
	
	SwingUtilities.invokeAndWait(new Runnable() {public void run() {
	  System.out.println("Doing invokeAndWait");
	}});
	*/
	
	
	st.resetSyntax();
	st.wordChars('a','z');
	st.wordChars('A','Z');
	st.wordChars('\u00A0','\u00ff');
	st.wordChars('?','?');
	st.whitespaceChars('\u0000','\u0020');
	st.quoteChar('"');
	st.parseNumbers();
	st.eolIsSignificant(true);
	
	/* read in, and process, tokens */
	
	while ((ttype=st.nextToken()) != st.TT_EOF) {
	  switch(ttype) {
	    
	  case st.TT_EOL:
	    break;
	    
	  case st.TT_WORD:
	    if (st.sval.compareTo("InputLabel0")==0) {
	      if (st.nextToken()=='=' && st.nextToken()=='"') {	
		final String GotInputLabel=new String(st.sval);
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.InputLabel[0]=GotInputLabel;
		}});
	      }
	    }
	    else if (st.sval.compareTo("InputLabel1")==0) {
	      if (st.nextToken()=='=' && st.nextToken()=='"') {	
		final String GotInputLabel=new String(st.sval);
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.InputLabel[1]=GotInputLabel;
		}});
	      }
	    }
	    else if (st.sval.compareTo("BlindLabel0")==0) {
	      if (st.nextToken()=='=' && st.nextToken()=='"') {	
		final String GotBlindLabel=new String(st.sval);
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.BlindLabel[0]=GotBlindLabel;
		}});
	      }
	    }
	    else if (st.sval.compareTo("BlindLabel1")==0) {
	      if (st.nextToken()=='=' && st.nextToken()=='"') {	
		final String GotBlindLabel=new String(st.sval);
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.BlindLabel[1]=GotBlindLabel;
		}});
	      }
	    }
	    else if (st.sval.compareTo("Message")==0) {
	      if (st.nextToken()=='=' && st.nextToken()=='"') {	
		final String GotMessage=new String(st.sval);
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.Message=GotMessage;
		}});
	      }
	    }
	    else if (st.sval.compareTo("RevealMessage0")==0) {
	      if (st.nextToken()=='=' && st.nextToken()=='"') {	
		final String GotMessage=new String(st.sval);
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.RevealMessage[0]=GotMessage;
		}});
	      }
	    }
	    else if (st.sval.compareTo("RevealMessage1")==0) {
	      if (st.nextToken()=='=' && st.nextToken()=='"') {	
		final String GotMessage=new String(st.sval);
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.RevealMessage[1]=GotMessage;
		}});
	      }
	    }
	    else if (st.sval.compareTo("SelectedInput")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {	
		final int number=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.externalsetbutton(number);
		}});
	      }
	    }
	    else if (st.sval.compareTo("SelectedInput?")==0) {
	      SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		panel.printselected();
	      }});
	    }
	    else if (st.sval.compareTo("JumpBack")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {	
		final int number=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  panel.setjumpback(number != 0);
		}});
	      }
	    }
	    else if (st.sval.compareTo("JumpBack?")==0) {
	      SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		panel.printjumpback();
	      }});
	    }
	    else if (st.sval.compareTo("BlindMode")==0) {
	      if (st.nextToken()=='=' && st.nextToken()==st.TT_NUMBER) {	
		final int number=(int)st.nval;
		SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		  if (number != 0) panel.State=STATE_BLIND;
		  else panel.State=STATE_VISIBLE;
		}});
	      }
	    }
	    else if (st.sval.compareTo("BlindRandomize")==0) {
	      
	      SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		Random R = new Random();
		/* boolean flag=R.nextBoolean(); (but nextBoolean requires JDK 1.2 and we want to be 1.1 compat) */
		boolean flag=(R.nextFloat() > 0.5);
		if (flag) {
		  panel.map[0]=1;
		  panel.map[1]=0;
		  panel.imap[0]=1;
		  panel.imap[1]=0;
		} else {
		  panel.map[0]=0;
		  panel.map[1]=1;
		  panel.imap[0]=0;
		  panel.imap[1]=1;
		}
	      }});
	      
	    }
	    else if (st.sval.compareTo("NoJumpBack")==0) {
	      /* jump-back is disabled -- turn off the checkbox */

	      SwingUtilities.invokeAndWait(new Runnable() {public void run() {
		panel.setjumpback(false);
		panel.JumpBackBox.setEnabled(false);
	      }});
	      
	    }
	    else if (st.sval.compareTo("BeginGUI")==0) {
	      SwingUtilities.invokeAndWait(new Runnable() {public void run() {		  
		if (panel.State==STATE_VISIBLE) {
		  panel.SetVisibleLabels();
		} else {
		  panel.SetBlindLabels();
		}
		/* Make sure that what is selected is consistent with 
		   our internal state */
		panel.InputButton[panel.internalselected].setSelected(true);
		panel.JumpBackBox.setSelected(panel.jumpback);
		

		frame = new JFrame(Title);
		frame.addWindowListener(new WindowAdapter() {
		  public void windowClosing(WindowEvent e) {System.exit(0);}
		});
		frame.getContentPane().add("Center", panel);
		frame.pack();
		frame.setVisible(true); 
		System.out.println("GUIReady");
	      }});
	    }
	    break;
	    
	  default:
	    System.err.println("Got unknown input token: " + ttype);
	    
	  }
	}
	System.exit(0);
      }
      catch(Exception except) {
	System.err.println("Error in input processing thread: "+except);
      }
    }
  }
  

  public static void main(String s[]) {
    ConsoleInput consoleInput;
    
    
    /* System.err.println("main started "); */
    
    try {
      //UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
      UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
      //UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
    } catch (Exception exc) {
      System.err.println("Error loading Look&Feel: " + exc);
    }
    
    
    panel = new syncswitch_panel();
    /* System.err.println("sio_panel created "); */

    /* We will want to remove these */

    /*
      frame = new JFrame(Title);
      frame.addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {System.exit(0);}
      });
      frame.getContentPane().add("Center", panel);
      frame.pack();
      frame.setVisible(true); 
      */

    /* start up the control-input thread */
    
    consoleInput = new ConsoleInput(panel);
    consoleInputThread=new Thread(consoleInput);
    consoleInputThread.start();
  }
}

