/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "io_spec.h"
#include "buffer.h"
#include "linklist.h"
#include "util.h"

int GotMagic=0;
double Volume=75.0; /* percent */

int InputPipe=0;
int OutputPipe=1;

int BlkSize=0;
int NumChannels=0;


struct SBuf *Buffer;
struct SBuf *OutBuf;

INT16 *DelayBuf=NULL;
int DBLen=0;
int DBNumChannels=0;


int main(int argc, char *argv[])
{
  int Cnt;

  int Ceil;
  int Floor;
  int LowMask;
  int CCnt;
  double ScaleFactor;

  int UseHighBits,UseLowBits;

  register int Data;
  int MaxVal;
  

  for (Cnt=1; argv[Cnt];Cnt++) {
    if (!strcmp(argv[Cnt],"-magic")) {
      GotMagic=1;
    }
    if (!strncmp(argv[Cnt],"-volume=",8)) {
      Volume=strtod(argv[Cnt]+8,NULL);
    }
  }

  if (!GotMagic) {
    fprintf(stderr,
	    "This program is meant to be run by Aglaophone\n"
	    "Pipes:  #0   input stream\n"
	    "        #1   output stream\n"
	    "        #2   error messages (do not connect)\n"
	    "\n"
	    "Usage:\n"
	    "   %s -magic -delay=%f\n"
	    "\n"
	    "This program multiplies the audio stream volume by the specified\n"
	    "percent\n"
	    ,argv[0],(double)Volume);
    exit(1);
  }
  

  Buffer=CreateSBuf();
  for (;;) {
    
    SBufPrepareHeader(Buffer);
    ReadSBuf(Buffer,InputPipe); 
    assert(SBufCheckHeader(Buffer));
    SBufReallocateIfNeeded(Buffer);
    SBufPrepareFinish(Buffer);
    
    ReadSBuf(Buffer,InputPipe); 
    
    
    
    /* copy out needed data */
    BlkSize=Buffer->Header->BlockSize;
    NumChannels=Buffer->Header->NumChannels;
    
    /* Process data here */
    for (Cnt=0;Cnt < BlkSize;Cnt++) {
      for (CCnt=0;CCnt < NumChannels;CCnt++) {
	SBufEntry(Buffer,Cnt,CCnt)*=Volume/100.0;
      }
    }
    
    SBufPrepare(Buffer);
    
    WriteSBuf(Buffer,OutputPipe);

    
  }
}
