/*
     Aglaophone -- a system for real time audio processing
     Copyright (C) 1998-1999 Stephen D. Holland
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; version 2 of the License.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "linklist.h"
#include "stringstack.h"

  struct Module {
    struct Node Node;
    char *Name; /* allocated separately */
    char *Command; /* allocated separately */
    struct Stack *ArgList; /* arguments are placed on this such that the 
			      first argument is on the top of the stack */
  };
  struct FileSpec {
    struct Node Node;
    char *Name; /* allocated separately */
    char *FileName; /* allocated separately */
    int Mode; 
    int fd; /* set later */
  };
  struct Connection {
    struct Node Node;
    char *InputName; /* allocated separately */
    int InputPort;
    int PipeRead;
    struct FileSpec *InpFile,*OutFile; /* One of these is set if the connection
					  is to a file, and there is no pipe */

    char *OutputName; /* allocated separately*/
    int OutputPort;
    int PipeWrite;
  };

  FILE *Config;

#define GET_INPUT PCB.input_code=fgetc(Config);

  struct List ModuleList;
  struct List ConnectionList;
  struct List FileList;

  struct Stack *Stk;


  struct Module *Mod;
  struct FileSpec *File;

  int MaxPort=2;


  char **gargv; /* global copy of argv */
  int gargc;


/*

 AnaGram Parsing Engine
 Copyright (c) 1993, 1996, Parsifal Software.
 All Rights Reserved.

*/



#ifndef VPP_H
#include "vpp.h"
#endif

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define RULE_CONTEXT (&((PCB).cs[(PCB).ssx]))
#define ERROR_CONTEXT ((PCB).cs[(PCB).error_frame_ssx])
#define CONTEXT ((PCB).cs[(PCB).ssx])



vpp_pcb_type vpp_pcb;
#define PCB vpp_pcb

#line 258 "/home1/sdh4/aglaophone/vpp/vpp.syn"
int main(int argc, char *argv[])
{
  int Port;
  int ProcCount=0;
  int pid;
  struct Module *Module,*InpMod,*OutMod;
  struct FileSpec *File;
  struct Connection *Conn;
  int NumArgs;
  char **newargv;
  int Cnt;
  char *Arg;
  int pipefds[2];


  gargv=argv;
  gargc=argc;
  
  if (!argv[1]) {
    printf("Usage: %s <configfile>\n",argv[0]);
    exit(1);
  }
  
  Config=fopen(argv[1],"r");

  if (!Config) {
    fprintf(stderr,"%s: Unable to open configuration file %s\n",argv[0],argv[1]);
    exit(1);
  }

  Stk=InitStack();
  NewList(&ModuleList);
  NewList(&ConnectionList);
  NewList(&FileList);

  vpp(); /* call the parser */
  if (PCB.exit_flag != AG_SUCCESS_CODE) {
    fprintf(stderr,"%s: Failed due to parse error in configuration file.\n",argv[0]);
    exit(1);
  }

  /* Make sure there aren't any connections to undefined modules */
  for (Conn=(struct Connection *)ConnectionList.lh_Head;Conn->Node.ln_Succ;Conn=(struct Connection *)Conn->Node.ln_Succ) {
    InpMod=NULL;
    OutMod=NULL;
    Conn->InpFile=NULL;
    Conn->OutFile=NULL;

    /* Select default ports if they have not been set.
       Default to port 0 for the input side of a connection
       Default to part 1 for the output side of a connection */
    
    if (Conn->InputPort < 0) Conn->InputPort=0;
    if (Conn->OutputPort < 0) Conn->OutputPort=1;
    

    for (Module=(struct Module *)ModuleList.lh_Head;Module->Node.ln_Succ;Module=(struct Module *)Module->Node.ln_Succ) {
      if (!strcmp(Module->Name,Conn->InputName)) {
	InpMod=Module;
      }
      if (!strcmp(Module->Name,Conn->OutputName)) {
	OutMod=Module;
      }
    }
    
    for (File=(struct FileSpec *)FileList.lh_Head;File->Node.ln_Succ;File=(struct FileSpec *)File->Node.ln_Succ) {
      if (!strcmp(File->Name,Conn->InputName) && (File->Mode & O_WRONLY)) {
	Conn->InpFile=File;
      }
      if (!strcmp(File->Name,Conn->OutputName) && (File->Mode==O_RDONLY || (File->Mode & O_RDONLY))) {
	Conn->OutFile=File;
      }
      
    }
    
    if (Conn->InpFile && Conn->OutFile) {
      fprintf(stderr,"%s: Can't perform file copy of %s to %s\n",argv[0],Conn->OutputName,Conn->InputName);
    }
    
    if (!InpMod && !Conn->InpFile) {
      fprintf(stderr,"%s: Couldn't find input module/file %s\n",argv[0],Conn->InputName);
      exit(1);
    }
    if (!OutMod && !Conn->OutFile) {
      fprintf(stderr,"%s: Couldn't find output module/file %s\n",argv[0],Conn->OutputName);
      exit(1);
    }
  }
  

  /* Avoid port numbering conflicts -- open ports up to the max port to
     /dev/null */
  do {
    Port=open("/dev/null",O_WRONLY);
  } while (Port <= MaxPort);

  /* Create pipes... */
  for (Conn=(struct Connection *)ConnectionList.lh_Head;Conn->Node.ln_Succ;Conn=(struct Connection *)Conn->Node.ln_Succ) {

    if (Conn->InpFile==NULL && Conn->OutFile==NULL) {
      /* we only create the pipe if we are connecting module->module. 
	 module->file or file->module doesn't require a pipe */
      pipe(pipefds);
      
      Conn->PipeRead=pipefds[0];
      Conn->PipeWrite=pipefds[1];

    }
    else {
      /* for a file->module or module->file connection, we need to 
	 open the file */
      
      if (Conn->InpFile) {
	assert(!Conn->OutFile);
	File=Conn->InpFile;
	if (!(File->Mode & O_WRONLY)) {
	  fprintf(stderr,"%s: Attempting to write to read-only file %s\n",argv[0],File->Name);
	  exit(1);
	}
      }

      if (Conn->OutFile) {
	assert(!Conn->InpFile);
	File=Conn->OutFile;
	if ((File->Mode & O_WRONLY)) {
	  fprintf(stderr,"%s: Attempting to read from write-only file %s\n",argv[0],File->Name);
	  exit(1);
	}
	
      }
      if (File->fd >= 0) {
	fprintf(stderr,"%s: File %s connected more than once\n",argv[0],File->Name);
	exit(1);
      }
      File->fd=open(File->FileName,File->Mode,0666);
      if (File->fd < 0) {
	fprintf(stderr,"%s: Error opening file %s\n",argv[0],File->FileName);
	exit(1);
      }
    }
  }

  /* fork off processes */
  for (Module=(struct Module *)ModuleList.lh_Head;Module->Node.ln_Succ;Module=(struct Module *)Module->Node.ln_Succ) {
    pid=fork();
    if (pid==0) {
      /* child process */
      
      /* hook up pipes */
      for (Conn=(struct Connection *)ConnectionList.lh_Head;Conn->Node.ln_Succ;Conn=(struct Connection *)Conn->Node.ln_Succ) {

	if (!Conn->InpFile && !Conn->OutFile) {
	  /* For a module->module connection, use the
	     pipes we have opened */
	  assert(Conn->PipeRead >= 0);
	  assert(Conn->PipeWrite >= 0);
	  
	  if (!strcmp(Module->Name,Conn->InputName)) {
	    dup2(Conn->PipeRead,Conn->InputPort);
	  } else {
	    close(Conn->PipeRead);
	  }
	  
	  if (!strcmp(Module->Name,Conn->OutputName)) {
	    dup2(Conn->PipeWrite,Conn->OutputPort);
	  } else {
	    close(Conn->PipeWrite);
	  }
	} 
	if (Conn->InpFile != NULL) {
	  /* For a file<-module connection,
	     just dup2 the file descriptor */
	  if (!strcmp(Module->Name,Conn->OutputName)) {
	    dup2(Conn->InpFile->fd,Conn->OutputPort);
	  }
	  
	}
	if (Conn->OutFile != NULL) {
	  /* For a file->module connection,
	     just dup2 the file descriptor */
	  if (!strcmp(Module->Name,Conn->InputName)) {
	    dup2(Conn->OutFile->fd,Conn->InputPort);
	  }
	  
	}


	
      }
      
      /* exec the program */
      NumArgs=CountStack(Module->ArgList);
      newargv=malloc(sizeof(char *)*(NumArgs+2));
      newargv[0]=strdup(Module->Command);
      for (Cnt=1;Cnt <= NumArgs;Cnt++) {
	Arg=PopString(Module->ArgList);
	newargv[Cnt]=strdup(Arg);
      }
      newargv[NumArgs+1]=NULL;

      execvp(Module->Command,newargv);
      fprintf(stderr,"%s: Error execing %s.\n",argv[0],Module->Command);
      exit(1);
    } else if (pid < 0) {
      /* error */
      fprintf(stderr,"%s: Error forking process.\n",argv[0]);
      exit(1);
    } else {
      ProcCount++;
    }
    
  }


  /* close our copy of all pipes */
  for (Conn=(struct Connection *)ConnectionList.lh_Head;Conn->Node.ln_Succ;Conn=(struct Connection *)Conn->Node.ln_Succ) {
    close(Conn->PipeRead);
    close(Conn->PipeWrite);
  }
  
  
  /* wait for all subprocesses to finish */
  while (ProcCount) {
    wait(NULL);
    ProcCount--;
  }
  exit(0);
  return 0;
}



#line 341 "vpp.c"
#ifndef CONVERT_CASE
#define CONVERT_CASE(c) (c)
#endif
#ifndef TAB_SPACING
#define TAB_SPACING 8
#endif

static void ag_rp_1(void) {
#line 116 "/home1/sdh4/aglaophone/vpp/vpp.syn"
File=calloc(sizeof(struct FileSpec),1);File->Name=strdup(PopString(Stk));File->Mode=O_RDONLY;File->fd=-1;
#line 352 "vpp.c"
}

static void ag_rp_2(void) {
#line 117 "/home1/sdh4/aglaophone/vpp/vpp.syn"
    
    if (!File->FileName) { 
      fprintf(stderr,"%s: No filename given for file %s.\n",gargv[0],Mod->Name);
      exit(1);
    }

    AddTail(&FileList,(struct Node *)File);
    
  
#line 366 "vpp.c"
}

#define ag_rp_3() (File->FileName=strdup(PopString(Stk)))

#define ag_rp_4(m) (File->Mode=m)

#define ag_rp_5() (O_RDONLY)

#define ag_rp_6() (O_WRONLY|O_CREAT|O_TRUNC)

#define ag_rp_7() (O_WRONLY|O_APPEND)

static void ag_rp_8(void) {
#line 142 "/home1/sdh4/aglaophone/vpp/vpp.syn"
Mod=calloc(sizeof(struct Module),1);Mod->Name=strdup(PopString(Stk));
#line 382 "vpp.c"
}

static void ag_rp_9(void) {
#line 143 "/home1/sdh4/aglaophone/vpp/vpp.syn"
    char *Str;
    Mod->ArgList=InitStack();
    
    while (Str=PopString(Stk)) {
      PushString(Mod->ArgList,Str);
    }
    
    if (!Mod->Command) { 
      fprintf(stderr,"%s: No command given for module %s.\n",gargv[0],Mod->Name);
    }
    
    AddTail(&ModuleList,(struct Node *)Mod);
  
#line 400 "vpp.c"
}

#define ag_rp_10() (Mod->Command=strdup(PopString(Stk)))

static void ag_rp_12(void) {
#line 172 "/home1/sdh4/aglaophone/vpp/vpp.syn"
char *String;String=PopString(Stk);AddStr(Stk,String);
#line 408 "vpp.c"
}

static void ag_rp_13(int i) {
#line 175 "/home1/sdh4/aglaophone/vpp/vpp.syn"
InitString(Stk);if (i < 0) i=0;if (i >= (gargc-1)) i=gargc-2;AddStr(Stk,gargv[i+1]);
#line 414 "vpp.c"
}

#define ag_rp_14(d) (d-'0')

#define ag_rp_15(i, d) (i*10+d-'0')

static void ag_rp_16(int OPort, int IPort) {
#line 192 "/home1/sdh4/aglaophone/vpp/vpp.syn"
     /* Input on right, output on left */
    
    struct Connection *Conn;

    Conn=malloc(sizeof(struct Connection));

    Conn->InputName=strdup(PopString(Stk));
    Conn->InputPort=IPort;    
    if (IPort > MaxPort) MaxPort=IPort;


    Conn->OutputName=strdup(PopString(Stk));
    Conn->OutputPort=OPort;
    if (OPort > MaxPort) MaxPort=OPort;


    Conn->PipeRead=-1;
    Conn->PipeWrite=-1;
    
    AddTail(&ConnectionList,(struct Node *)Conn);
    

  
#line 446 "vpp.c"
}

static void ag_rp_17(int IPort, int OPort) {
#line 217 "/home1/sdh4/aglaophone/vpp/vpp.syn"
     /* Input on left, output on right */
    struct Connection *Conn;

    Conn=malloc(sizeof(struct Connection));
    Conn->OutputName=strdup(PopString(Stk));
    Conn->OutputPort=OPort;
    if (OPort > MaxPort) MaxPort=OPort;


    Conn->InputName=strdup(PopString(Stk));
    Conn->InputPort=IPort;    
    if (IPort > MaxPort) MaxPort=IPort;

    Conn->PipeRead=-1;
    Conn->PipeWrite=-1;
    
    AddTail(&ConnectionList,(struct Node *)Conn);
  
#line 469 "vpp.c"
}

#define ag_rp_18() (-1)

#define ag_rp_19(s) (s)

static void ag_rp_20(int c) {
#line 242 "/home1/sdh4/aglaophone/vpp/vpp.syn"
InitString(Stk);AddChar(Stk,c);
#line 479 "vpp.c"
}

#define ag_rp_21(c) (AddChar(Stk,c))

#define ag_rp_22() (InitString(Stk))

#define ag_rp_23(c) (AddChar(Stk,c))

#define ag_rp_24(d) (d-'0')

#define ag_rp_25(u, d) (u*10+d-'0')


#define READ_COUNTS 
#define WRITE_COUNTS 
static vpp_vs_type ag_null_value;
#define V(i,t) (*(t *) (&(PCB).vs[(PCB).ssx + i]))
#define VS(i) (PCB).vs[(PCB).ssx + i]

#ifndef GET_CONTEXT
#define GET_CONTEXT CONTEXT = (PCB).input_context
#endif

typedef enum {
  ag_action_1,
  ag_action_2,
  ag_action_3,
  ag_action_4,
  ag_action_5,
  ag_action_6,
  ag_action_7,
  ag_action_8,
  ag_action_9,
  ag_action_10,
  ag_action_11,
  ag_action_12
} ag_parser_action;

static int ag_ap;



static const unsigned char ag_rpx[] = {
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  1,  2,  0,  0,  3,  4,  5,  6,  7,  8,  9,  0,  0, 10,
   10,  0,  0,  0,  0, 11, 12, 13, 14,  0,  0,  0, 15, 16, 17, 18, 19, 20,
    0, 21, 22, 23, 24
};

static const unsigned char ag_key_itt[] = {
 0
};

static const unsigned short ag_key_pt[] = {
0
};

static const unsigned char ag_key_ch[] = {
    0, 47, 99,102,109,255, 42,255,117,255,103,255,112,114,255,109,255,109,
  110,255,111,255,110,255,101,255,108,255,105,255,101,117,255,100,255,111,
  255, 45, 47, 60, 97, 99,102,109,111,114,255, 99,102,109,255, 47,255, 47,
  102,109,255,117,255,103,255,114,255,109,255,109,255,111,255, 47, 97, 99,
  255, 45, 47, 60,255,102,109,255, 45, 60,255,117,255,103,255,114,255,109,
  255,109,255,111,255, 97, 99,255, 47, 97,111,114,255, 97,111,114,255,117,
  255,103,255,114,255,109,255,109,255,111,255, 97, 99,102,109,255,117,255,
  103,255,114,255,109,255,109,255,111,255, 47, 97, 99,102,109,255
};

static const unsigned char ag_key_act[] = {
  0,3,3,3,3,4,3,4,3,4,1,4,3,2,4,3,4,1,3,4,2,4,3,4,1,4,2,4,2,4,0,3,4,2,4,
  2,4,3,3,3,2,2,2,2,3,3,4,3,3,3,4,3,4,3,3,3,4,3,4,1,4,2,4,3,4,1,4,2,4,3,
  2,2,4,3,3,3,4,3,3,4,3,3,4,3,4,1,4,2,4,3,4,1,4,2,4,2,2,4,3,3,3,3,4,3,3,
  3,4,3,4,1,4,2,4,3,4,1,4,2,4,2,2,3,3,4,3,4,1,4,2,4,3,4,1,4,2,4,3,2,2,3,
  3,4
};

static const unsigned char ag_key_parm[] = {
    0,  1, 51, 24, 38,  0,  2,  0, 44,  0, 45,  0, 37,  0,  0, 42,  0, 43,
   51,  0,  0,  0, 31,  0, 24,  0,  0,  0,  0,  0, 33, 38,  0,  0,  0,  0,
    0,  3,  1,  4,  0,  0,  0,  0, 36, 35,  0, 51, 24, 38,  0,  1,  0,  1,
   31, 33,  0, 44,  0, 45,  0,  0,  0, 42,  0, 43,  0,  0,  0,  1,  0,  0,
    0,  3,  1,  4,  0, 31, 33,  0,  3,  4,  0, 44,  0, 45,  0,  0,  0, 42,
    0, 43,  0,  0,  0,  0,  0,  0,  1, 37, 36, 35,  0, 37, 36, 35,  0, 44,
    0, 45,  0,  0,  0, 42,  0, 43,  0,  0,  0,  0,  0, 31, 33,  0, 44,  0,
   45,  0,  0,  0, 42,  0, 43,  0,  0,  0,  1,  0,  0, 31, 33,  0
};

static const unsigned char ag_key_jmp[] = {
    0,  0,  2, 13, 17,  0, 23,  0, 36,  0,  8,  0, 31, 10,  0, 41,  0, 15,
   45,  0, 17,  0, 54,  0, 22,  0, 24,  0, 26,  0,  0, 58,  0, 30,  0, 33,
    0, 25, 27, 29, 12, 20, 28, 35, 61, 70,  0, 74, 85, 89,  0, 95,  0, 97,
   99,107,  0,113,  0, 57,  0, 59,  0,118,  0, 63,  0, 65,  0,111, 61, 67,
    0,122,124,126,  0,128,136,  0,140,142,  0,144,  0, 83,  0, 85,  0,149,
    0, 89,  0, 91,  0, 87, 93,  0,153,155,161,170,  0,174,180,189,  0,193,
    0,107,  0,109,  0,198,  0,113,  0,115,  0,111,117,202,210,  0,216,  0,
  124,  0,126,  0,221,  0,130,  0,132,  0,214,128,134,225,233,  0
};

static const unsigned char ag_key_index[] = {
    1,  0,  6, 37, 47,  0,  0,  6,  6, 51, 51, 51, 47,  0,  0,  0, 51, 51,
   51,  0,  0,  0, 53,  0, 69, 53, 73, 73,  1, 69, 77,  0, 80, 95, 77, 73,
   51, 51, 95, 98, 51,  1,  0,  0, 51, 51, 51, 51,  1,103,  0, 51, 51,  0,
    0,  0,  0, 53,  0,  0,119, 51, 51,136,  0,136
};

static const unsigned char ag_key_ends[] = {
42,0, 111,110,110,101,99,116,105,111,110,115,0, 105,108,101,0, 
111,100,117,108,101,0, 47,0, 62,0, 42,0, 45,0, 112,101,110,100,0, 
109,101,110,116,0, 97,110,100,0, 110,101,99,116,105,111,110,115,0, 
97,109,101,0, 108,101,0, 118,101,114,119,114,105,116,101,0, 
101,97,100,0, 111,110,110,101,99,116,105,111,110,115,0, 105,108,101,0, 
111,100,117,108,101,0, 42,0, 42,0, 105,108,101,110,97,109,101,0, 
111,100,101,0, 42,0, 109,101,110,116,0, 97,110,100,0, 62,0, 
42,0, 45,0, 105,108,101,110,97,109,101,0, 111,100,101,0, 62,0, 
45,0, 109,101,110,116,0, 97,110,100,0, 42,0, 112,112,101,110,100,0, 
118,101,114,119,114,105,116,101,0, 101,97,100,0, 112,112,101,110,100,0, 
118,101,114,119,114,105,116,101,0, 101,97,100,0, 109,101,110,116,0, 
97,110,100,0, 105,108,101,110,97,109,101,0, 111,100,101,0, 42,0, 
109,101,110,116,0, 97,110,100,0, 105,108,101,110,97,109,101,0, 
111,100,101,0, 
};
#define AG_TCV(x) (((int)(x) >= -1 && (int)(x) <= 255) ? ag_tcv[(x) + 1] : 0)

static const unsigned char ag_tcv[] = {
   10, 63, 63, 63, 63, 63, 63, 63, 63, 63, 62, 19, 63, 63, 62, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 62, 63, 59,
   15, 48, 63, 63, 63, 63, 63, 63, 63, 63, 63, 55, 63, 50, 50, 50, 50, 50,
   50, 50, 50, 50, 50, 63, 63, 63, 64, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 26, 63,
   29, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63,
   63, 63, 63, 63, 63
};

#ifndef SYNTAX_ERROR
#define SYNTAX_ERROR fprintf(stderr,"%s, line %d, column %d\n", \
  (PCB).error_message, (PCB).line, (PCB).column)
#endif

#ifndef FIRST_LINE
#define FIRST_LINE 1
#endif

#ifndef FIRST_COLUMN
#define FIRST_COLUMN 1
#endif


#ifndef PARSER_STACK_OVERFLOW
#define PARSER_STACK_OVERFLOW {fprintf(stderr, \
   "\nParser stack overflow, line %d, column %d\n",\
   (PCB).line, (PCB).column);}
#endif

#ifndef REDUCTION_TOKEN_ERROR
#define REDUCTION_TOKEN_ERROR {fprintf(stderr, \
    "\nReduction token error, line %d, column %d\n", \
    (PCB).line, (PCB).column);}
#endif


typedef enum
  {ag_accept_key, ag_set_key, ag_jmp_key, ag_end_key, ag_no_match_key,
   ag_cf_accept_key, ag_cf_set_key, ag_cf_end_key} key_words;

#ifndef GET_INPUT
#define GET_INPUT ((PCB).input_code = getchar())
#endif


static unsigned ag_look_ahead(void) {
  if ((PCB).rx < (PCB).fx) {
    return (unsigned) CONVERT_CASE((PCB).lab[(PCB).rx++]);
  }
  GET_INPUT;
  (PCB).fx++;
  return (unsigned) CONVERT_CASE((PCB).lab[(PCB).rx++] = (PCB).input_code);
}

static void ag_get_key_word(int ag_k) {
  int save_index = (PCB).rx;
  const  unsigned char *sp;
  unsigned ag_ch;
  while (1) {
    switch (ag_key_act[ag_k]) {
    case ag_cf_end_key:
      sp = ag_key_ends + ag_key_jmp[ag_k];
      do {
        if ((ag_ch = *sp++) == 0) {
          int ag_k1 = ag_key_parm[ag_k];
          int ag_k2 = ag_key_pt[ag_k1];
          if (ag_key_itt[ag_k2 + ag_look_ahead()]) goto ag_fail;
          (PCB).rx--;
          (PCB).token_number = (vpp_token_type) ag_key_pt[ag_k1 + 1];
          return;
        }
      } while (ag_look_ahead() == ag_ch);
      goto ag_fail;
    case ag_end_key:
      sp = ag_key_ends + ag_key_jmp[ag_k];
      do {
        if ((ag_ch = *sp++) == 0) {
          (PCB).token_number = (vpp_token_type) ag_key_parm[ag_k];
          return;
        }
      } while (ag_look_ahead() == ag_ch);
    case ag_no_match_key:
ag_fail:
      (PCB).rx = save_index;
      return;
    case ag_cf_set_key: {
      int ag_k1 = ag_key_parm[ag_k];
      int ag_k2 = ag_key_pt[ag_k1];
      ag_k = ag_key_jmp[ag_k];
      if (ag_key_itt[ag_k2 + (ag_ch = ag_look_ahead())]) break;
      save_index = --(PCB).rx;
      (PCB).token_number = (vpp_token_type) ag_key_pt[ag_k1+1];
      break;
    }
    case ag_set_key:
      save_index = (PCB).rx;
      (PCB).token_number = (vpp_token_type) ag_key_parm[ag_k];
    case ag_jmp_key:
      ag_k = ag_key_jmp[ag_k];
      ag_ch = ag_look_ahead();
      break;
    case ag_accept_key:
      (PCB).token_number =  (vpp_token_type) ag_key_parm[ag_k];
      return;
    case ag_cf_accept_key: {
      int ag_k1 = ag_key_parm[ag_k];
      int ag_k2 = ag_key_pt[ag_k1];
      if (ag_key_itt[ag_k2 + ag_look_ahead()]) (PCB).rx = save_index;
      else {
        (PCB).rx--;
        (PCB).token_number =  (vpp_token_type) ag_key_pt[ag_k1+1];
      }
      return;
    }
    }
    while(ag_key_ch[ag_k] < ag_ch) ag_k++;
    if (ag_key_ch[ag_k] != ag_ch) {
      (PCB).rx = save_index;
      return;
    }
  }
}


static void ag_track(void) {
  int ag_k = 0;
  while (ag_k < (PCB).rx) {
    int ag_ch = (PCB).lab[ag_k++];
    switch (ag_ch) {
    case '\n':
      (PCB).column = 1, (PCB).line++;
    case '\r':
    case '\f':
      break;
    case '\t':
      (PCB).column += (TAB_SPACING) - ((PCB).column - 1) % (TAB_SPACING);
      break;
    default:
      (PCB).column++;
    }
  }
  ag_k = 0;
  while ((PCB).rx < (PCB).fx) (PCB).lab[ag_k++] = (PCB).lab[(PCB).rx++];
  (PCB).fx = ag_k;
  (PCB).rx = 0;
}


static void ag_prot(void) {
  int ag_k = 34 - ++(PCB).btsx;
  if (ag_k <= (PCB).ssx) {
    (PCB).exit_flag = AG_STACK_ERROR_CODE;
    PARSER_STACK_OVERFLOW;
    return;
  }
  (PCB).bts[(PCB).btsx] = (PCB).sn;
  (PCB).bts[ag_k] = (PCB).ssx;
  (PCB).vs[ag_k] = (PCB).vs[(PCB).ssx];
  (PCB).ss[ag_k] = (PCB).ss[(PCB).ssx];
}

static void ag_undo(void) {
  if ((PCB).drt == -1) return;
  while ((PCB).btsx) {
    int ag_k = 34 - (PCB).btsx;
    (PCB).sn = (PCB).bts[(PCB).btsx--];
    (PCB).ssx = (PCB).bts[ag_k];
    (PCB).vs[(PCB).ssx] = (PCB).vs[ag_k];
    (PCB).ss[(PCB).ssx] = (PCB).ss[ag_k];
  }
  (PCB).token_number = (vpp_token_type) (PCB).drt;
  (PCB).ssx = (PCB).dssx;
  (PCB).sn = (PCB).dsn;
  (PCB).drt = -1;
}


static const unsigned char ag_tstt[] = {
62,19,15,1,0,5,6,7,8,
64,63,62,59,55,50,48,29,26,15,0,17,18,
64,63,62,59,55,50,48,29,26,19,15,0,13,14,
62,19,15,1,0,6,
51,38,24,0,9,20,21,22,23,
64,63,62,59,55,50,48,29,26,15,0,
19,0,
64,63,62,59,55,50,48,29,26,19,15,0,
2,0,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
51,38,24,10,0,20,21,22,23,
63,59,0,25,
26,0,
63,59,0,25,
63,62,59,50,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
63,62,59,50,19,15,1,0,6,7,8,
26,0,
0,52,
26,0,
0,27,
63,59,29,0,25,53,54,
0,39,
62,19,15,1,0,6,7,8,
63,59,55,50,0,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
0,28,
50,0,56,
4,3,0,
0,40,
33,31,29,0,30,
50,0,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
45,44,43,42,29,0,41,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
63,59,0,25,54,
63,59,0,25,54,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
62,19,15,1,0,6,7,8,
37,36,35,0,34,
59,48,0,32,46,47,
62,19,15,1,0,6,7,
62,19,15,1,0,6,7,
59,48,0,32,46,47,
59,48,0,32,46,47,
59,48,0,32,46,47,
59,48,0,32,46,47,
62,19,15,1,0,6,7,8,
50,0,49,
0,60,
48,0,47,
62,19,15,1,0,6,
62,19,15,1,0,6,
62,50,19,15,1,0,6,7,8,
64,63,62,59,55,50,48,29,26,19,15,0,
62,19,15,1,0,6,7,8,

};


static unsigned const char ag_astt[425] = {
  1,1,1,1,8,0,1,1,1,1,1,1,1,1,1,1,1,1,1,8,1,1,1,1,1,1,1,1,1,1,1,1,1,8,1,1,9,
  9,1,1,5,3,1,1,1,7,1,1,1,1,1,9,9,9,9,9,9,9,9,9,9,5,3,7,9,9,9,9,9,9,9,9,9,9,
  9,5,3,7,1,1,1,1,8,1,1,1,1,1,1,1,8,1,1,1,1,1,1,1,8,1,1,1,1,1,1,3,7,3,3,3,3,
  2,2,7,1,1,7,2,2,7,1,10,1,10,10,1,1,1,8,1,1,1,1,1,1,1,8,1,1,1,10,1,10,10,1,
  1,1,8,1,1,1,1,7,8,1,1,7,4,1,2,2,1,7,1,3,1,4,1,1,1,1,1,8,1,1,1,10,10,1,10,4,
  1,1,1,1,8,1,1,1,1,1,1,1,5,1,1,3,1,1,1,1,8,1,1,1,8,1,2,7,1,1,1,7,8,1,1,1,1,
  7,3,10,4,1,1,1,1,8,1,1,1,1,1,1,1,8,1,1,1,1,1,1,1,1,7,3,1,1,1,1,8,1,1,1,1,1,
  1,1,8,1,1,1,1,1,1,1,5,1,1,2,2,2,7,1,1,2,2,7,1,1,1,1,1,1,8,1,1,1,1,1,1,1,8,
  1,1,1,1,1,1,1,8,1,1,1,1,1,1,1,8,1,1,1,1,1,1,1,5,1,1,2,2,2,2,7,1,1,1,7,2,1,
  2,1,1,1,1,7,1,1,1,1,1,1,7,1,1,1,1,7,3,1,3,1,1,7,3,1,3,1,1,7,2,1,2,1,1,7,2,
  1,2,1,1,1,1,5,1,1,2,2,7,1,4,1,1,5,2,9,9,1,1,4,3,9,9,1,1,4,3,1,10,1,1,1,5,1,
  1,2,10,10,10,1,10,10,10,10,10,10,10,7,1,1,1,1,5,1,1,3
};


static const unsigned char ag_pstt[] = {
3,3,1,2,4,0,3,3,4,
5,5,5,5,5,5,5,5,5,5,6,5,6,
7,7,7,7,7,7,7,7,7,7,7,8,7,8,
2,2,1,2,4,2,
10,11,9,4,12,12,12,12,12,
13,13,13,13,13,13,13,13,13,13,15,
16,6,
8,8,8,8,8,8,8,8,8,8,8,10,
11,8,
3,3,1,2,13,3,3,13,
3,3,1,2,14,3,3,14,
3,3,1,2,15,3,3,15,
10,11,9,5,12,18,18,18,18,
52,52,13,16,
17,14,
52,52,15,18,
53,3,53,53,3,1,2,19,3,3,19,
3,3,1,2,20,3,3,20,
53,3,53,53,3,1,2,21,3,3,21,
22,19,
23,23,
24,21,
22,25,
52,52,28,23,26,47,27,
31,29,
3,3,1,2,30,3,3,30,
53,53,31,53,50,
3,3,1,2,32,3,3,32,
3,3,1,2,3,3,3,45,
3,3,1,2,33,3,3,33,
34,34,
57,31,35,
36,37,32,
38,38,
39,40,41,34,25,
58,51,
3,3,1,2,42,3,3,42,
3,3,1,2,43,3,3,43,
44,45,46,47,48,38,34,
3,3,1,2,49,3,3,49,
3,3,1,2,50,3,3,50,
3,3,1,2,3,3,3,23,
52,52,42,26,51,
52,52,43,26,52,
3,3,1,2,53,3,3,53,
3,3,1,2,54,3,3,54,
3,3,1,2,55,3,3,55,
3,3,1,2,56,3,3,56,
3,3,1,2,3,3,3,32,
30,29,28,49,57,
59,58,50,26,60,26,
61,61,1,2,51,61,61,
62,62,1,2,52,62,62,
59,58,53,38,60,38,
59,58,54,37,60,37,
59,58,55,36,60,36,
59,58,56,35,60,35,
3,3,1,2,3,3,3,27,
43,58,63,
55,64,
58,39,41,
2,2,1,2,49,2,
2,2,1,2,48,2,
3,44,3,1,2,3,3,3,42,
56,56,56,65,56,56,56,56,56,56,56,64,
3,3,1,2,3,3,3,54,

};


static const unsigned short ag_sbt[] = {
     0,   9,  22,  36,  42,  51,  62,  64,  76,  78,  86,  94, 102, 111,
   115, 117, 121, 132, 140, 151, 153, 155, 157, 159, 166, 168, 176, 181,
   189, 197, 205, 207, 210, 213, 215, 220, 222, 230, 238, 245, 253, 261,
   269, 274, 279, 287, 295, 303, 311, 319, 324, 330, 337, 344, 350, 356,
   362, 368, 376, 379, 381, 384, 390, 396, 405, 417, 425
};


static const unsigned short ag_sbe[] = {
     4,  19,  33,  40,  45,  61,  63,  75,  77,  82,  90,  98, 106, 113,
   116, 119, 128, 136, 147, 152, 153, 156, 157, 162, 166, 172, 180, 185,
   193, 201, 205, 208, 212, 213, 218, 221, 226, 234, 243, 249, 257, 265,
   271, 276, 283, 291, 299, 307, 315, 322, 326, 334, 341, 346, 352, 358,
   364, 372, 377, 379, 382, 388, 394, 401, 416, 421, 425
};


static const unsigned char ag_fl[] = {
  1,1,2,0,1,3,1,1,2,0,1,3,1,2,0,1,3,1,2,1,1,1,0,10,0,2,3,4,1,1,1,0,10,0,
  2,3,3,3,3,1,1,2,3,1,2,7,0,2,6,6,1,3,1,2,4,0,2,1,2
};

static const unsigned char ag_ptt[] = {
    0,  7,  7,  8,  8,  5,  6, 13, 13, 14, 14,  6, 17, 17, 18, 18,  6,  9,
    9, 20, 20, 20, 27, 23, 28, 28, 30, 30, 34, 34, 34, 39, 21, 40, 40, 41,
   41, 41, 41, 32, 32, 32, 47, 49, 49, 22, 52, 52, 53, 53, 54, 54, 25, 25,
   46, 60, 60, 56, 56
};




static void ag_ra(void)
{
  switch(ag_rpx[ag_ap]) {
  case   1: ag_rp_1(); break;
  case   2: ag_rp_2(); break;
  case   3: ag_rp_3(); break;
  case   4: ag_rp_4(V(2,int)); break;
  case   5: V(0,int) = ag_rp_5(); break;
  case   6: V(0,int) = ag_rp_6(); break;
  case   7: V(0,int) = ag_rp_7(); break;
  case   8: ag_rp_8(); break;
  case   9: ag_rp_9(); break;
  case  10: ag_rp_10(); break;
  case  11: ag_rp_12(); break;
  case  12: ag_rp_13(V(1,int)); break;
  case  13: V(0,int) = ag_rp_14(V(0,int)); break;
  case  14: V(0,int) = ag_rp_15(V(0,int), V(1,int)); break;
  case  15: ag_rp_16(V(0,int), V(4,int)); break;
  case  16: ag_rp_17(V(0,int), V(4,int)); break;
  case  17: V(0,int) = ag_rp_18(); break;
  case  18: V(0,int) = ag_rp_19(V(2,unsigned)); break;
  case  19: ag_rp_20(V(0,int)); break;
  case  20: ag_rp_21(V(1,int)); break;
  case  21: ag_rp_22(); break;
  case  22: ag_rp_23(V(1,int)); break;
  case  23: V(0,unsigned) = ag_rp_24(V(0,int)); break;
  case  24: V(0,unsigned) = ag_rp_25(V(0,unsigned), V(1,int)); break;
  }
}

#define TOKEN_NAMES vpp_token_names
const char *const vpp_token_names[65] = {
  "ConfigFile",
  "",
  "",
  "",
  "",
  "ConfigFile",
  "ws",
  "",
  "",
  "Sections",
  "EOF",
  "whitespace",
  "anythingbuteof",
  "",
  "",
  "",
  "anythingbuteol",
  "",
  "",
  "",
  "Section",
  "ModuleSection",
  "ConnectionSection",
  "FileSection",
  "",
  "String",
  "",
  "",
  "FileData",
  "",
  "FileDatum",
  "",
  "argument",
  "",
  "modespec",
  "",
  "",
  "",
  "",
  "",
  "ModuleData",
  "ModuleDatum",
  "",
  "",
  "",
  "",
  "QuotedString",
  "CommandLineOption",
  "",
  "unsigned integer",
  "Digit",
  "",
  "ConnectionData",
  "ConnectionDatum",
  "ConnectionSide",
  "",
  "Unsigned",
  "StringCharNonDigit",
  "StringChar",
  "",
  "DequotedString",
  "QuotedStringChar",
  "",
  "",
  "",

};

static char ag_msg[82];

static void ag_diagnose(void) {
  int ag_snd = (PCB).sn, ag_k;

  ag_k = ag_sbt[ag_snd];
  if (*TOKEN_NAMES[ag_tstt[ag_k]] && ag_astt[ag_k + 1] == ag_action_8) {
    sprintf(ag_msg, "Missing %s", TOKEN_NAMES[ag_tstt[ag_k]]);
  }
  else if ((PCB).token_number && *TOKEN_NAMES[(PCB).token_number]) {
    sprintf(ag_msg, "Unexpected %s", TOKEN_NAMES[(PCB).token_number]);
  }
  else if (isprint((*(PCB).lab)) && (*(PCB).lab) != '\\') {
    sprintf(ag_msg, "Unexpected \'%c\'", (char) (*(PCB).lab));
  }
  else sprintf(ag_msg, "Unexpected input");
  (PCB).error_message = ag_msg;


}
static int ag_action_1_r_proc(void);
static int ag_action_2_r_proc(void);
static int ag_action_3_r_proc(void);
static int ag_action_4_r_proc(void);
static int ag_action_1_s_proc(void);
static int ag_action_3_s_proc(void);
static int ag_action_1_proc(void);
static int ag_action_2_proc(void);
static int ag_action_3_proc(void);
static int ag_action_4_proc(void);
static int ag_action_5_proc(void);
static int ag_action_6_proc(void);
static int ag_action_7_proc(void);
static int ag_action_8_proc(void);
static int ag_action_9_proc(void);
static int ag_action_10_proc(void);
static int ag_action_11_proc(void);
static int ag_action_8_proc(void);


static int (*const  ag_r_procs_scan[])(void) = {
  ag_action_1_r_proc,
  ag_action_2_r_proc,
  ag_action_3_r_proc,
  ag_action_4_r_proc
};

static int (*const  ag_s_procs_scan[])(void) = {
  ag_action_1_s_proc,
  ag_action_2_r_proc,
  ag_action_3_s_proc,
  ag_action_4_r_proc
};

static int (*const  ag_gt_procs_scan[])(void) = {
  ag_action_1_proc,
  ag_action_2_proc,
  ag_action_3_proc,
  ag_action_4_proc,
  ag_action_5_proc,
  ag_action_6_proc,
  ag_action_7_proc,
  ag_action_8_proc,
  ag_action_9_proc,
  ag_action_10_proc,
  ag_action_11_proc,
  ag_action_8_proc
};


static int ag_action_10_proc(void) {
  int ag_t = (PCB).token_number;
  (PCB).btsx = 0, (PCB).drt = -1;
  do {
    ag_track();
    if ((PCB).rx < (PCB).fx) {
      (PCB).input_code = (PCB).lab[(PCB).rx++];
      (PCB).token_number = (vpp_token_type) AG_TCV((PCB).input_code);}
    else {
      GET_INPUT;
      (PCB).lab[(PCB).fx++] = (PCB).input_code;
      (PCB).token_number = (vpp_token_type) AG_TCV((PCB).input_code);
      (PCB).rx++;
    }
    if (ag_key_index[(PCB).sn]) {
      unsigned ag_k = ag_key_index[(PCB).sn];
      unsigned char ag_ch = (unsigned char) CONVERT_CASE((PCB).input_code);
      while (ag_key_ch[ag_k] < ag_ch) ag_k++;
      if (ag_key_ch[ag_k] == ag_ch) ag_get_key_word(ag_k);
    }
  } while ((PCB).token_number == (vpp_token_type) ag_t);
  (PCB).rx = 0;
  return 1;
}

static int ag_action_11_proc(void) {
  int ag_t = (PCB).token_number;

  (PCB).btsx = 0, (PCB).drt = -1;
  do {
    (*(int *) &(PCB).vs[(PCB).ssx]) = *(PCB).lab;
    (PCB).ssx--;
    ag_track();
    ag_ra();
    if ((PCB).exit_flag != AG_RUNNING_CODE) return 0;
    (PCB).ssx++;
    if ((PCB).rx < (PCB).fx) {
      (PCB).input_code = (PCB).lab[(PCB).rx++];
      (PCB).token_number = (vpp_token_type) AG_TCV((PCB).input_code);}
    else {
      GET_INPUT;
      (PCB).lab[(PCB).fx++] = (PCB).input_code;
      (PCB).token_number = (vpp_token_type) AG_TCV((PCB).input_code);
      (PCB).rx++;
    }
    if (ag_key_index[(PCB).sn]) {
      unsigned ag_k = ag_key_index[(PCB).sn];
      unsigned char ag_ch = (unsigned char) CONVERT_CASE((PCB).input_code);
      while (ag_key_ch[ag_k] < ag_ch) ag_k++;
      if (ag_key_ch[ag_k] == ag_ch) ag_get_key_word(ag_k);
    }
  }
  while ((PCB).token_number == (vpp_token_type) ag_t);
  (PCB).rx = 0;
  return 1;
}

static int ag_action_3_r_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  (PCB).btsx = 0, (PCB).drt = -1;
  (PCB).reduction_token = (vpp_token_type) ag_ptt[ag_ap];
  ag_ra();
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_3_s_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  (PCB).btsx = 0, (PCB).drt = -1;
  (PCB).reduction_token = (vpp_token_type) ag_ptt[ag_ap];
  ag_ra();
  return (PCB).exit_flag == AG_RUNNING_CODE;;
}

static int ag_action_4_r_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  (PCB).reduction_token = (vpp_token_type) ag_ptt[ag_ap];
  return 1;
}

static int ag_action_2_proc(void) {
  (PCB).btsx = 0, (PCB).drt = -1;
  if ((PCB).ssx >= 34) {
    (PCB).exit_flag = AG_STACK_ERROR_CODE;
    PARSER_STACK_OVERFLOW;
  }
  (*(int *) &(PCB).vs[(PCB).ssx]) = *(PCB).lab;
  (PCB).ss[(PCB).ssx] = (PCB).sn;
  (PCB).ssx++;
  (PCB).sn = ag_ap;
  ag_track();
  return 0;
}

static int ag_action_9_proc(void) {
  if((PCB).drt == -1) {
    (PCB).drt=(PCB).token_number;
    (PCB).dssx=(PCB).ssx;
    (PCB).dsn=(PCB).sn;
  }
  ag_prot();
  (PCB).vs[(PCB).ssx] = ag_null_value;
  (PCB).ss[(PCB).ssx] = (PCB).sn;
  (PCB).ssx++;
  (PCB).sn = ag_ap;
  (PCB).rx = 0;
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_2_r_proc(void) {
  (PCB).ssx++;
  (PCB).sn = ag_ap;
  return 0;
}

static int ag_action_7_proc(void) {
  --(PCB).ssx;
  (PCB).rx = 0;
  (PCB).exit_flag = AG_SUCCESS_CODE;
  return 0;
}

static int ag_action_1_proc(void) {
  ag_track();
  (PCB).exit_flag = AG_SUCCESS_CODE;
  return 0;
}

static int ag_action_1_r_proc(void) {
  (PCB).exit_flag = AG_SUCCESS_CODE;
  return 0;
}

static int ag_action_1_s_proc(void) {
  (PCB).exit_flag = AG_SUCCESS_CODE;
  return 0;
}

static int ag_action_4_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  (PCB).reduction_token = (vpp_token_type) ag_ptt[ag_ap];
  (PCB).btsx = 0, (PCB).drt = -1;
  (*(int *) &(PCB).vs[(PCB).ssx]) = *(PCB).lab;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  else (PCB).ss[(PCB).ssx] = (PCB).sn;
  ag_track();
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbe[(PCB).sn] + 1;
    unsigned ag_t2 = ag_sbt[(PCB).sn+1] - 1;
    do {
      unsigned ag_tx = (ag_t1 + ag_t2)/2;
      if (ag_tstt[ag_tx] < (const unsigned char)(PCB).reduction_token) ag_t1 = ag_tx + 1;
      else ag_t2 = ag_tx;
    } while (ag_t1 < ag_t2);
    ag_ap = ag_pstt[ag_t1];
    if ((ag_s_procs_scan[ag_astt[ag_t1]])() == 0) break;
  }
  return 0;
}

static int ag_action_3_proc(void) {
  int ag_sd = ag_fl[ag_ap] - 1;
  (PCB).btsx = 0, (PCB).drt = -1;
  (*(int *) &(PCB).vs[(PCB).ssx]) = *(PCB).lab;
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  else (PCB).ss[(PCB).ssx] = (PCB).sn;
  ag_track();
  (PCB).reduction_token = (vpp_token_type) ag_ptt[ag_ap];
  ag_ra();
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbe[(PCB).sn] + 1;
    unsigned ag_t2 = ag_sbt[(PCB).sn+1] - 1;
    do {
      unsigned ag_tx = (ag_t1 + ag_t2)/2;
      if (ag_tstt[ag_tx] < (const unsigned char)(PCB).reduction_token) ag_t1 = ag_tx + 1;
      else ag_t2 = ag_tx;
    } while (ag_t1 < ag_t2);
    ag_ap = ag_pstt[ag_t1];
    if ((ag_s_procs_scan[ag_astt[ag_t1]])() == 0) break;
  }
  return 0;
}

static int ag_action_8_proc(void) {
  ag_undo();
  (PCB).rx = 0;
  (PCB).exit_flag = AG_SYNTAX_ERROR_CODE;
  ag_diagnose();
  SYNTAX_ERROR;
  {(PCB).rx = 1; ag_track();}
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_5_proc(void) {
  int ag_sd = ag_fl[ag_ap];
  if((PCB).drt == -1) {
    (PCB).drt=(PCB).token_number;
    (PCB).dssx=(PCB).ssx;
    (PCB).dsn=(PCB).sn;
  }
  if (ag_sd) (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  else {
    ag_prot();
    (PCB).ss[(PCB).ssx] = (PCB).sn;
  }
  (PCB).rx = 0;
  (PCB).reduction_token = (vpp_token_type) ag_ptt[ag_ap];
  ag_ra();
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbe[(PCB).sn] + 1;
    unsigned ag_t2 = ag_sbt[(PCB).sn+1] - 1;
    do {
      unsigned ag_tx = (ag_t1 + ag_t2)/2;
      if (ag_tstt[ag_tx] < (const unsigned char)(PCB).reduction_token) ag_t1 = ag_tx + 1;
      else ag_t2 = ag_tx;
    } while (ag_t1 < ag_t2);
    ag_ap = ag_pstt[ag_t1];
    if ((ag_r_procs_scan[ag_astt[ag_t1]])() == 0) break;
  }
  return (PCB).exit_flag == AG_RUNNING_CODE;
}

static int ag_action_6_proc(void) {
  int ag_sd = ag_fl[ag_ap];
  (PCB).reduction_token = (vpp_token_type) ag_ptt[ag_ap];
  if((PCB).drt == -1) {
    (PCB).drt=(PCB).token_number;
    (PCB).dssx=(PCB).ssx;
    (PCB).dsn=(PCB).sn;
  }
  if (ag_sd) {
    (PCB).sn = (PCB).ss[(PCB).ssx -= ag_sd];
  }
  else {
    ag_prot();
    (PCB).vs[(PCB).ssx] = ag_null_value;
    (PCB).ss[(PCB).ssx] = (PCB).sn;
  }
  (PCB).rx = 0;
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbe[(PCB).sn] + 1;
    unsigned ag_t2 = ag_sbt[(PCB).sn+1] - 1;
    do {
      unsigned ag_tx = (ag_t1 + ag_t2)/2;
      if (ag_tstt[ag_tx] < (const unsigned char)(PCB).reduction_token) ag_t1 = ag_tx + 1;
      else ag_t2 = ag_tx;
    } while (ag_t1 < ag_t2);
    ag_ap = ag_pstt[ag_t1];
    if ((ag_r_procs_scan[ag_astt[ag_t1]])() == 0) break;
  }
  return (PCB).exit_flag == AG_RUNNING_CODE;
}


void init_vpp(void) {
  (PCB).rx = (PCB).fx = 0;
  (PCB).ss[0] = (PCB).sn = (PCB).ssx = 0;
  (PCB).exit_flag = AG_RUNNING_CODE;
  (PCB).line = FIRST_LINE;
   (PCB).column = FIRST_COLUMN;
  (PCB).btsx = 0, (PCB).drt = -1;
}

void vpp(void) {
  init_vpp();
  (PCB).exit_flag = AG_RUNNING_CODE;
  while ((PCB).exit_flag == AG_RUNNING_CODE) {
    unsigned ag_t1 = ag_sbt[(PCB).sn];
    if (ag_tstt[ag_t1]) {
      unsigned ag_t2 = ag_sbe[(PCB).sn] - 1;
      if ((PCB).rx < (PCB).fx) {
        (PCB).input_code = (PCB).lab[(PCB).rx++];
        (PCB).token_number = (vpp_token_type) AG_TCV((PCB).input_code);}
      else {
        GET_INPUT;
        (PCB).lab[(PCB).fx++] = (PCB).input_code;
        (PCB).token_number = (vpp_token_type) AG_TCV((PCB).input_code);
        (PCB).rx++;
      }
      if (ag_key_index[(PCB).sn]) {
        unsigned ag_k = ag_key_index[(PCB).sn];
        unsigned char ag_ch = (unsigned char) CONVERT_CASE((PCB).input_code);
        while (ag_key_ch[ag_k] < ag_ch) ag_k++;
        if (ag_key_ch[ag_k] == ag_ch) ag_get_key_word(ag_k);
      }
      do {
        unsigned ag_tx = (ag_t1 + ag_t2)/2;
        if (ag_tstt[ag_tx] > (const unsigned char)(PCB).token_number)
          ag_t1 = ag_tx + 1;
        else ag_t2 = ag_tx;
      } while (ag_t1 < ag_t2);
      if (ag_tstt[ag_t1] != (const unsigned char)(PCB).token_number)
        ag_t1 = ag_sbe[(PCB).sn];
    }
    ag_ap = ag_pstt[ag_t1];
    (ag_gt_procs_scan[ag_astt[ag_t1]])();
  }
}


